
#include <string.h>
#include <mex.h>
#include <omp.h>
#include <unistd.h>

#include "flexdata.h"
#include "flexdata_client_server.h"


#define COMMAND prhs[0]

#define LOAD_BASEDIRS prhs[1]
#define LOAD_IMAGEMASK prhs[2]
#define LOAD_MMAP prhs[3]
#define LOAD_SERVERS prhs[3]

#define CATPQ_IDX prhs[1]
#define CATPQ_FROM_BB prhs[2]

#define DETECT_IDX prhs[1]
#define DETECT_ALLS_PQ prhs[2]
#define DETECT_L2SQ prhs[3]


/* global variable, removed on cleanup */
static flexdata_t *flexdata = NULL; 
static flexdata_client_t *flexdata_c = NULL; 

void cleanup() {
  flexdata_delete(flexdata);       
  flexdata_client_delete(flexdata_c); 
  flexdata = NULL; flexdata_c = NULL; 
  mexUnlock(); 
}


void mexFunction (int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray*prhs[])

{
  char command[256]; 
  if(nrhs < 1 || mxGetString (COMMAND, command, 256) != 0) 
    mexErrMsgTxt("did not understand command"); 
  
  
  /*****************************************************************
   * Startup / shutdown
   */

  if(!strcmp(command, "load") || !strcmp(command, "load-servers")) {
    
    cleanup();    

    if(!(nrhs == 4 && mxIsCell(LOAD_BASEDIRS) && mxIsInt32(LOAD_IMAGEMASK)))
      mexErrMsgTxt("invalid load command");     
       
    if(!(mxGetNumberOfElements(LOAD_IMAGEMASK) == mxGetNumberOfElements(LOAD_BASEDIRS)))
      mexErrMsgTxt("invalid load arguments"); 

    int nchannel = mxGetNumberOfElements(LOAD_IMAGEMASK);
    char **basedirs = mxMalloc(sizeof(char*) * nchannel); 
    char *buf = mxMalloc(nchannel * 1024), *bufi = buf; 
    int i;
    for(i = 0; i < nchannel; i++) {
      mxArray *dir = mxGetCell(LOAD_BASEDIRS, i); 
      if(mxGetString(dir, bufi, 1024) != 0) 
        mexErrMsgTxt("basedirs should be strings"); 
      basedirs[i] = bufi; 
      bufi += strlen(bufi) + 1; 
    }
    
    if(!strcmp(command, "load")) { 
      if(!(mxIsLogical(LOAD_MMAP) && mxGetNumberOfElements(LOAD_MMAP) == 1)) 
        mexErrMsgTxt("invalid mmap argument");     
      int do_mmap = *mxGetLogicals(LOAD_MMAP);
      flexdata = flexdata_new(nchannel, basedirs, mxGetData(LOAD_IMAGEMASK), do_mmap); 
      mxFree(basedirs); mxFree(buf);    
      if(!flexdata) mexErrMsgTxt("Loading flexdata failed"); 
    } else {
      /* client-server version */
      
      if(!mxIsStruct(LOAD_SERVERS)) mexErrMsgTxt("invalid server list"); 
      int nserver = mxGetNumberOfElements(LOAD_SERVERS);
      flexdata_server_t *servers = mxMalloc(sizeof(flexdata_server_t) * nserver); 

      for(i = 0; i < nserver; i++) {
        mxArray *aname = mxGetField(LOAD_SERVERS, i, "name"); 
        mxArray *aport = mxGetField(LOAD_SERVERS, i, "port"); 
        mxArray *alast = mxGetField(LOAD_SERVERS, i, "last_image"); 
        
        if(!(aname && mxGetString(aname, servers[i].name, 256) == 0 && 
             aport && alast))
          mexErrMsgTxt("invalid server list entry"); 

        servers[i].last_image = (int)mxGetScalar(alast); 
        servers[i].port = (int)mxGetScalar(aport); 
      }
      

      flexdata_c = flexdata_client_new(nserver, servers, nchannel, basedirs, mxGetData(LOAD_IMAGEMASK)); 
      if(!flexdata_c) 
        mexErrMsgTxt("flexdata_client_new failed"); 
      
      mxFree(servers); 

    }
    
    mexLock(); 

  } else if(!strcmp(command, "cleanup")) {
    cleanup(); 

  /*****************************************************************
   * Operations
   */

  } else if(!strcmp(command, "detect")) {
    mxAssert(flexdata || flexdata_c, "no flexdata loaded"); 

    int tot_nsq = flexdata ? flexdata->tot_nsq : flexdata_c->tot_nsq; 

    if(!(nrhs == 4 && mxIsInt32(DETECT_IDX)))
      mexErrMsgTxt("invalid detect index");
    if(!(mxIsDouble(DETECT_ALLS_PQ) && 
         mxGetM(DETECT_ALLS_PQ) == 256 && 
         mxGetN(DETECT_ALLS_PQ) == tot_nsq))
      mexErrMsgTxt("invalid alls_pq argument");

    if(!(mxIsDouble(DETECT_L2SQ) && 
         mxGetM(DETECT_L2SQ) == 256 && 
         mxGetN(DETECT_L2SQ) == tot_nsq))
      mexErrMsgTxt("invalid detect argument");     
    if(nlhs != 2) mexErrMsgTxt("need 2 outputs");     
    
    int nimage = mxGetNumberOfElements(DETECT_IDX); 
    const *imind = mxGetData(DETECT_IDX); 

    int nbox = flexdata ? 
      flexdata_count_nbox(flexdata->nimage, flexdata->nbox, nimage, imind) : 
      flexdata_count_nbox(flexdata_c->nimage, flexdata_c->nbox, nimage, imind); 
    
    plhs[0] = mxCreateNumericMatrix (nbox, 1, mxDOUBLE_CLASS, mxREAL);
    plhs[1] = mxCreateNumericMatrix (nbox, 2, mxDOUBLE_CLASS, mxREAL);

    int ret; 
    if(flexdata) {
      /* for some reason, this is not set properly on input */
      omp_set_num_threads(omp_get_num_procs()); 
      ret = flexdata_compute_bb_scores(flexdata, nimage, imind, 
                                           mxGetPr(DETECT_ALLS_PQ), 
                                           mxGetPr(DETECT_L2SQ), 
                                           mxGetPr(plhs[0]), 
                                           mxGetPr(plhs[1])); 
    } else {
      ret = flexdata_client_compute_bb_scores(flexdata_c, nimage, imind, 
                                           mxGetPr(DETECT_ALLS_PQ), 
                                           mxGetPr(DETECT_L2SQ), 
                                           mxGetPr(plhs[0]), 
                                           mxGetPr(plhs[1]));
    } 
      
    if(!ret) mexErrMsgTxt("flexdata_compute_bb_scores error"); 
        



  } else if(!strcmp(command, "catpq")) {
    mxAssert(flexdata || flexdata_c, "no flexdata loaded"); 
    
    if(!(nrhs == 3 && 
         mxIsInt32(CATPQ_IDX) && mxGetN(CATPQ_IDX) == 2 && 
         mxIsLogical(CATPQ_FROM_BB) && mxGetNumberOfElements(CATPQ_FROM_BB) == 1))
      mexErrMsgTxt("invalid catpq argument");     

    if(nlhs != 2) mexErrMsgTxt("need 2 outputs");     

    int from_bb = *mxGetLogicals(CATPQ_FROM_BB);
    int nsample = mxGetM(CATPQ_IDX); 
    const int *idx = mxGetData(CATPQ_IDX); 
    
    
    int nchannel = flexdata ? flexdata->nchannel : flexdata_c->nchannel; 
    int tot_nsq = flexdata ? flexdata->tot_nsq : flexdata_c->tot_nsq; 
    plhs[0] = mxCreateNumericMatrix (nchannel, nsample, mxSINGLE_CLASS, mxREAL);
    plhs[1] = mxCreateNumericMatrix (tot_nsq, nsample, mxUINT8_CLASS, mxREAL);
    int ret; 

    if(flexdata) {
      /* for some reason, this is not set properly on input */
      omp_set_num_threads(omp_get_num_procs()); 

      ret = flexdata_sample_descriptors(flexdata, nsample, from_bb, idx, idx + nsample, 
                                        mxGetData(plhs[0]), mxGetData(plhs[1]));
    } else {
      ret = flexdata_client_sample_descriptors(flexdata_c, nsample, from_bb, idx, idx + nsample, 
                                        mxGetData(plhs[0]), mxGetData(plhs[1]));
    }
    if(!ret) mexErrMsgTxt("flexdata_sample_descriptors error"); 
  } else 
    mexErrMsgTxt("Unknown command");   

}
