#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/socket.h>
#include <arpa/inet.h>
 
#include <sys/types.h>
#include <netdb.h>
#include <unistd.h>


#include "flexdata_client_server.h"

typedef enum {
  COMMAND_LOAD = 1234,
  COMMAND_CATPQ, 
  COMMAND_DETECT
} command_t; 


#define NEWA(type,n) (type*)malloc(sizeof(type)*(n))
#define NEWAC(type,n) (type*)calloc(n,sizeof(type))
#define NEW(type) NEWA(type,1)



static int recv_all(int sock, void *ptr, size_t sz) {
  char *c = ptr;
  while(sz) {
    size_t ret = recv(sock, c, sz, 0);
    if(ret == (size_t)(-1) || ret == 0) return 0;
    c += ret; 
    sz -= ret;
  }
  return 1; 
}

#define RECV(var, n) \
  if(!recv_all(sock, (var), sizeof((var)[0])*(n))) {     \
  fprintf(stderr, "receive error line %d\n", __LINE__); \
  goto fail; \
}

static int send_all(int sock, const void *ptr, size_t sz) {
  const char *c = ptr;
  while(sz) {
    size_t ret = send(sock, c, sz, 0);
    if(ret == (size_t)(-1) || ret == 0) return 0;
    c += ret; 
    sz -= ret;
  }
  return 1; 
}

#define SEND(var, n) \
  if(!send_all(sock, (var), sizeof((var)[0])*(n))) {     \
  fprintf(stderr, "send error line %d\n", __LINE__); \
  goto fail; \
}



/*****************************************************************
 *  Client code  
 *****************************************************************/

int connect_to_server(const char *servname, int port) {
  int sock;
     
  /* Address lookup */
  struct addrinfo ai, *ao;
  memset(&ai, 0, sizeof(ai)); 
  ai.ai_family = AF_INET; 
  ai.ai_socktype = SOCK_STREAM;
  char port_string[10]; 
  sprintf(port_string, "%d", port); 
  if(getaddrinfo(servname, port_string, &ai, &ao) != 0) {
    fprintf(stderr, "server %s:%d ", servname, port);
    perror("getaddrinfo"); 
    return -1; 
  }

  sock = socket(ao->ai_family, ao->ai_socktype, ao->ai_protocol);
  if (sock == -1) {
    fprintf(stderr, "server %s:%d ", servname, port);
    perror("socket"); 
    return -1;
  }

  if (connect(sock, ao->ai_addr, ao->ai_addrlen) < 0) {
    fprintf(stderr, "server %s:%d ", servname, port);
    perror("socket"); 
    close(sock);
    return -1;
  }
  freeaddrinfo(ao);
  fprintf(stderr, "connected to server %s:%d\n", servname, port);
  return sock;
}



flexdata_client_t *flexdata_client_new(int nserver, 
                                       flexdata_server_t *servers,
                                       int nchannel, 
                                       const char **dirnames, 
                                       const int *immasks) {
  int i; 
  flexdata_client_t *fc = NEWAC(flexdata_client_t, 1); 
  fc->nserver = nserver; 
  fc->nchannel = nchannel; 
  fc->servers = NEWAC(flexdata_server_t, nserver); 
  char *dirname_buf = NEWA(char, nchannel * 1024); 
  int dirname_buf_sz; 
  {
    char *w = dirname_buf; 
    for(i = 0; i < nchannel; i++) {
      strcpy(w, dirnames[i]); 
      w += strlen(dirnames[i]) + 1; 
    }
    dirname_buf_sz = w - dirname_buf;
  }

  
  for(i = 0; i < nserver; i++) {
    flexdata_server_t *s = fc->servers + i;
    const flexdata_server_t *sin = servers + i;
    strcpy(s->name, sin->name); 
    s->port = sin->port; 
    s->last_image = sin->last_image; 
    int sock = s->sock = connect_to_server(s->name, s->port); 
    if(sock <= 0) goto fail; 
    s->sock = sock;
         

    /* emit load command */
    command_t cmd = COMMAND_LOAD; 
    SEND(&cmd, 1);
    SEND(&nchannel, 1); 
    SEND(&dirname_buf_sz, 1); 
    SEND(dirname_buf, dirname_buf_sz); 
    SEND(immasks, nchannel); 
  }
  
  /* all servers are loading */

  /* collect boxes */
  for(i = 0; i < nserver; i++) {
    int nimage;     
    int sock = fc->servers[i].sock;
    RECV(&nimage, 1); 
    RECV(&fc->tot_nsq, 1); 
    uint64 *nbox = NEWA(uint64, nimage); 
    RECV(nbox, nimage); 
    if(i == 0) {
      fc->nimage = nimage; 
      fc->nbox = nbox; 
    } else {
      if(nimage != fc->nimage || memcmp(fc->nbox, nbox, sizeof(*nbox) * nimage) != 0) {
        fprintf(stderr, "inconsistent nb of boxes\n"); 
        goto fail;         
      }
      free(nbox);        
    }
  } 
  fc->cum_nbox = NEWA(uint64, fc->nimage + 1); 
  fc->im_to_serv = NEWA(int, fc->nimage);
  fc->cum_nbox[0] = 0;
  
  int s = 0; 
  for(i = 0; i < fc->nimage; i++) {    
    fc->cum_nbox[i + 1] = fc->cum_nbox[i] + fc->nbox[i];
    while(s < fc->nserver && i >= fc->servers[s].last_image) s++; 
    fc->im_to_serv[i] = s;
  }
  if(s == fc->nserver) {
    fprintf(stderr, "inconsistent last_image\n"); 
    goto fail; 
  }
  
  free(dirname_buf); 
  return fc; 
 fail: 
  flexdata_client_delete(fc); 
  free(dirname_buf); 
  return NULL; 
}

void flexdata_client_delete(flexdata_client_t *fc) {
  if(!fc) return; 
  int i;
  for(i = 0; i < fc->nserver; i++) {
    if(fc->servers[i].sock > 0) close(fc->servers[i].sock); 
  }
  free(fc->nbox); 
  free(fc->cum_nbox); 
  free(fc->im_to_serv);
  free(fc); 
}

int flexdata_client_sample_descriptors(flexdata_client_t * fc,
                                       int nsample,
                                       int from_bb,
                                       const int *imind,
                                       const int *bbind,
                                       float *npts,
                                       uint8 *pqcodes) {
  int nserver = fc->nserver; 
  int *iminds = NEWA(int, nsample); 
  int *bbinds = NEWA(int, nsample);   
  int ret = 0; 
  long i, j;

  for(i = 0; i < nserver; i++) {
    int ns = 0; 
    for(j = 0; j < nsample; j++) {      
      if(fc->im_to_serv[imind[j]] == i) {
        iminds[ns] = imind[j]; 
        bbinds[ns] = bbind[j];
        ns++; 
      }
    }
    if(ns == 0) continue; 
    int sock = fc->servers[i].sock;
    command_t cmd = COMMAND_CATPQ; 
    SEND(&cmd, 1);
    SEND(&ns, 1); 
    SEND(&from_bb, 1); 
    SEND(iminds, ns); 
    SEND(bbinds, ns);     
  }

  /* servers processing */

  for(i = 0; i < nserver; i++) {
    int *map = iminds; 
    uint64 ns = 0; 
    for(j = 0; j < nsample; j++) {      
      if(fc->im_to_serv[imind[j]] == i) {
        map[ns] = j; 
        ns++; 
      }
    }
    if(ns == 0) continue; 
    float *nptss = NEWA(float, fc->nchannel * ns); 
    uint8 *pqcodess = NEWA(uint8, fc->tot_nsq * ns); 
    int sock = fc->servers[i].sock;
    RECV(nptss, fc->nchannel * ns); 
    RECV(pqcodess, fc->tot_nsq * ns); 

    for(j = 0; j < ns; j++) {
      size_t ld = fc->nchannel; 
      memcpy(npts + map[j] * ld, nptss + j * ld, sizeof(*npts) * ld); 
      ld = fc->tot_nsq; 
      memcpy(pqcodes + map[j] * ld, pqcodess + j * ld, ld); 
    }    
    free(nptss); 
    free(pqcodess);     
  }
  ret = 1;

 fail:
  free(iminds); free(bbinds);  
  return ret; 
}


int flexdata_client_compute_bb_scores(const flexdata_client_t * fc,
                                      int nimage,
                                      const int *imind,
                                      const double *alls_pq,
                                      const double *l2normsq,
                                      double *scores,
                                      double *npts_stats) {
  int nserver = fc->nserver; 
  int *iminds = NEWA(int, nimage); 
  int ret = 0; 
  int i, j;

  for(i = 0; i < nserver; i++) {
    int ni = 0; 
    for(j = 0; j < nimage; j++) {      
      if(fc->im_to_serv[imind[j]] == i) {
        iminds[ni] = imind[j]; 
        ni++; 
      }
    }
    if(ni == 0) continue; 
    int sock = fc->servers[i].sock;
    command_t cmd = COMMAND_DETECT; 
    SEND(&cmd, 1);
    SEND(&ni, 1); 
    SEND(iminds, ni); 
    SEND(alls_pq, 256 * fc->tot_nsq);     
    SEND(l2normsq, 256 * fc->tot_nsq);     
  }

  /* servers processing */

  /* cumulated nb of boxes per image in imind */
  uint64 * cum_nbox = NEWA(uint64, nimage + 1);   
  cum_nbox[0] = 0;
  for(i = 0; i < nimage; i++) 
    cum_nbox[i + 1] = cum_nbox[i] + fc->nbox[imind[i]]; 

  for(i = 0; i < nserver; i++) {
    int *map = iminds;
    int ni = 0; 
    for(j = 0; j < nimage; j++) {      
      if(fc->im_to_serv[imind[j]] == i) {
        map[ni] = j; 
        ni++; 
      }
    }
    if(ni == 0) continue; 
    int sock = fc->servers[i].sock;
    for(j = 0; j < ni; j++) {
      int si = map[j];
      int imi = imind[si];
      RECV(scores + cum_nbox[si], fc->nbox[imi]); 
    }
    for(j = 0; j < ni; j++) {
      int si = map[j];
      int imi = imind[si];
      RECV(npts_stats + cum_nbox[si] * 2, fc->nbox[imi] * 2); 
    }

  }
  ret = 1;

 fail:
  free(iminds); 
  free(cum_nbox); 
  return ret; 

}


/*****************************************************************
 *  Server code  
 *****************************************************************/

#ifdef SERVER_EXECUTABLE
 
static flexdata_t *server_new_connection(int sock) {
  flexdata_t *f = NULL; 
  command_t command; 
  RECV(&command, 1); 
  if(command != COMMAND_LOAD)  {
    fprintf(stderr, "new connection: weird command received %d\n", command); 
    goto fail; 
  }
  int dirname_buf_sz, nchannel;
  RECV(&nchannel, 1); 
  RECV(&dirname_buf_sz, 1); 
  char *dirname_buf = NEWA(char, dirname_buf_sz); 
  RECV(dirname_buf, dirname_buf_sz); 
  int *immasks = NEWA(int, nchannel); 
  RECV(immasks, nchannel); 
  char **dirnames = NEWA(char*, nchannel);
  {
    int i; 
    char *w = dirname_buf; 
    for(i = 0; i < nchannel; i++) {
      dirnames[i] = w; 
      w += strlen(w) + 1; 
    }
  }
      
  flexdata_t *fx = flexdata_new(nchannel, dirnames, immasks, 1); 
  if(!fx) goto fail; 
  SEND(&fx->nimage, 1); 
  SEND(&fx->tot_nsq, 1); 
  SEND(fx->nbox, fx->nimage); 
  f = fx;
 fail: 
  free(dirname_buf); 
  free(dirnames); 
  free(immasks); 
  return f;
}

int server_catpq(flexdata_t *f, int sock) {
  int ret = 0; 

  int nsample;
  int from_bb;
  int *imind = NULL;
  int *bbind = NULL;
  float *npts = NULL;
  uint8 *pqcodes = NULL;
  
  RECV(&nsample, 1); 
  printf("processing catpq command for %d samples\n", nsample);
  RECV(&from_bb, 1); 
  imind = NEWA(int, nsample); 
  RECV(imind, nsample); 
  bbind = NEWA(int, nsample); 
  RECV(bbind, nsample); 
  
  /* start processing */
  npts = NEWA(float, f->nchannel * (uint64)nsample); 
  pqcodes = NEWA(uint8, f->tot_nsq * (uint64)nsample); 
      
  int retx = flexdata_sample_descriptors(f, nsample, from_bb, imind, bbind, npts, pqcodes);       
  if(!retx) goto fail; 
  
  /* send back result */
  SEND(npts, f->nchannel * (uint64)nsample); 
  SEND(pqcodes, f->tot_nsq * (uint64)nsample); 
  
  ret = 1;
 fail:
  free(imind); free(bbind); 
  free(npts); free(pqcodes); 

  return ret;  
}

int server_detect(flexdata_t *f, int sock) {
  int ret = 0;
  int nimage;
  int *imind = NULL;
  double *alls_pq = NULL;
  double *l2normsq = NULL;
  double *scores = NULL;
  double *npts_stats = NULL;
  
  RECV(&nimage, 1); 
  printf("Processing detect command for %d images\n", nimage); 
  imind = NEWA(int, nimage); 
  RECV(imind, nimage); 
  alls_pq = NEWA(double, 256 * f->tot_nsq);     
  RECV(alls_pq, 256 * f->tot_nsq);     
  l2normsq = NEWA(double, 256 * f->tot_nsq);     
  RECV(l2normsq, 256 * f->tot_nsq);     
  
  int tot_nbox = flexdata_count_nbox(f->nimage, f->nbox, nimage, imind); 
  printf("  total %d boxes\n", tot_nbox); 
  scores = NEWA(double, tot_nbox); 
  npts_stats = NEWA(double, tot_nbox * 2);
  
  int retx = flexdata_compute_bb_scores(f, nimage, imind, alls_pq, l2normsq, scores, npts_stats); 
  if(!retx) goto fail; 
  
  SEND(scores, tot_nbox); 
  SEND(npts_stats, tot_nbox * 2);      
  
  ret = 1; 
 fail: 
  free(imind); free(alls_pq); free(l2normsq); 
  free(scores); free(npts_stats); 
  return ret; 
}

int server_process_command(flexdata_t *f, int sock) {
  command_t cmd; 
  size_t ret = recv(sock, &cmd, sizeof(cmd), 0);

  if(ret == 0) {
    /*  The return value will be 0 when the peer  has  performed  an
        orderly shutdown.*/
    return 0;
  }

  switch(cmd) {
  case COMMAND_CATPQ: 
    return server_catpq(f, sock); 
  case COMMAND_DETECT: 
    return server_detect(f, sock); 
    break;
  default:
    fprintf(stderr, "Unknown command %d\n", cmd); 
    return 0;
  }
} 



void flexdata_server_start(int port) {
  int socket_desc;
  struct sockaddr_in server, client;
     
  //Create socket
  socket_desc = socket(AF_INET, SOCK_STREAM, 0);
  assert(socket_desc != -1);
     
  //Prepare the sockaddr_in structure
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons( port );
     
  //Bind
  if( bind(socket_desc,(struct sockaddr *)&server, sizeof(server)) < 0) {
    perror("bind failed");
    exit(1); 
  }
  printf("Waiting for connections on port %d\n", port);
     
  //Listen
  listen(socket_desc, 3);
  
  do {
    int c = sizeof(struct sockaddr_in);
    int client_sock = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c);
    if (client_sock < 0) {
      perror("accept failed");
      exit(2);
    }

    flexdata_t *f = server_new_connection(client_sock); 
    if(!f) {
      close(client_sock); 
      continue; 
    }

    printf("server ready, flexdata with %d images, %d channels, total nsq %d, total %ld boxes\n", 
           f->nimage, f->nchannel, f->tot_nsq, f->cum_nbox[f->nimage]);
    
    while(server_process_command(f, client_sock)) ; 

    printf("closing connection\n"); 

    flexdata_delete(f);
    close(client_sock); 

  } while(1); 

}





int main(int argc, char**args) {
  int port; 
  if(argc != 2 || sscanf(args[1], "%d", &port) != 1) {
    fprintf(stderr, "usage: %s port\n", args[0]); 
    return 1; 
  }
  flexdata_server_start(port);
    
  return 0; 
}

#endif
