
function compute_descriptors()

  index = dataset_definition();
  
  % make output directory
  mkdir data/fisher_vectors
  
  for desc = {'sift', 'color'}
    desc = char(desc);

    % SPM grid
    if isequal(desc, 'sift')
      % grids =  {[1 1],[4 4]}; % for full dataset
      grids =  {[1 1], [2 1]}; % for toy dataset
    else
      grids = {[1 1]};
    end
    
    pca_file = ['data/trained/local_PCA_' desc '.mat'];
    load(pca_file)   

    gmm_file = ['data/trained/local_GMM_' desc '.mat'];
    load(gmm_file);

    for src = {'boxes', 'masks'}
      src = char(src); 
      pq_file = [ 'data/trained/PQ_centroids_' desc '_for_' src '.mat']; 
      load(pq_file); 
      
      compute_descriptors_1(index, desc, src, PCA, dict, pq, grids); 
            
    end
        
  end

end


  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Compute for a given descriptor and source

function compute_descriptors_1(index, desc, src, PCA, dict, pq, grids)
  
  ncell = sum(prod(vertcat(grids{:}),2)); % number of spatial cells
  nimg = length(index.imnames);
  
  fprintf('**** Processing %d images with descriptor %s, source %s, %d grid cells\n', ...
          nimg, desc, src, ncell); 
    
  cn_setrandseed(101);

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Prepare bboxes to extract
  
  x = load('data/masks/bbs.mat'); 
  candbbs = x.boxes;
  
  extractbbs = struct();  
  for imi = nimg:-1:1
    
    extractbbs(imi).orgimi = imi;
    imsize = index.imsizes(imi, :); 
    
    extractbbs(imi).nbb = size(candbbs(imi).bbs, 1);
    extractbbs(imi).ngt = size(index.gtboxes(imi).bbs, 1);
        
    % extract_name = bb gt im, in this order. 
    extractbbs(imi).bbs = [
        candbbs(imi).bbs                   % candidate windows
        int16(index.gtboxes(imi).bbs)      % ground-truth windows
        ];
    if isequal(src, 'boxes')
      extractbbs(imi).bbs = [ extractbbs(imi).bbs ; int16([1 1 imsize([2 1])])];         % full image
    end
    
    assert(isa(extractbbs(imi).bbs,'int16'));
  end

  
  all_dirs = {};

  % prepare output directories
  for grid = grids
    
    idim = grid{1}(1);
    jdim = grid{1}(2);
    
    % order of i,j loops matters
    for jgrid = 1:jdim
      for igrid = 1:idim
        
        c_outroot = sprintf('data/fisher_vectors/%s_%s_%dx%di%dj%d', desc, src, idim, jdim, igrid, jgrid);
        fprintf('  output descriptors to %s\n', c_outroot);
        
        mkdir(c_outroot); 
                
        all_dirs{end + 1} = c_outroot; 
        
      end
    end
  end
 
  % do the processing with chunks in parallel.  
  ncell = length(all_dirs); 
  
  % For multiprocessing, images are divided in batches (chunks), which are encoded independently
  step = 50; % images per chunk
  chunk_begins = 1:step:nimg; 
  chunk_ends = min(chunk_begins + step - 1, nimg); 
  
  nchunk = length(chunk_begins); 
    
  params = {};  
  for chunki = 1:nchunk
    image_nums = chunk_begins(chunki) : chunk_ends(chunki);
    params{chunki} = {index, desc, src, PCA, dict, pq, grids, extractbbs(image_nums), all_dirs, chunki, image_nums};
  end
  
  startmulticoremaster(@compute_chunk_descriptors, params);  

  fprintf('  merging per-image data\n');

  for ci = 1:ncell
    fprintf('    merging chunks for cell=%d/%d\n', ci, ncell); 
    
    merge_file_chunks(all_dirs{ci}, 'bb', chunk_begins, chunk_ends); 
    merge_file_chunks(all_dirs{ci}, 'gt', chunk_begins, chunk_ends); 
    if isequal(src, 'boxes')      
      merge_file_chunks(all_dirs{ci}, 'im', chunk_begins, chunk_ends); 
    end
  end
  
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Process a chunk of images

function compute_chunk_descriptors(index, desc, src, PCA, dict, pq, grids, ...
                                   extractbbs, all_dirs, chunki, image_nums) 

  ncell = length(all_dirs); 

  % check if we already processed this
  all_done = true; 
  for celli = 1:ncell    
    all_done = all_done && files_ok(all_dirs{celli}, 'bb', chunki) && ...
                           files_ok(all_dirs{celli}, 'gt', chunki) && ...
                           (~isequal(src, 'boxes') || files_ok(all_dirs{celli}, 'im', chunki));
  end

  if all_done
    fprintf('Warn: skipping chunk %d: already processed\n', chunki)
    return
  end

  % open all output files
  all_writer = {};
  for celli = 1:ncell    
    all_writer{celli}.bb = open_files(all_dirs{celli}, 'bb', chunki); 
    all_writer{celli}.gt = open_files(all_dirs{celli}, 'gt', chunki); 
    if isequal(src, 'boxes')
      all_writer{celli}.im = open_files(all_dirs{celli}, 'im', chunki); 
    end
  end      
  for outi = 1:length(image_nums)
    imi = image_nums(outi);

    compute_image_descriptors(index.imgroot, desc, src, PCA, pq, dict, grids, ...
                              index.imnames{imi}, index.imsizes(imi, :), extractbbs(outi), all_writer); 
  end

  % close outputs
  for celli = 1:ncell
    close_files(all_writer{celli}.bb, all_dirs{celli}, 'bb', chunki); 
    close_files(all_writer{celli}.gt, all_dirs{celli}, 'gt', chunki); 
    if isequal(src, 'boxes')
      close_files(all_writer{celli}.im, all_dirs{celli}, 'im', chunki); 
    end
  end    

end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Process one image

function compute_image_descriptors(imgroot, desc, src, PCA, pq, dict, grids, ...
                                   imname, imsize, eb, all_writer)

  ncell = length(all_writer); 

  nobj = size(eb.bbs, 1);   
  img_filename = [imgroot imname '.jpg'];    

  hasim = '';
  if isequal(src, 'boxes')
    hasim = ' + im 1';
  end    
  fprintf('  computing %d (bb %d + gt %d%s) compressed Fisher vectors from %s features for image %s\n', ...
          nobj, eb.nbb, eb.ngt, hasim, desc, imname);
    
  % Handle mask 
  if isequal(src, 'masks')
    masks_filename = ['data/masks/' imname '.mat'];    
    maskout = load(masks_filename);  
    masks = maskout.masks;  

    if eb.ngt > 0
      % we need a mask for the GT windows also
      
      seg_filename = [ 'data/segmentation/' imname '.mat'];
      trees = load( seg_filename);
      trees = trees.trees; 
      
      [x_bbs, masks_2] = fv_candbox_superpixelsinside( struct2('ovth',1','superpixelseg','avg','ifempty','zeros'), ...
                                                       eb.bbs(eb.nbb + 1:end, :), trees );
      assert(length(masks_2) == eb.ngt)
      for i = 1:eb.ngt
        masks{i + eb.nbb} = masks_2{i}; 
      end
    end
    
  else
    masks = [];
  end

  % compute FVs for the given bboxes
  [per_cell_descs, per_cell_npts] = fv_from_samples(img_filename, imsize, ...
                                                    desc, eb.bbs, PCA, dict, grids, masks);

  
  % store to flexdata writers  
  for celli = ncell:-1:1
    xdesc = per_cell_descs{celli}; 
    xcellnpts = per_cell_npts{celli};       
    
    % PQ quantization
    xpqc = pq_assign(pq, single(xdesc)); % (nsq ntotal)
    xcellnpts2 = xcellnpts(:); 
    
    files = all_writer{celli}; 
    
    i = 1; j = eb.nbb;
    write_fv(files.bb, rowvec(xcellnpts2(i:j)), xpqc(:, i:j));
    
    i = j + 1; j = j + eb.ngt;
    write_fv(files.gt, rowvec(xcellnpts2(i:j)), xpqc(:, i:j));
    
    if isequal(src, 'boxes')
      i = j + 1; j = j + 1;
      write_fv(files.im, rowvec(xcellnpts2(i:j)), xpqc(:, i:j));
    end     
    
  end

  
end




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Operations on the 5 files in that encode the descriptors (flexdata format)
% Originally, flexdata is much more flexible than this fixed format, but let's 
% keep the code simple

function fh = open_files(dirname, fieldname, chunki)
  fh(1) = fopen(sprintf('%s/__%d__npts_%s.info', dirname, chunki, fieldname), 'w'); 
  fh(2) = fopen(sprintf('%s/__%d__npts_%s.data', dirname, chunki, fieldname), 'w'); 
  fh(3) = fopen(sprintf('%s/__%d__pq_%s.info', dirname, chunki, fieldname), 'w'); 
  fh(4) = fopen(sprintf('%s/__%d__pq_%s.bloscinfo', dirname, chunki, fieldname), 'w'); 
  fh(5) = fopen(sprintf('%s/__%d__pq_%s.blosc', dirname, chunki, fieldname), 'w'); 
  assert(min(fh) > -1);
end


function close_files(fh, dirname, fieldname, chunki)
  for i = 1:5
    fclose(fh(i));
  end
  % put an empty file to say we are done
  okfile = fopen(sprintf('%s/__%d__npts_%s.ok', dirname, chunki, fieldname), 'w'); fclose(okfile);   
  okfile = fopen(sprintf('%s/__%d__pq_%s.ok', dirname, chunki, fieldname), 'w'); fclose(okfile);   
end

function ok = files_ok(dirname, fieldname, chunki)
  ok = exist(sprintf('%s/__%d__npts_%s.ok', dirname, chunki, fieldname), 'file') && ...
       exist(sprintf('%s/__%d__pq_%s.ok', dirname, chunki, fieldname), 'file');
end


function write_fv(fh, npts, desc)
  % write Fisher vector + nb of pts 
    
  %%%%%% write npts
  % metadata    
  fwrite(fh(1), size(npts), 'uint64'); 
  % data 
  fwrite(fh(2), npts, 'single'); 
  
  %%%%%% write descriptor 
  
  % metadata
  fwrite(fh(3), size(desc), 'uint64');

  compressed_data = compress_blosc_mex(desc); 
  % remember compressed size
  fwrite(fh(4), length(compressed_data), 'uint64');
  % data
  fwrite(fh(5), compressed_data, 'uint8');
  
end


function merge_file_chunks(dirname, fieldname, chunk_begins, chunk_ends) 
  % all files are merged by concatenating them. Info files get an additional header.

  % open output files
  fh(1) = fopen(sprintf('%s/npts_%s.info', dirname, fieldname), 'w'); 
  fh(2) = fopen(sprintf('%s/npts_%s.data', dirname, fieldname), 'w'); 
  fh(3) = fopen(sprintf('%s/pq_%s.info', dirname, fieldname), 'w'); 
  fh(4) = fopen(sprintf('%s/pq_%s.bloscinfo', dirname, fieldname), 'w'); 
  fh(5) = fopen(sprintf('%s/pq_%s.blosc', dirname, fieldname), 'w'); 
  assert(min(fh) > -1);
  
  % write headers
  N = chunk_ends(end);
  header_npts = [ 1     % version; /* 1 */
                  2     % iftype;  /* 0: unsigned integer, 1: signed integer, 2: floating */
                  4     % nbytes; /* 1,2,4,8  number of bytes per item. Example, iftype=2 & nbyte=8 -> double */
                  2     % ndims; /* Number of dimensions in the "sizes" array. */
                  N ];  % N;     /*   Number of instances. */
  fwrite(fh(1), header_npts, 'uint64'); 
  header_codes = [ 1     % version; /* 1 */
                   0     % iftype;  /* 0: unsigned integer, 1: signed integer, 2: floating */
                   1     % nbytes; /* 1,2,4,8  number of bytes per item. Example, iftype=2 & nbyte=8 -> double */
                   2     % ndims; /* Number of dimensions in the "sizes" array. */
                   N ];  % N;     /*   Number of instances. */
  fwrite(fh(3), header_codes, 'uint64');   
  fwrite(fh(4), N, 'uint64');   
  
  
  % merge chunks
  for chunki = 1:length(chunk_ends)
    okname1 = sprintf('%s/__%d__npts_%s.ok', dirname, chunki, fieldname); 
    okname2 = sprintf('%s/__%d__pq_%s.ok', dirname, chunki, fieldname); 
    assert(exist(okname1, 'file') & exist(okname2, 'file')); % minimal sanity check
    append_to_handle(fh(1), sprintf('%s/__%d__npts_%s.info', dirname, chunki, fieldname)); 
    append_to_handle(fh(2), sprintf('%s/__%d__npts_%s.data', dirname, chunki, fieldname)); 
    append_to_handle(fh(3), sprintf('%s/__%d__pq_%s.info', dirname, chunki, fieldname)); 
    append_to_handle(fh(4), sprintf('%s/__%d__pq_%s.bloscinfo', dirname, chunki, fieldname)); 
    append_to_handle(fh(5), sprintf('%s/__%d__pq_%s.blosc', dirname, chunki, fieldname));     
  end   

  % close outputs
  for i = 1:5
    fclose(fh(i));
  end
  
  % put an empty file to say we are done
  okfile = fopen(sprintf('%s/npts_%s.ok', dirname, fieldname), 'w'); fclose(okfile);   
  okfile = fopen(sprintf('%s/pq_%s.ok', dirname, fieldname), 'w'); fclose(okfile);   
  
  % delete temp files
  system(sprintf('rm %s/__*__*_%s.*', dirname, fieldname)); 
  
end

function append_to_handle(fh, filename)
  % writes the contents of filename to the open file handle fh
  f = fopen(filename, 'r'); 
  assert(f > -1);
  a = fread(f, 'uint8'); 
  fwrite(fh, a, 'uint8');  
  fclose(f);
end
