
function evaluate_classifiers()

  index = dataset_definition();

  descriptors = prepare_descriptors(1); 

  % geometrical data for training boxes
  
  test_candbbs = load_candidate_boxes(index, index.all_testmask);
     
  fprintf('Loaded %d test candidate boxes\n', sample_size(test_candbbs));

  sum_AP = 0;

  for classi = 1:length(index.clsnames)

    fprintf('*** testing classifier for class %s\n', index.clsnames{classi});
    
    filename = [ 'data/classifiers/classifier_' index.clsnames{classi} '.mat'];
    fprintf('  Loading classifier %s\n', filename);
    load(filename);
    
    
    % score all images 
    S = classify_all_boxes(classifier, descriptors, test_candbbs);

    detinfo = {};
    for imi = 1:length(S)
      Si = S(imi);
      orgimi = Si.orgimi; 

      %%% filter detection bboxes
      
      % remove boxes that have a dimension < 10/255th of image size 
      Si = filter_by_minsz(Si, 10); 

      
      % map detection bounding boxes to full image coordinates
      bbfull = bbs_to_full(Si.bbuint8, index.imsizes(orgimi, :));

      selind = samples_nonmax_suppression(bbfull, Si.scores, 0.3, -Inf, 10);
      
      det_x1y1x2y2 = bbfull(selind, :); 
      scores = Si.scores(selind); 
      
      %%% At this point, the scored boxes for imi are available 
      %%% (for use in a real detection application...)
                
      
      %%% prepare relevant ground truth
            
      gtboxes = index.gtboxes(orgimi);
      gtm = (gtboxes.clsind == classi); 
      gt_x1y1x2y2 = gtboxes.bbs(gtm,:);
      difficultind = find(gtboxes.diffmask(gtm));
            
      detinfoi = eval_detections(det_x1y1x2y2, scores, gt_x1y1x2y2, difficultind);
      detinfo{imi} = detinfoi;
    end

    detinfo = [detinfo{:}];
  
    
    %%% Per-box scoring
    
    scores      = [detinfo.scores]';
    gt          = [detinfo.gt]';

    %%% remove difficult GT detections
    
    validmask       = ~[detinfo.isdiffdet];
    scores          = scores(validmask);
    gt              = gt(validmask);
    
    numitem = size(gt,1);
    ntotalp = sum(gt,1);
    ntotaln = numitem - ntotalp;
    [scores,si]=sort(scores,1,'descend');
    gt = gt(si);
    tp = cumsum(gt,1);     % true positive
    fp = cumsum(1 - gt,1); % false positive
    
    % remove undetected boxes
    vmask = (scores ~= -inf);
    tp = tp(vmask);
    fp = fp(vmask);
    scores = scores(vmask);
    nsame = sum(sum(diff(scores,1,1)==0,1));
    if nsame > 0 
      fprintf('Warn: score pairs are equal, where PreRecall,etc. evaluation is unreliable.');
    end
    
    % compute precision and recall
    rec = tp ./ ntotalp;
    pre = tp ./ (fp + tp);

    % compute AP
    
    if 1 % VOC 2007 evaluation (interpolated AP)
      nbins = 11; % 11 in VOC.
      ap = 0;
      for t = linspace(0,1,nbins)
        p = max( pre( rec>=t ) ); 
        if isempty(p)
          p = 0;
        end
        ap = ap + p / nbins;
      end
    else % VOC 2010 evaluation
      mrec=[0 ; rec ; 1];
      mpre=[0 ; pre ; 0];
      for i=numel(mpre)-1:-1:1
        mpre(i)=max(mpre(i),mpre(i+1));
      end
      i=find(mrec(2:end)~=mrec(1:end-1))+1;
      ap=sum((mrec(i)-mrec(i-1)).*mpre(i));
    end

    sum_AP = sum_AP + ap; 

    fprintf('   AP=%.2f %% maxRecall=%.2f %%\n', 100 * ap, 100 * max(rec));
        
  end
  fprintf('   mAP=%.2f %%\n', 100 * sum_AP / length(index.clsnames));
    
end


function ss = sample_size(samples)
  ss = 0;
  for i = 1:length(samples)
    ss = ss + size(samples(i).bbuint8, 1);
  end
end

function Si = filter_by_minsz(Si, minsz)
  th = double(minsz);
  x = double( Si.bbuint8 ); 
  
  m = ( (x(:,4)-x(:,2)+1)>=th ) & ( (x(:,3)-x(:,1)+1)>=th );
  
  Si.bbuint8 = Si.bbuint8(m, :); 
  Si.scores = Si.scores(m);  
  
end


function bbfull = bbs_to_full(bbs, imsize)
  % scale back from uint8 image coordinates to pixel coordinates
      
  bbs = double(bbs);
  xscale = (imsize(2) - 1) / 255;
  yscale = (imsize(1) - 1) / 255;
  scales = [xscale yscale xscale yscale];
  bbfull = bsxfun(@times, bbs, scales) + 1;
  
end

function detinfo = eval_detections(det_x1y1x2y2, scores, gt_x1y1x2y2, difficultind) 

  minoverlap = 0.5;
  detinfo.scores = scores'; 
  detinfo.det_x1y1x2y2 = det_x1y1x2y2;
  
  % compute overlaps
      
  ovrmat = cn_bboverlap(gt_x1y1x2y2, det_x1y1x2y2); 
  [c_maxovr c_gtind] = max(ovrmat,[],1);
  c_ngt = size(gt_x1y1x2y2, 1);
  c_ndet = size(det_x1y1x2y2, 1);
  
  scores = reshape(scores, 1, []);
  [foo sortind] = sort(scores,'descend');

  % add extra fields
  c_hasdetected = false(1,c_ngt); % whether has been detected
  detinfo.gt = false(1,c_ndet); % whether a correct detection
  detinfo.evalcase = zeros(1,c_ndet); % remaining ones will be evalcaseconsts.diffdet, which is 0.
  detinfo.isdiffdet = false(1,c_ndet); % difficult mask -- to be skipped
    % NO! detinfo.scores = detinfo.scores(sortind); 

  % create a mask for difficult objects in groundtruth
  c_gtisdiffdet = false(1,c_ngt); 
  c_gtisdiffdet(difficultind) = 1;

  % VOCevaldet also works in this order.
  for obji = reshape(sortind,1,[])
    % associate with gt object that has the largest overlap
    % DONT try to find some gt object that hasnt been detector (which would increase
    % accurracy but this is not VOC-way of doing it.)
    %
    % see VOCevaldet to see how difficult is handled. the same here.
    
    % c_maxovr is empty when there is no ground truth object
    if ~isempty(c_maxovr) && c_maxovr(obji) >= minoverlap % VOC uses >= 
      
      % get associated bounding box
      c_curgti = c_gtind(obji);
      
      if c_gtisdiffdet(c_curgti)
        % Associated groundtruth object is difficult
        %
        % Note that both VOC 2007 and VOC 2010 allows multiple detections of the same difficult object.
        % Any object whose max overlap (above >= .5) is a difficult object will be dropped from the
        % evaluation. none of them will be included in the evaluations.
        %
        % In this case, evalcase is neither tp nor fp since it may or may not be the unique 
        % detection on this groundtruth object.
        
        % set detection as 'difficult' too so that it wont be used in evaluation
        % either in a positive or negative way.
        detinfo.isdiffdet(obji) = true;
        
        % better we assign it now as been detected for efficiency
        c_hasdetected(c_curgti) = true;
        
      else
        % a normal groundtruth object
        
        if ~(c_hasdetected(c_curgti))
          c_hasdetected(c_curgti) = true;
          detinfo.gt(obji) = true;
          detinfo.evalcase(obji) = 1;
        else %  has been detected before, 
             % detinfo.gt(obji) = false; --> no need it is already false
             detinfo.evalcase(obji) = -2;
        end
        
        detinfo.gti(obji) = c_curgti;
      end
      
    else
      % false positive
      % detinfo.gt(obji) = false; -- no need
      detinfo.evalcase(obji) = -1;
      detinfo.gti(obji) = -1;
    end
    
  end

  % add undetected ground truth objects (false negatives at all thresholds) with -inf score as detections.
  % we will skip those that are labeled as 'difficult' but not here. it is a consideration after doing all these operations.
  c_nundetobj = c_ngt - sum(c_hasdetected); % which is sum(~c_hasdetected)
  if c_nundetobj > 0
    % false negative. gt=true since these are relevant objects with -inf score.
    detinfo.gt         = [detinfo.gt        true(1,c_nundetobj)];
    detinfo.scores     = [detinfo.scores    -inf(1,c_nundetobj)];
    detinfo.evalcase   = [detinfo.evalcase  repmat(-3,1,c_nundetobj)];
    detinfo.isdiffdet  = [detinfo.isdiffdet c_gtisdiffdet(~c_hasdetected)];
  end

  % evalcase:
  %  +1: tp
  %  -1: fp at background
  %  -2: fp due to 2+ detection of the same object,
  %  -3: missed object (fn)
  
end
