function classifier = cn_pqliblinear_prep(classifier,pqcenters,opt)
% classifier = cn_pqliblinear_prep(classifier,pqcenters,opt)
%
% Precompute alls_pq which allows fast classification, and required by cn_pqliblinear_test().
%
% This function is automatically called by cn_pqliblinear_train().
% However, alls_pq may be too large to store on disk. Therefore, one may prefer to 
% save classifier without alls_pq field and re-generate it later.
%
% INPUT
% classifier
%   .z2c        (1 Z)
%   .allw_pq    (sum(ds_z.*nsq_z) ncls)     Uncompressed weight vector.
%   .allw_e     (de ncls)                   Used only for sanity checking, de=0 is OK.
%   .[ds_c]     (1 C)                       Only if opt=matlab.
%   .[nsq_c]    (1 C)                       Only if opt=matlab.
% pqcenters     {(ds_c ks nsq_c)}
% opt           'mex' (default) | 'matlab'
%
% OUTPUT
% classifier
%   +.alls_pq   (ncls ks nsq_total)         Look up table to use in scoring.
%
% SEE cn_pqliblinear_train
% 
% R.G.Cinbis, Nov 2012

% 2013-04-20, 15:03: Now 'mex' implementation is the detault one. 

% sanity checks on the classifier struct
ncls = size(classifier.allw_e,2);
assert(ncls==size(classifier.allw_pq,2));

if nargin < 3 || strcmp(opt,'mex')

    % todo: try converting pqcenters into double.
    allwt = classifier.allw_pq.';
    if isa(allwt,'double') 
        for i = 1:length(pqcenters)
            pqcenters{i} = double( pqcenters{i} ); % does not copy if this already in double.
        end
    end
    classifier.alls_pq = cn_pqliblinear_prepmex(allwt,uint64(classifier.z2c),pqcenters);

elseif strcmp(opt,'matlab')

    ds_c      = classifier.ds_c;
    nsq_c     = classifier.nsq_c;
    z2c       = classifier.z2c;
    Z         = length(z2c);
    ds_z      = ds_c(z2c);
    nsq_z     = nsq_c(z2c);
    nsq_total = sum(nsq_z);
    allw_pq   = classifier.allw_pq;
    ks        = 256;

    classifier.alls_pq = zeros(ncls,ks,nsq_total);
    i0 = 0;
    j0 = 0;
    for z = 1:Z
        i2 = i0 + ds_z(z) * nsq_z(z);
        j2 = j0 + nsq_z(z);

        tmpw = reshape( allw_pq((i0+1):i2,:), [ds_z(z) nsq_z(z) ncls] ); % (ds_z nsq_z ncls)

        % permute -> (ds_z ncls nsq_z)
        tmpw = permute(tmpw,[1 3 2]);

        % Then:  (ds_z ncls nsq_z)' (ds_z ks nsq_z) -> (ncls ks nsq_z)
        classifier.alls_pq(:,:,(j0+1):j2) = mtimesx( tmpw, 't', double(pqcenters{z2c(z)}), 'speed');

        i0 = i2;
        j0 = j2;
    end

else
    error('failure');
end


