#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstring>
#include <stdint.h> // to get uint8_t
#include <limits>
#include <cstdarg>
#include "cn_matlabarrays.hpp"
#include "cn_pqliblinear_shared.hpp"
#include "cn_math.hpp"
using namespace std;

/**
 * Author: R. Gokberk Cinbis.
 */


#define VERBOSE 0

/* 
 * Implements L1 and L2 loss SVM solver based on PQ compressed descriptors.
 * Based on weighted liblinear, therefore, accepts weights.
 * Uses a dual solver, however, no <xi,xj> inner products are actually needed. It keeps solution in the primal
 * version.
 *
 * In addition to pq-compressed vectors, we allow adding a number of not-compressed dimensions, which 
 * can be used for bias, etc. purposes. For standard bias term, it should be set to all 1.
 * 
 * -- INPUT --
 * centers          {(ds_c ks nsq_c)}       single      (1 C) PQ centers.
 * z2c              (1 Z)                   mwSize      Map pq blocks within codes array to pq-centers.
 * codes            (nsq_total_code N)      uint8       PQ codes.
 * extras           (de N)                  double      Extra, uncompressed dimension defined per-example.
 * y                (1 N)                   double      Labels {-1,+1}
 * samplew          (1 N)                   double      Sample weights.
 * type             1: L2R_L2LOSS_SVC_DUAL
 *                  3: L2R_L1LOSS_SVC_DUAL
 * eps              scalar                  double      Convergence threshold.
 * quiet            scalar                  bool
 *   
 * -- OUTPUT --
 * wpq              (sum(ds_i*nsq_i) 1)     double
 * wextras          (de 1)
 *
 */

#include "mex.h"

/* ----------------- liblinear definitions ----------------- */

typedef signed char schar;

#if 0
template <class T> static inline void swap(T& x, T& y) { T t=x; x=y; y=t; }
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#define INF HUGE_VAL

// ------  print functions  -------  

void print_null(const char *s) {}
void print_string_matlab(const char *s) {mexPrintf(s);}
static void (*liblinear_print_string) (const char *) = &print_string_matlab;

static void info(const char *fmt,...)
{
    char buf[BUFSIZ];
    va_list ap;
    va_start(ap,fmt);
    vsprintf(buf,fmt,ap);
    va_end(ap);
    (*liblinear_print_string)(buf);
}


/* ----------------- my definitions ----------------- */

#define ks          256                 // we fix codes to uint8

//     0       1                    2               3                    4         5               6       7 
enum { L2R_LR, L2R_L2LOSS_SVC_DUAL, L2R_L2LOSS_SVC, L2R_L1LOSS_SVC_DUAL, MCSVM_CS, L1R_L2LOSS_SVC, L1R_LR, L2R_LR_DUAL }; /* solver_type */

struct problem
{
    size_t l, n;                // (was int) l: number of samples, n: original dimensionality of the descriptors
    int *y;                     // labels
    const double *W;            // instance weights, vector (1 numdata) 
    // removed: x,bias

    // things that I've added:
    vector <const float *> centers;  // PQ centers {(ds_c ks nsq_c)} 
    const uint8_t *codes;       // PQ codes, col major ie. (nsq_total_code numdata)
    const double *wscr;         // Scaling for compressed descriptors/scores.
    const double *extras;       // Uncompressed dimensions matrix of size (de N).
    vector <mwSize> nsq_c;      // number of subquantizers for pq centers.
    vector <mwSize> ds_c;       // dimensionality of a single subquantizer for pq centers.
    vector <mwIndex> z2c;       // mapping from pq code blocks to pq centers
    mwSize nsq_total_code;         // sum(nsq_c(z2c)) Total code length.
    mwSize C;                   // length(centers)
    mwSize Z;                   // number blocks, ie length(z2c)
    mwSize de;                  // number of extra dimensions
    mwSize dtotalpq;            // total dimensionality encoded in pq.

};

// GOKBERK: Precompute sum(centers.^2,1). Result in double. Not time critical.
// result[c] is (ks nsq_c[c]) dimensional.
vector<double *> *computeCentersNormSq(const problem * prob) {
    vector <double *> *centersNormSq = new vector<double *>;
    centersNormSq->reserve( prob->C );
    for (mwIndex c = 0; c < prob->C; ++c) {

        const float *tmp_center = prob->centers[c]; // (ds ks nsq)
        const mwSize nsq = prob->nsq_c[c]; 
        const mwSize ds = prob->ds_c[c]; 

        double *tmp_out = (double *) cn::mex::matCalloc(ks*nsq,sizeof(double)); // compute & keep in double. init to 0.
        centersNormSq->push_back(tmp_out);

        for (mwIndex sqi = 0; sqi < nsq; ++sqi) { // {c}(:,sqi)
            for (mwIndex ki = 0; ki < ks; ++ki) {
                for (mwIndex d = 0; d < ds; ++d) {
                    double x = *tmp_center;
                    *tmp_out += x * x;
                    ++tmp_center;
                }
                ++tmp_out;
            }
        }

    }
    return centersNormSq;
}


template <class WscrCase>
inline double getWscr(const problem *prob,const mwIndex &z,const mwIndex &i) {
    mexErrMsgTxt("BUG");
    return prob->wscr; // just to avoid compile errors.
}
template <>
inline double getWscr<WSCR_Z_N>(const problem *prob,const mwIndex &z,const mwIndex &i) {
    return prob->wscr[prob->Z * i + z];
}
template <>
inline double getWscr<WSCR_1_N>(const problem *prob,const mwIndex &z,const mwIndex &i) {
    return prob->wscr[i];
}
template <>
inline double getWscr<WSCR_Z_1>(const problem *prob,const mwIndex &z,const mwIndex &i) {
    return prob->wscr[z];
}
template <>
inline double getWscr<WSCR_1_1>(const problem *prob,const mwIndex &z,const mwIndex &i) {
    return *(prob->wscr);
}


// GOKBERK: Get Squared L2 Norm of i-th vector. Not time critical.
template <class WscrCase>
double getSquaredNorm(const problem *prob, const vector<double *> &centersNormSq, const mwIndex i) {
    double cur = 0;
    const uint8_t * tmp_code = prob->codes + prob->nsq_total_code * i;
    for (mwIndex z = 0; z < prob->Z; ++z) {

        // current subquantizer
        mwIndex c = prob->z2c[z];
        const double * tmp_norm = centersNormSq[c]; // (ks nsq)
        const mwSize nsq = prob->nsq_c[c]; 

        double tmp_cur = 0; // norm for current z
        for (mwIndex sqi = 0; sqi < nsq; ++sqi) {
            tmp_cur += tmp_norm[ *tmp_code ];
            ++tmp_code; // next subquantizer
            tmp_norm += ks; // next subquantizer
        }

        // scale by wscr(z,i)^2
        const double &xwscr = getWscr<WscrCase>(prob,z,i);
        cur += tmp_cur * xwscr * xwscr;

    }

    // squared norm of the uncompressed vector
    const double * tmp_extra = prob->extras + prob->de * i;
    for (mwIndex d = 0; d < prob->de; ++d) {
        cur += *tmp_extra * *tmp_extra;
        ++tmp_extra;
    }
    return cur;
}

// GOKBERK: w' * x_i based pq-code of x_i and also extras.
// Each vector is assumed to have col-major ordering of size (ds nsq Z)
template <class WscrCase>
double w_dot_x(const problem *prob, const double *w, const mwIndex i) {

    double G = 0;
    const uint8_t *tmp_code = prob->codes + prob->nsq_total_code * i;
    const double *tmp_w = w;
    for (mwIndex z = 0; z < prob->Z; ++z) {
        double tmpG = 0;

        // current subquantizer
        mwIndex c = prob->z2c[z];
        const float * tmp_centerslice = prob->centers[c]; // (ds ks nsq)
        const mwSize nsq = prob->nsq_c[c]; 
        const mwSize ds = prob->ds_c[c]; 
        const mwSize dsks = ds * ks;

        for (mwIndex sqi = 0; sqi < nsq; ++sqi) {
            const float * tmp_centercol = tmp_centerslice + ds * *tmp_code;
            cn::ddot(ds,tmp_w,tmp_centercol,tmpG);
            tmp_w += ds; // next block's weight vector
            ++tmp_code; // next subquantizer code
            tmp_centerslice += dsks; // centers(:,:,sqi) for the next sqi
        }

        // scale by wscr(z,i)
        G += tmpG * getWscr<WscrCase>(prob,z,i);
    }

    // pq compressed part is completed. now extra dims.
    const double * tmp_extra = prob->extras + prob->de * i;
    cn::ddot(prob->de,tmp_w,tmp_extra,G);

    return G;
}

// GOKBERK: w = w + x_i * scale
// Each vector is assumed to have col-major ordering of size (sum(ds_i*nsq_i) Z)
template <class WscrCase>
void increment_w_by_x(const problem *prob,double *w,const mwIndex i,const double scale) {

    const uint8_t *tmp_code = prob->codes + prob->nsq_total_code * i;
    double *tmp_w = w;
    for (mwIndex z = 0; z < prob->Z; ++z) {

        // current subquantizer
        mwIndex c = prob->z2c[z];
        const float * tmp_centerslice = prob->centers[c]; // (ds ks nsq)
        const mwSize nsq = prob->nsq_c[c]; 
        const mwSize ds = prob->ds_c[c]; 
        const mwSize dsks = ds * ks;

        const double tmp_scale = scale * getWscr<WscrCase>(prob,z,i); // scale by wscr(z,i)
        for (mwIndex sqi = 0; sqi < nsq; ++sqi) {
            const float * tmp_centercol = tmp_centerslice + ds * *tmp_code;
            for (mwIndex d = 0; d < ds; ++d) {
                *(tmp_w++) += tmp_scale * *(tmp_centercol++); // next dim
            }
            ++tmp_code; // next subquantizer code
            tmp_centerslice += dsks; // next subquantizer centers
        }
    }

    // pq compressed part is completed. now extra dims.
    const double * tmp_extra = prob->extras + prob->de * i;
    for (mwIndex d = 0; d < prob->de; ++d) {
        *(tmp_w++) += scale * *(tmp_extra++); // next dim
    }

}

// w            Preallocated (1 numdim)
// solver_type  L2R_L2LOSS_SVC_DUAL or L2R_L1LOSS_SVC_DUAL
template <class WscrCase>
static void solve_l2r_l1l2_svc(
        const problem *prob, double *w, double eps, 
        int solver_type)
{
    int l = prob->l;
    int w_size = prob->n;
    int i, s, iter = 0;
    double C, d, G;
    double *QD = new double[l];
    int max_iter = 1000;
    int *index = new int[l];
    double *alpha = new double[l];
    schar *y = new schar[l];
    int active_size = l;

    // PG: projected gradient, for shrinking and stopping
    double PG;
    double PGmax_old = INF;
    double PGmin_old = -INF;
    double PGmax_new, PGmin_new;

    // default solver_type: L2R_L2LOSS_SVC_DUAL
    double *diag = new double[l];
    double *upper_bound = new double[l];
    double *C_ = new double[l];
    for(i=0; i<l; i++) 
    {
        if(prob->y[i]>0)
            C_[i] = prob->W[i]; // * Cp. Global cost parameter Cp = 1.
        else 
            C_[i] = prob->W[i]; // * Cn. Global cost parameter Cn = 1.
        diag[i] = 0.5/C_[i];
        upper_bound[i] = INF;
    }
    if(solver_type == L2R_L1LOSS_SVC_DUAL) // hinge loss
    {
        for(i=0; i<l; i++) 
        {
            diag[i] = 0;
            upper_bound[i] = C_[i];
        }
    } else if (solver_type != L2R_L2LOSS_SVC_DUAL) {
        mexErrMsgTxt("BUG, invalid solver_type!");
    }

    // GOKBERK: Precompute sum(centers.^2,1)
    vector<double *> *centersNormSq = computeCentersNormSq(prob);
   
    for(i=0; i<w_size; i++)
        w[i] = 0;
    for(i=0; i<l; i++)
    {
        alpha[i] = 0;
        if(prob->y[i] > 0)
        {
            y[i] = +1; 
        }
        else
        {
            y[i] = -1;
        }

        /*feature_node *xi = prob->x[i];
        while (xi->index != -1)
        {
            QD[i] += (xi->value)*(xi->value);
            xi++;
        }*/
        // GOKBERK
        QD[i] = diag[i] + getSquaredNorm<WscrCase>(prob, *centersNormSq, i);    

        index[i] = i;
    }

    for (mwIndex c = 0; c < centersNormSq->size(); ++c) 
        mxFree(centersNormSq->at(c));
    delete centersNormSq; // GOKBERK: not needed anymore

    while (iter < max_iter)
    {
        PGmax_new = -INF;
        PGmin_new = INF;

        for (i=0; i<active_size; i++)
        {
            int j = i+rand()%(active_size-i); // j: [i,N-i-1]
            swap(index[i], index[j]);
        }

        for (s=0; s<active_size; s++)
        {
            i = index[s];
            G = 0;
            schar yi = y[i];


            // GOKBERK: DO DOT PRODUCT BETWEEN w and xi. We can't precompute dot products.
            /*
            feature_node *xi = prob->x[i];
            while(xi->index!= -1)
            {
                G += w[xi->index-1]*(xi->value);
                xi++;
            }
            */
            G += w_dot_x<WscrCase>( prob, w, i );

            G = G*yi-1;

            C = upper_bound[i];
            G += alpha[i]*diag[i];

            PG = 0;
            if (alpha[i] == 0)
            {
                if (G > PGmax_old)
                {
                    active_size--;
                    swap(index[s], index[active_size]);
                    s--;
                    continue;
                }
                else if (G < 0)
                    PG = G;
            }
            else if (alpha[i] == C)
            {
                if (G < PGmin_old)
                {
                    active_size--;
                    swap(index[s], index[active_size]);
                    s--;
                    continue;
                }
                else if (G > 0)
                    PG = G;
            }
            else
                PG = G;

            PGmax_new = max(PGmax_new, PG);
            PGmin_new = min(PGmin_new, PG);

            if(fabs(PG) > 1.0e-12)
            {
                double alpha_old = alpha[i];
                alpha[i] = min(max(alpha[i] - G/QD[i], 0.0), C);
                d = (alpha[i] - alpha_old)*yi;

                // GOKBERK Add d*xi on to weight vector w
                increment_w_by_x<WscrCase>(prob,w,i,d);
                /*
                xi = prob->x[i];
                while (xi->index != -1)
                {
                    w[xi->index-1] += d*xi->value;
                    xi++;
                }
                */
            }
        }

        iter++;
        if(iter % 10 == 0)
            info(".");

        if(PGmax_new - PGmin_new <= eps)
        {
            if(active_size == l)
                break;
            else
            {
                active_size = l;
                info("*");
                PGmax_old = INF;
                PGmin_old = -INF;
                continue;
            }
        }
        PGmax_old = PGmax_new;
        PGmin_old = PGmin_new;
        if (PGmax_old <= 0)
            PGmax_old = INF;
        if (PGmin_old >= 0)
            PGmin_old = -INF;
    }

    info("\noptimization finished, #iter = %d\n",iter);
    if (iter >= max_iter)
        info("\nWARNING: reaching max number of iterations\nUsing -s 2 may be faster (also see FAQ)\n\n");

    // calculate objective value

    double v = 0;
    int nSV = 0;
    for(i=0; i<w_size; i++)
        v += w[i]*w[i];
    for(i=0; i<l; i++)
    {
        v += alpha[i]*(alpha[i]*diag[i] - 2);
        if(alpha[i] > 0)
            ++nSV;
    }
    info("Objective value = %lf\n",v/2);
    info("nSV = %d\n",nSV);

    delete [] upper_bound;
    delete [] diag;
    delete [] C_;
    delete [] QD;
    delete [] alpha;
    delete [] y;
    delete [] index;
}

enum { a_centers = 0, a_z2c, a_codes, a_wscr, a_extras, a_y, a_samplew, a_type, a_eps, a_quiet };

void mexFunction(int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[])
{
    if((nlhs != 2) || (nrhs != 10))    
        mexErrMsgTxt("Invalid #input/#output params.");

    // NO! problem * prob = (problem *) mxMalloc(sizeof(problem)); 
    // Since prob contains objects, this fails.
    // This is fine: problem * prob = new problem; 
    problem prob_stat;
    problem *prob = &prob_stat;

    // all inputs arrays are dense
    for (mwIndex i = 0 ; i < (mwSize) nrhs; ++i ) {
        cn::mex::assert2(!mxIsSparse(prhs[i]),"no argument can be sparse");
    }

    // centers
    if (!mxIsCell(prhs[a_centers]))
        mexErrMsgTxt("pq_centers should be a cell array");
    prob->C = mxGetNumberOfElements(prhs[a_centers]);   

    // reserve vectors
    prob->centers.reserve(prob->C);
    prob->ds_c.reserve(prob->C);
    prob->nsq_c.reserve(prob->C);

    // get centers
    for (mwIndex c = 0; c < prob->C; ++c) {
        const mxArray *arr = mxGetCell(prhs[a_centers],c); // (ds_c ks_c nsq_c)
        if (arr==NULL)
            mexErrMsgTxt("pqcenters cell array contains NULL array pointers!");
        cn::mex::assert2(!mxIsSparse(arr),"pqcenters members cannot be sparse");

        prob->ds_c.push_back( cn::mex::getdim(arr,1) );
        cn::mex::assert2(cn::mex::getdim(arr,2)==ks,"size(centers,2)!=ks");
        prob->nsq_c.push_back( cn::mex::getdim(arr,3) ); 

        prob->centers.push_back( cn::mex::getDenseData<float>(arr) );
    }

    // mapping
    prob->Z = mxGetNumberOfElements( prhs[a_z2c] );
    const double *z2c_d = cn::mex::getDenseData<double>( prhs[a_z2c] );
    prob->z2c.reserve(prob->Z);
    prob->nsq_total_code = 0;
    for (mwIndex z = 0; z < prob->Z; ++z) {
        mwIndex c = (mwIndex) z2c_d[z]-1;
        prob->z2c.push_back(c);
        if (c > prob->C)
            mexErrMsgTxt("invalid entries in z2c");
        prob->nsq_total_code += prob->nsq_c.at(c);
#if VERBOSE
        printf("nsq_c[%d]=%d\n",c,prob->nsq_c[c]);
#endif
    }

    // codes
#if VERBOSE
    printf("nsq_total_code=%d size(codes,1)=%d\n",prob->nsq_total_code,cn::mex::getdim(prhs[a_codes],1));
#endif
    cn::mex::assert2(cn::mex::getdim(prhs[a_codes],1)==prob->nsq_total_code,"Incompatible size(codes,1)");
    prob->l = cn::mex::getdim(prhs[a_codes],2); 
    prob->codes = cn::mex::getDenseData<uint8_t>(prhs[a_codes]);

    // 2012-09-24: make randomization stable, never change this initialization! without this, results were not stable.
    srand( (unsigned int) prob->l );

    // wscr (Z N) | (1 N) | (Z 1) | (1 1)
    EnumWscrCase wscrShape = getWscrCase(prhs[a_wscr],prob->Z,prob->l);
    prob->wscr = cn::mex::getDenseData<double>(prhs[a_wscr]);

    // extra dimensions  (de N)
    prob->de = cn::mex::getdim(prhs[a_extras],1);
    cn::mex::assert2(mxGetN(prhs[a_extras])==prob->l,"size(extras,2)!=N");
    prob->extras = cn::mex::getDenseData<double>(prhs[a_extras]);

    // total dimensionality
    prob->dtotalpq = 0;
    for (mwIndex z = 0; z < prob->Z; ++z) {
        mwIndex c = prob->z2c[z];
        prob->dtotalpq += prob->ds_c[c] * prob->nsq_c[c];
    }
    prob->n = prob->dtotalpq + prob->de;

    // labels
    cn::mex::assert2(mxGetNumberOfElements(prhs[a_y])==prob->l);
    const double *y_double = cn::mex::getDenseData<double>(prhs[a_y]);
    prob->y = (int *) mxMalloc(sizeof(int)*prob->l);
    for (mwIndex i = 0; i < prob->l; ++i) {
        if (y_double[i]>0)
            prob->y[i] = 1;
        else
            prob->y[i] = -1;
    }

    // samplew
    cn::mex::assert2(mxGetNumberOfElements(prhs[a_samplew])==prob->l);
    prob->W = cn::mex::getDenseData<double>(prhs[a_samplew]);

    // type
    int solver_type = (int) cn::mex::getscalar(prhs[a_type]);

    // other options
    double eps = cn::mex::getscalar(prhs[a_eps]);
    bool quiet = cn::mex::getscalar(prhs[a_quiet]) != 0;

    // print function
    if (quiet)
        liblinear_print_string = &print_null;
    else
        liblinear_print_string = print_string_matlab; // default printing to matlab display

    // alloc weight vector, init to 0.
    double *w = (double *) cn::mex::matCalloc(prob->n,sizeof(double));

    // call solver
    //printf("type=%d\n",solver_type);
    if ((solver_type==1) || (solver_type==3)) {
        switch(wscrShape) {
            case ENUM_WSCR_Z_N:
                solve_l2r_l1l2_svc<WSCR_Z_N>(prob, w, eps, solver_type);
                break;
            case ENUM_WSCR_1_N:
                solve_l2r_l1l2_svc<WSCR_1_N>(prob, w, eps, solver_type);
                break;
            case ENUM_WSCR_Z_1:
                solve_l2r_l1l2_svc<WSCR_Z_1>(prob, w, eps, solver_type);
                break;
            case ENUM_WSCR_1_1:
                solve_l2r_l1l2_svc<WSCR_1_1>(prob, w, eps, solver_type);
                break;
            default:
                mexErrMsgTxt("Error");
        }
    } else {
        mexErrMsgTxt("Unsupported solver type");
    }           

    // set outputs
    mwSize out0sz [2];
    out0sz[0] = prob->dtotalpq;
    out0sz[1] = 1;

    plhs[0] = mxCreateNumericArray(2,out0sz,mxDOUBLE_CLASS,mxREAL); 
    plhs[1] = mxCreateDoubleMatrix(prob->de,1,mxREAL); 

    // copy weights vector to the output
    double *tmp_out1 = cn::mex::getDenseData<double>(plhs[0]); // weight vector for pq compressed dimensions
    double *tmp_out2 = cn::mex::getDenseData<double>(plhs[1]); // weight vector for extras
    double *tmp_w = w;
    for (mwIndex d = 0; d < prob->dtotalpq; ++d) {
        *(tmp_out1++) = *(tmp_w++);
    }
    for (mwIndex d = prob->dtotalpq; d < prob->n; ++d) {
        *(tmp_out2++) = *(tmp_w++);
    }

    // clear the memory
    mxFree(prob->y);
    mxFree(w);

}

