
/**
 * Author: R. Gokberk Cinbis.
 */



#include "mex.h"
#include <omp.h>
#include <cmath>
#include <math.h>
#include <vector>
#include <algorithm>
#include <cstring>
#include "cn_threads.hpp"
#include "cn_matlabarrays.hpp"
#include "cn_rand.hpp"
#include "cn_math.hpp"
#include "cn_sort.hpp"
#include <vector>
#include <iostream>
#include <string>
#include <sstream>
#include <limits>
#include <algorithm>
#include "cn_siginthandler.hpp"
#include "cn_memory.hpp"
#include "cn_math.hpp"

using std::vector;
using std::string;
using std::runtime_error;
using std::numeric_limits;
using cn::RandGenTwister;
using cn::transposematrix;

#define DEBUG 0

// Data type of the new cluster centers to accumulate descriptors.
// We may add a lot of numbers, so, we fix this to double.
typedef double Tnewcent;   

#if CN_WINDOWS
#define printf mexPrintf
#endif

/* Using Jakob's Distance Calculation Trick:
 *
 * Find the closest center c to each point x:
 *      argmin_c( || x - c || )
 *      = argmin_c( sum_d xd^2 + cd^2 - 2*xd*cd || )
 *      = argmin_c( sum_d (.5*cd^2) - xd*cd || )
 *      = argmax_c( sum_d xd*cd - (.5*cd^2) || )
 *
 * In the code, (.5*cd^2) is stored as "Cnorms"
 */

// simple class handle malloc() safely
template <class T>
class MallocHandler {
    public:
        MallocHandler(mwSize nitem): data(NULL) {
            data = (T *) malloc( sizeof(T) * nitem);
            if (data==NULL)
                throw std::runtime_error("MallocHandler::Allocation failed");
        }

        ~MallocHandler() {
            if (data!=NULL)
                free(data);
        }
        T *ptr() { return this->data; }
    private:
        T *data;

};

// at each call:
//      return current pointer (before increment) and increment tmp by sizeof(T)*numel
template <class T>
T *gettemp( mwSize numel, uint8_t **tmp ) {
    T *res = (T *) *tmp;
    *tmp += sizeof(T) * numel;
    return (T *) res;
}

// print out a given matrix
template <class T>
void printMatrix(const T *X,const mwSize M,const mwSize N) {
    for (mwSize i = 0; i<M; i++) {
        for (mwSize j =0; j<N; j++) {
            printf("%.4f ",(double) *(X + j * M + i));
        }
        printf("\n");
    }
    printf("\n");
}
template <>
void printMatrix <uint32_t> (const uint32_t *X,const mwSize M,const mwSize N) {
    for (mwSize i = 0; i<M; i++) {
        for (mwSize j =0; j<N; j++) {
            printf("%d ",(int) *(X + j * M + i));
        }
        printf("\n");
    }
    printf("\n");
}

// randomly initialize centers using M = XT(:,cn_randmask(N,K)); 
template <class Tcent, class Tx>
void initcenters(Tcent *C,const Tx *XT,const mwSize ds,const uint32_t K,const mwSize N,RandGenTwister &rgen) {
    MallocHandler<mwSize> h_ind(K);
    mwSize *ind = h_ind.ptr();
    rgen.randpermfirstk(N,(mwSize) K,ind);
    cn::sort(K,ind,"ascend");
    Tcent *C2 = C;
    for (uint32_t k = 0; k < K; ++k) {
        const Tx *XT2 = XT + ds * ind[k];
        for ( mwSize d = 0; d < ds; ++d )  {
            *(C2++) = *(XT2++);
        }
    }
}

// Input:   (ds K) C 
// Output:  (1  K) Cnorms = .5 * sum( C.^2 , 1 )
template <class Tcent>
void compCnorms(const Tcent *C,const mwSize ds,const uint32_t K,Tcent *Cnorms) {
    // printf("C:\n"); printMatrix(C,ds,K); Tcent * Cnorms_org = Cnorms;
    for (mwSize k = 0; k < K; ++k) {
        Tcent cur = 0;
        for (mwSize d = 0; d < ds; ++d) {
            Tcent val = *(C++);
            cur += val * val;
        }
        *(Cnorms++) = (cur / 2);
    }
    // printf("Cnorms:\n"); printMatrix(Cnorms_org,1,K);
}

// newcent += x
template <class Tnewcent,class Tx> 
void incrementcenter(Tnewcent *C,const Tx *X,const mwSize ds) {
    for (mwSize d = 0; d < ds; ++d) {
        *(C++) += *(X++);
    }
}

// tmpC(:,k) /= countC(k)
// C = tmpC
template <class Tnewcent,class Tcent> 
void normalizeAndCopyCenters( Tcent * C, Tnewcent * tmpC, mwSize * countC, const mwSize ds, const uint32_t K ) {
   
    // tmpC(:,k) /= countC(k)
    Tnewcent *tmpC2 = tmpC;
    for (mwSize k = 0 ; k < K ; ++k ) {
        Tnewcent curc = *(countC++); // we can assume that curc >= 1 thanks to replaceEmptyClusters().
        for (mwSize d = 0 ; d < ds; ++d ) {
            *(tmpC2++) /= curc;
        }
    }
   
    // C = tmpC
    if ( sizeof(Tcent) == sizeof(Tnewcent) ) {
        memcpy( C, tmpC, sizeof(Tcent) * ds * K );
    } else { 
        mwSize nel = ds * K;
        for ( mwSize i = 0 ; i < nel ; ++i ) {
            *(C++) = (Tcent) *(tmpC++);
        }
    }
}


// If isXtranspose=true  => X: (ds N)
// If isXtranspose=false => X: (N ds)
//
// returns true, if any id was changed.
//
// Just like jakob_kmeans, there is no need to change countC on non-empty clusters. It is fine for a point to be
// a part of more than one cluster center, for this boundary condition.
template <class Tx>
bool replaceEmptyClusters(Tnewcent *tmpC,const Tx *X,const bool isXtranspose,mwSize *countC,
        const uint32_t K,const mwSize N,const mwSize ds,RandGenTwister &rgen,uint32_t *ids) {
    // replace empty clusters with random points
    bool res = false;
    for (uint32_t k = 0; k < K; ++k ) {
        // printf("countC[%d]=%ld\n",k,countC[k]);
        if ( countC[k] == 0 ) {
            mwSize pti = rgen.randi(N); // [0,N)

            // this is not 100% compatible with jakob_kmeans but somewhat close to it.
            // it is costly to try to get the same behaviour here.
            ids[pti] = k;
            res = true;

#if DEBUG
            printf("Replacing empty cluster %d with point %d \n", (int) k, (int) pti); 
#endif

            Tnewcent *tmpC2 = tmpC + ds * k;
            if (isXtranspose) {
                const Tx *X2 = X + ds * pti;
                for ( mwSize d = 0; d < ds; ++d )  {
                    *(tmpC2++) = *(X2++);
                }
            } else {
                const Tx *X2 = X + pti;
                for ( mwSize d = 0; d < ds; ++d )  {
                    *(tmpC2++) = *X2;
                    X2 += N;
                }
            }
            countC[k] = 1;
        }
    }
    return res;
}

// sort columns of C of size (ds N) such that first one has the highest countC.
// tmpC (of at least the same size with C) is used only as a temporary area.
// on return, "countC" is no more valid and tmpCorg contains garbage values.
template <class Tcent>
void sortcentersbyfreq( Tcent * C, Tnewcent * tmpCorg, mwSize * countC, const mwSize ds, const uint32_t K ) {
    MallocHandler <uint32_t> h_order( K );
    uint32_t *order = h_order.ptr();
    cn::stable_getorder( K, countC, order, "descend" );
    if ( sizeof(Tcent)>sizeof(Tnewcent) ) {
        throw runtime_error("bug::unsupported case");
    }

    Tcent *tmpC = (Tcent *) tmpCorg;
    for (mwIndex to = 0; to < K; ++to ) {
        const uint32_t &from = order[to];
        memcpy(tmpC + ds * to,C + ds*from,sizeof(Tcent)*ds);
    }
    memcpy(C,tmpC,sizeof(Tcent)*ds*K);
}

// Takes transpose beforehand to speed-up the process.  An alternative would be using in-place tranpose, but,
// because of the difficult of that operation (which is not implemented even by matlab for non-square matrices), 
// I haven't tried that.
//
// X        (N ds) Sub array to process
// tmp      Temporary area of length TMP_FIRSTTRANSPOSE
// C        (ds K) Output centers. If already_initialized==true, C also provides input, initial centers.
template <class Tx,class Tcent>
bool kmeans_with_trans(const Tx *X,uint8_t *tmp,Tcent *C,const mwSize N,const mwSize ds,const uint32_t K,
        RandGenTwister &rgen,const mwSize maxiter,const bool already_initialized) {

    // temp arrays
#define TMP_WITH_TRANS  sizeof(Tx)*ds*N+sizeof(uint32_t)*N+sizeof(Tnewcent)*ds*K+sizeof(mwSize)*K+sizeof(Tcent)*K
    Tx *XT          = gettemp<Tx>(ds*N,&tmp);
    uint32_t *ids   = gettemp<uint32_t>(N,&tmp);    // ids, needed to track any changes
    Tnewcent *tmpC  = gettemp<Tnewcent>(ds*K,&tmp); // temporary centers array  
    mwSize *countC  = gettemp<mwSize>(K,&tmp);      // number of points assigned to each cluster
    Tcent *Cnorms   = gettemp<Tcent>(K,&tmp);     

    transposematrix(X,N,ds,XT);

    if (!already_initialized)
        initcenters(C,XT,ds,K,N,rgen);

    // Initially, assume all points were assigned to the 0-th cluster.
    // If this is not the case, simply will cause converged = false.
    // If this is really the case (bad initialization) and K > 1, this will require re-initialization, which is
    // also fine as it will set converged = false.
    memset( ids, 0, sizeof(uint32_t)*N );

    // printMatrix(C,ds,K);
    bool converged = false;
    mwSize iteri = 1;
    while((!converged) && (iteri < maxiter)) {
        converged = true;
        iteri++; 

        // compute .5*C.^2 
        compCnorms(C,ds,K,Cnorms);

        memset(tmpC,0,sizeof(Tnewcent)*ds*K);
        memset(countC,0,sizeof(mwSize)*K);

        // find closest center to each point while updating cluster centers
        const Tx *xi = XT;
        uint32_t bestk = 0; // define here just to initialize only once
        for (mwSize i = 0; i < N; ++i ) { 
            Tcent maxsofar = -numeric_limits<Tcent>::infinity();

            // find the closest center
            const Tcent *C2 = C;
            for (uint32_t k = 0; k < K; ++k) {
                Tcent tmp = 0; // should be initialized 
                cn::ddot(ds,xi,C2,tmp);
                tmp -= Cnorms[k];
                C2 += ds;
                if (tmp > maxsofar) {
                    maxsofar = tmp;
                    bestk = k;
                }
                // printf("ids[%d]=%d maxsofar=%f k=%d cur=%f\n",(int)i,(int)bestk,(double)maxsofar,(int)k,(double)tmp);
            }

            // check convergence
            if (bestk != ids[i]) {
                converged = false;
            }
            ids[i] = bestk;
            // printf("-> ids[%d]=%d max=%f\n",(int)i,(int)bestk,(double)maxsofar);

            // accumulate the centers, we'll normalize at the end
            incrementcenter( tmpC + ds * bestk, xi, ds );
            countC[bestk] += 1;
            
            xi += ds;

        } // over i

        // printMatrix <uint32_t> (ids,1,N);

        if (replaceEmptyClusters(tmpC,XT,true,countC,K,N,ds,rgen,ids))
            converged = false;

        // normalize new centers and copy them to C
        normalizeAndCopyCenters( C, tmpC, countC, ds, K );
        // printMatrix(C,ds,K);

    } // until convergence

#if DEBUG
    if ( converged ) {
        printf("Converged at iteri=%d\n",(int) iteri );
    } else {
        printf("Maximum number of iteratios reached at iteri=%d\n",(int) iteri );
    }
#endif

    // sort cernters based on their frequency
    sortcentersbyfreq( C, tmpC, countC, ds, K );

    return false;

}


// Here, we will NOT do a full X.' before k-means.
//#define TMP_NO_TRANS 


template 
<class Tx,class Tcent>
struct CppParams {
    const Tx * X;      // (N D)
    mwSize N;       // number of training examples
    mwSize D;
    mwSize nsq;
    mwSize ds;      // dimensionality of each chunk
    uint32_t K;     // uint32_t is good since it reduces memory footprint
    double randseed;
    double maxiter; // double since it supports Inf easily.
    Tcent * centers; // (ds K nsq)
    bool already_initialized;
};


// multithreaded wrapper
template 
<class Tx,class Tcent>
void kmeansparallel(CppParams <Tx,Tcent> &p) {

    // avoid overhead, if any
    const Tx * X = p.X;
    Tcent *C = p.centers;
    mwSize N = p.N;
    mwSize nsq = p.nsq;
    mwSize ds = p.ds;
    uint32_t K = p.K;
    double randseed = p.randseed;
    double maxiter = p.maxiter;
    bool already_initialized = p.already_initialized;

    // set #threads
    mwSize maxthread = cn::setmaxthreads();
    if (maxthread > nsq) {
        maxthread = nsq;
    }
    if (maxthread < 1)
        maxthread = 1;

#if DEBUG
        maxthread = 1;
#endif

    mwSize tmpSize = TMP_WITH_TRANS; // number of temp bytes needed per thread.

    // this will indicate the progress
    MallocHandler <bool> h_hastrained( nsq );
    bool *hastrained = h_hastrained.ptr();
    cn::siginthandler sigint;

    // temporary arrays
    while(true) { // loop until work is done. a work may interrupt in case of allocation problem.

        omp_set_num_threads(maxthread);
        memset(hastrained,0,sizeof(bool)*nsq); // progress indicator
        string errmsg("");
        bool iserr = false;

        // create threads
#pragma omp parallel
        {
            mwSize workeri = omp_get_thread_num();
            mwSize nthread = omp_get_num_threads();

            // allocate temporary std objects
            try {
                // if allocaation fails, re-try with a smaller number of threads

                MallocHandler <uint8_t> h_curtmp( tmpSize );
                uint8_t *curtmp = h_curtmp.ptr();

                mwSize end = cn::jobendi(nsq);
                for (mwSize sqi = cn::jobstarti(nsq); sqi < end; ++sqi) {

                    if (sigint.state() || iserr)
                        break; // ctrl+c

#if 0
                    // printf("clustering sqi=%d\n",(int) sqi );
                    if ((nthread >= 10) && (workeri < 3))
                        throw std::runtime_error("throwing exception for debug purposes within the thread, to simulate out-of-memory");
#endif 


                    // randseed = randseed + sqi + 1
                    // This is cleaner and easier (and also jakob_kmeans-compatible) for parallelization than
                    // randseed+sqi.
                    RandGenTwister rgen(randseed + sqi + 1);

                    Tcent *curC = C + ds * K * sqi;
                    const Tx *curX = X + N * ds * sqi;

                    kmeans_with_trans(curX,curtmp,curC,N,ds,K,rgen,maxiter,already_initialized);
                    hastrained[sqi] = true;

                    if (workeri==0) { // print out progress
#pragma omp critical
                        {                            
                            mwSize ndone = 0;
                            for (mwSize sqi = 0; sqi < nsq; ++sqi) {
                                if (hastrained[sqi])
                                    ++ndone;
                            }
                            printf("cn_kmeansparallel: %.2f%s (ie. %d/%d subquantizers) done (nthread=%d)\n",((double)ndone)/nsq*100,"%",(int) ndone,(int) nsq,(int) nthread);
                        }
                    } // if workeri==0

                } // loop over sqi

            } catch (std::exception &e) {
#pragma omp critical
                {
                    iserr = true;
                    std::stringstream ss;
                    ss << "Exception occured at thread " << workeri << ":" << e.what() << std::endl;
                    errmsg += ss.str();
                }
            } catch (...) {
#pragma omp critical
                {
                    iserr = true;
                    std::stringstream ss;
                    ss << "Unknown exception occurred at thread " << workeri << std::endl;
                    errmsg += ss.str();
                }
            }

        } // omp parallel region

        if (sigint.state()) {
            break;

        } else if ( iserr ) {
            // allocation had failed.
            printf("The following error(s) occurred within the threads: \n%s\n", errmsg.c_str() );

            // try again, if possible
            if (maxthread <= 1) {
                throw runtime_error("Probably, not enough memory!"); // fall-back to the other implementation?
            }
            mwSize maxthread2 = (mwSize) ceil( ( (double) maxthread ) / 2.0 );
            printf("Will try reducing nthread from %d to %d \n", (int) maxthread, (int) maxthread2 );
            maxthread = maxthread2;

        } else {
            for (mwSize sqi = 0; sqi < nsq; ++sqi) {
                if ( ! hastrained[sqi] )
                    throw runtime_error("bug: ~all(hastrained)");
            }
            break;
        }

    } // loop until succeed

    if (sigint.state()) {
        throw runtime_error("cn_kmeansparallel_mex interrupted by user");
    }

}


struct MatlabParams {
    public: 
    const mxArray * data;
    const mxArray * nsq;
    const mxArray * K;
    const mxArray * randseed;
    const mxArray * maxiter;
    const mxArray * init_centers;
};

template 
<class Tx,class Tcent>
void mexFunction2(int nlhs, mxArray *plhs[], MatlabParams &q ) {

    CppParams <Tx,Tcent> p;

    // check type sanity
    if (sizeof(Tcent) < sizeof(Tx))
        mexErrMsgTxt("Tcent is set up incorrectly!");
    if (sizeof(Tnewcent) < sizeof(Tcent))
        mexErrMsgTxt("Tnewcent is set up incorrectly!");

    // N,D,X
    if (mxGetNumberOfDimensions(q.data)>2)
        mexErrMsgTxt("data should be NxD");
    p.N = mxGetM(q.data);
    p.D = mxGetN(q.data);
    p.X = cn::mex::getDenseData<Tx>(q.data);

    // nsq, ds
    p.nsq = (mwSize) cn::mex::getscalar(q.nsq);
    if (p.nsq > p.D)
        mexErrMsgTxt("nsq is too high");
    else if ( (p.nsq < 1) && (p.D > 0) )
        mexErrMsgTxt("nsq cannot be 0 or less");
    else if ( (p.D % p.nsq) != 0 )
        mexErrMsgTxt("mod(D,nsq)==0 is required!");
    p.ds = p.D / p.nsq;

    // check 1 <= K <= N
    double K_double = cn::mex::getscalar(q.K);
    if (K_double > 4294967295.0)
        mexErrMsgTxt("K is too high. At most intmax('uint32') is supported");
    else if (K_double < 1)
        mexErrMsgTxt("K cant be less than 1");
    p.K = (mwSize) cn::mex::getscalar(q.K);
    if (p.K > p.N)
        mexErrMsgTxt("K cant be more than the number of data samples");

    // randseed,maxiter
    p.randseed = cn::mex::getscalar(q.randseed);
    p.maxiter = cn::mex::getscalar(q.maxiter);

    mxClassID Tcent_mex = cn::mex::type2MxClassID<Tcent>();

    // allocate output
    if (nlhs > 1)
        mexErrMsgTxt("too many output args");
    plhs[0] = cn::mex::createArrayWoInit(Tcent_mex,p.ds,p.K,p.nsq);
    p.centers = (Tcent *) cn::mex::getDenseData<Tcent>(plhs[0]);

    // initial centers
    if (q.init_centers != NULL) {
        const Tcent * init_centers = cn::mex::getDenseData<Tcent>(q.init_centers);
        if ( (mxGetNumberOfDimensions(q.init_centers)>3) || (cn::mex::getdim(q.init_centers,1)!=p.ds) 
                || (cn::mex::getdim(q.init_centers,2)!=p.K) || (cn::mex::getdim(q.init_centers,3)!=p.nsq) )
            mexErrMsgTxt("Initial centers, if provided, should be (ds K nsq)");
        memcpy(p.centers,init_centers,sizeof(Tcent)*p.D*p.K); // initialize
        p.already_initialized = true;
    } else {
        p.already_initialized = false;
    }

    kmeansparallel(p);

}




const mxArray *getfield(const mxArray *prhs[],const char *fieldname) {
    const mxArray *result = mxGetField(prhs[0],0,fieldname);
    if (result == NULL) {
        string e = "mxGetField: Missing field: ";
        e += fieldname;
        mexErrMsgTxt(e.c_str());
    }
    return result;
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

    if (nrhs!=1)
        mexErrMsgTxt("invalid number of input args");

    MatlabParams q;
    q.data = getfield(prhs,"data");
    q.nsq = getfield(prhs,"nsq");
    q.K = getfield(prhs,"K");
    q.randseed = getfield(prhs,"randseed");
    q.maxiter = getfield(prhs,"maxiter");
    q.init_centers = mxGetField(prhs[0],0,"centers"); // null if not provided.

    if (mxIsDouble(q.data)) {
        mexFunction2<double,double>(nlhs,plhs,q); // there is no question here, everything is in double
    } else {
        // Faster but results differ from jakob_kmeans due to lower-precision
        // mexFunction2<float,float>(nlhs,plhs,q);
        
        // keeping centers always in double precision (even if data is float) may provide better precision without
        // increasing memory consumption of data. however, it usually slows down due to continuously upcasting
        // from float to double. an alternative would be to convert into double only once but that I haven't
        // implemented but should give the same results.
        mexFunction2<float,double>(nlhs,plhs,q); // keep centers in float
    }

}

