
/**
 * Author: R. Gokberk Cinbis
 */



#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstring>
#include <stdint.h> // to get uint8_t
#include <limits>
#include <cstdarg>
#include "cn_matlabarrays.hpp"
#include "cn_macros.hpp"
#include "cn_threads.hpp"
#include "cn_pq.hpp"
#include "cn_blas.hpp"

using std::runtime_error;
using std::vector;
using cn::mex::getcell;
using cn::mex::numel;
using cn::mex::ndims;
using cn::mex::getscalar;
const uint64_t ks = cn::pq::ks;

/**
 * Matlab arguments.
 */
template <class T>
struct matargs_t {
    uint64_t n;
    uint64_t k;

    uint64_t C;
    vector <uint64_t> ds_c;  // (1 C)
    vector <uint64_t> nsq_c; // (1 C)

    uint64_t Z;
    vector <uint64_t> z2c; // (1 Z) base 0.

    uint64_t nsq_total;
    uint64_t d_total;

    const uint8_t *XT;
    const T *wscr;
    const T *W;

    vector <const T *> centers; // (1 C)

    matargs_t() {}; // default constructor
    matargs_t( const matargs_t &from ); // disable copy constructor
};


// requires: args.C
// se
template<class T>
void getz2c(const mxArray *m_z2c,matargs_t <T> &args) {
    args.Z = numel(m_z2c);
    args.z2c.reserve(args.Z);
    args.nsq_total = 0;
    args.d_total = 0;
    const double *z2c = cn::mex::getDenseData<double>( m_z2c );
    for ( uint64_t z = 0 ; z < args.Z; ++z ) {
        uint64_t c = (uint64_t) (z2c[z] + 0.5); // double to uint64 with rounding.
        if ( c > args.C )
            mexErrMsgTxt("invalid z2c entry, has to be <= C");
        if ( c < 1 )
            mexErrMsgTxt("invalid z2c entry, has to be base 1");
        c -= 1; // convert to base 0.
        args.z2c.push_back( c ); 
        args.nsq_total += args.nsq_c[c];
        args.d_total += args.ds_c[c]  * args.nsq_c[c];
    }
}


template<class T>
void getcenters(const mxArray *m_centers,matargs_t<T> &args) {
    args.C = numel(m_centers);
    args.ds_c.reserve(args.C);
    args.nsq_c.reserve(args.C);
    args.centers.reserve(args.C);
    for ( uint64_t c = 0; c < args.C; ++c ) {
        const mxArray *A = getcell(m_centers,c);
        args.ds_c.push_back( cn::mex::getdim(A,1) );
        args.nsq_c.push_back( cn::mex::getdim(A,3) );
        cn::mex::assert2( ndims(A) <= 3, "ndims(centers{c}) <= 3" );
        cn::mex::assert2( cn::mex::getdim(A,2) == ks, "ks incompability in centers array" );
        args.centers.push_back( cn::mex::getDenseData<T>(A) );
    }
}

template <class T>
mxArray *process(const mxArray *m_XT,const mxArray *m_wscr,const mxArray *m_W,const mxArray *m_centers,const mxArray *m_z2c) {
    matargs_t <T> args;

    // first: centers & z2c
    getcenters(m_centers,args);
    getz2c(m_z2c,args);

    if (args.nsq_total != cn::mex::getdim(m_XT,2))
        throw runtime_error("pq code length is incompatible");
    args.n = cn::mex::getdim(m_XT,1);
    args.XT = cn::mex::getDenseData<uint8_t>(m_XT);

    // wscr
    if ( mxIsEmpty(m_wscr) ) {
        args.wscr = NULL;
    } else {
        if (args.n != cn::mex::getdim(m_wscr,1))
            throw runtime_error("Invalid data: size(wscr,1)!=n");
        if (args.Z != cn::mex::getdim(m_wscr,2))
            throw runtime_error("Invalid data: size(wscr,2)!=Z");
        args.wscr = cn::mex::getDenseData<T>(m_wscr);
    }

    // W
    if (args.n != cn::mex::getdim(m_W,1))
        throw runtime_error("Inner product dimensions are incompatible");
    args.k = cn::mex::getdim(m_W,2);
    args.W = cn::mex::getDenseData<T>(m_W);

    // alloc output
    mxArray * m_out = cn::mex::createArrayWoInit(cn::mex::type2MxClassID<T>(),args.d_total,args.k);
    T *out = cn::mex::getDenseData<T>(m_out);

    cn::pq::XTTmtimesW(
            args.ds_c,
            args.nsq_c,
            args.centers,
            args.z2c,
            args.n,
            args.k,
            args.XT,
            args.wscr,
            args.W,
            out);

    return m_out;
}

enum { a_X = 0, a_wscr, a_W, a_centers, a_z2c };
void mexFunction(int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[])
{
    if ( nrhs != 5 )
        mexErrMsgTxt("invalid number of input");
    if ( nlhs != 1 )
        mexErrMsgTxt("invalid number of output");

    const mxArray *m_XT = prhs[a_X];
    const mxArray *m_wscr = prhs[a_wscr];
    const mxArray *m_W = prhs[a_W];
    const mxArray *m_centers = prhs[a_centers];
    const mxArray *m_z2c = prhs[a_z2c];

    mxClassID T = mxGetClassID( m_W );

    try {
        if ( T == mxDOUBLE_CLASS ) {
            plhs[0] = process<double>(m_XT,m_wscr,m_W,m_centers,m_z2c);
        } else { 
            plhs[0] = process<float>(m_XT,m_wscr,m_W,m_centers,m_z2c);
        }

    } catch(runtime_error &err) { 
        mexErrMsgTxt(err.what());
    }

}



