#include "cn_fvgmm.hpp"

/**
 * Author: R. Gokberk Cinbis.
 */


namespace cn { namespace fvgmm { 

    namespace utils {

        // return sqrt(max(val,eps)) assuming val is non-negative.
        double mysqrt(const double &val) {
            if (val < 1e-16)
                return 1e-8;
            return sqrt(val);
        }

        // precompute fisher vector normalizer. store in the same ordering: (1+2*D K)
        void setfvnormalizer(const mwSize K,const mwSize D,const double *mix,const double *vars,double *fvnormalizer)
        {
#pragma omp parallel for
            for ( mwIndex k = 0; k < K; ++k ) {
                const double &pi = mix[k];
                double *out = fvnormalizer + (1+2*D) * k;
                const double *v = vars + D * k;

                // sqrt( mix * (1 - mix) )
                *(out++) = mysqrt(pi * (1-pi));

                // sqrt( mix * mu )
                for (mwIndex d = 0; d < D; ++d) {
                    *(out++) = mysqrt( pi * v[d] );
                }

                // sqrt(2*pi) * vars
                for (mwIndex d = 0; d < D; ++d) {
                    *(out++) = mysqrt(pi * 2) * v[d];
                }

            }

        }

    }

    PatchDesc::PatchDesc() {
        mix = NULL;
    }

    /// Set dictionary parameters. 
    void PatchDesc::setdictionary(const mwSize D,const mwSize K,const double *mix,const double *means,const double *vars) {
        this->D = D;
        this->K = K;
        this->mix = mix;
        this->means = means;
        this->vars = vars;
        fvlen_k = 1+2*D;

        // precompute fisher vector normalizer. store in the same ordering: (1+2*D K)
        fvnormalizer_sh.reset( new double[(1+2*D) * K], DeleterArray<double> );
        fvnormalizer_ptr = fvnormalizer_sh.get();
        utils::setfvnormalizer(K,D,mix,vars,this->fvnormalizer_ptr);

        fv_mix_zeropost_sh.reset( new double[K], DeleterArray<double> );
        fv_mix_zeropost_ptr = fv_mix_zeropost_sh.get();
        for ( mwIndex k = 0; k < K; ++k ) {
            fv_mix_zeropost_ptr[k] = (0 - this->mix[k]) / fvnormalizer_ptr[fvlen_k*k]; 
        }
    }

    void PatchDesc::getpatchfv_full(const double *desc,const mwSize sp,const double *q,const mwIndex *qind, double * out) {

        // make a first pass to just set derivatives wrt mixture as if p(k|x)=0 for them.
        // assume that out is zero-initialize, therefore, there is no need to initialize for fv_mu and fv_var.
        // then non-zero entries will be overwritten below.
        for (mwSize k = 0; k < K; ++k) { 
            out[fvlen_k*k] = fv_mix_zeropost_ptr[k];
        }

        for (mwSize ki = 0; ki < sp; ++ki) { 
            const mwSize &k = qind[ki];
            getpatchfv_k(desc,q[ki],k,out + fvlen_k * k);
        }
    }

    void PatchDesc::getpatchfv_k(const double *desc,const double q, const mwSize k, double * out) {
        if ( k >= K ) 
            throw runtime_error("0 <= k < K is required.");

        double *fv_mu = out + 1;      // means fv
        double *fv_var = out + 1 + D;  // variances fv (actually fv w.r.t. precision)

        const mwSize Dk = D * k;
        const double *means = this->means + Dk;
        const double *vars= this->vars + Dk;
        const double *fvnormalizer = this->fvnormalizer_ptr + fvlen_k * k;

        // fv_mix = (p(k|x)-pi) / sqrt(pi*(1-pi))
        out[0] = (q - this->mix[k]) / *(fvnormalizer++); 

        // fv_mu = p(k|x) * (x - mu) / sqrt(pi * var)
        for (mwIndex d = 0; d < D; ++d) 
            fv_mu[d] = q * (desc[d] - means[d]) / *(fvnormalizer++);

        // fv_var = p(k|x) * (var - (x-mu)^2) / ( var * sqrt(2*pi) )
        for (mwIndex d = 0; d < D; ++d) {
            const double diff = desc[d] - means[d];
            fv_var[d] = q * (vars[d] - diff * diff) / *(fvnormalizer++);
        }

    } 

    void PatchDesc::getpatchfv_hardassign(const double *desc,const mwSize k, double * out) {
        if ( k >= K ) 
            throw runtime_error("0 <= k < K is required.");

        double *fv_mu = out + 1;      // means fv
        double *fv_var = out + 1 + D;  // variances fv (actually fv w.r.t. precision)

        const mwSize Dk = D * k;
        const double *means = this->means + Dk;
        const double *vars= this->vars + Dk;
        const double *fvnormalizer = this->fvnormalizer_ptr + fvlen_k * k;

        // fv_mix = (p(k|x)-pi) / sqrt(pi*(1-pi))
        out[0] = (1 - this->mix[k]) / *(fvnormalizer++); 

        // fv_mu = p(k|x) * (x - mu) / sqrt(pi * var)
        for (mwIndex d = 0; d < D; ++d) 
            fv_mu[d] = (desc[d] - means[d]) / *(fvnormalizer++);

        // fv_var = p(k|x) * (var - (x-mu)^2) / ( var * sqrt(2*pi) )
        for (mwIndex d = 0; d < D; ++d) {
            const double diff = desc[d] - means[d];
            fv_var[d] = (vars[d] - diff * diff) / *(fvnormalizer++);
        }

    } 


} }

