#pragma once

/**
 * Author: R. Gokberk Cinbis.
 */


#include "cn_memory.hpp"
#include <cmath>
#include <string>

using std::runtime_error;

/**
 * Gaussian Mixture Model implementation, especially useful for posterior calculations.
 * Object is thread-safe, therefore, can be used for parallel tasks.
 */

namespace cn {

    template <class T>
    class MixOfGauss {

        public:

            /**
             * Create based on a pre-trained mixture model.
             *
             * mix: (1 K)
             * mu:  (D K)
             * var: (D K)
             */
            MixOfGauss( uint64_t a_D, uint64_t a_K, shared_ptr<const T> a_mix, shared_ptr<const T> a_mu, shared_ptr<const T> a_var)  
                : D(a_D), K(a_K), mix(a_mix), mu(a_mu), var(a_var) {
                
                precalc();

            }

            /**
             * Compute posteriors wrt dictionary for a single data point.
             *
             * x    (D 1)   Data point.
             * x2   (D 1)   Precomputed x^2.
             * q    (K 1)   Output array to store posteriors, preallocated.
             */
            void calcpostSingle( const T *x, const T *x2, T * q ) const {

                // logcond = (XT.^2).'*info.w_x2;
                // logcond = logcond + XT.'*info.w_x;
                // logcond = bsxfun(@plus,logcond,info.bias);
               
                T *w_x = this->w_x.get();
                T *w_x2 = this->w_x2.get();
                T *bias = this->bias.get();
                T *logprior = this->logprior.get();

                {
                    // compute p(x,k)
                    T *tmp = q;
                    for (uint64_t k = 0; k < K; ++k) {
                        const T *a = x;
                        const T *a2 = x2;
                        *tmp = *(bias++) + *(logprior++);
                        for (uint64_t d = 0; d < D; ++d) {
                            *tmp += *(w_x++) * *(a++) + *(w_x2++) * *(a2++);
                        }
                        ++tmp;
                    }
                }
                 
                // softmax(q) gives the posterior
               softmaxinplace(K,q);
                
            }

            uint64_t getK() const {
                return K;
            }

        private:

            uint64_t D;
            uint64_t K;
            shared_ptr<const T> mix;
            shared_ptr<const T> mu;
            shared_ptr<const T> var;

            // precalculations
            shared_ptr<T> logprior;
            shared_ptr<T> w_x;  // (D K)
            shared_ptr<T> w_x2; // (D K)
            shared_ptr<T> bias; // (1 K)
            shared_ptr<T> prec; // (D K)


            /**
             * Do pre-calculations for fast posterior calculation.
             * SEE cn_gaussdiag_precalc()
             */
            void precalc() {

                setprec();
                setbias();
                w_x = times( D*K, mu, prec );
                w_x2 = times( D*K, -0.5, prec );
                setlogprior();

                /*
                // Seems OK compared to matlab.
                printMatrix(D,K,prec.get(),"prec");
                printMatrix(1,K,bias.get(),"bias");
                printMatrix(D,K,w_x.get(),"w_x");
                printMatrix(D,K,w_x2.get(),"w_x2");
                printMatrix(1,K,logprior.get(),"logprior");
                */

            }


            /**
             * Compute precisions from variances.
             */
            void setprec() {
                prec.reset( new T[ D * K ], DeleterArray<T> ); // uninitilized
                const T *x_var = var.get();;
                T *x_out = prec.get();
                for ( uint64_t k = 0; k < K; ++k ) {
                    for ( uint64_t d = 0; d < D; ++d ) {
                        *(x_out++) = 1 / *(x_var++); // dont regularize. if regularization is needed, that should have been done during training.
                    }
                }
            }


            /**
             * bias =-0.5*sum(prec.*(mu.^2) , 1) + (-D/2)*log(2*pi) + (-1/2)*sum(log(vars),1)
             *      =-0.5* ( sum(prec.*(mu.^2) , 1) + D*log(2*pi) + sum(log(vars),1) )
             */
            void setbias() {

                if (!prec)
                    throw runtime_error("bug: prec is needed");

                bias.reset( new T [ 1*K ], DeleterArray<T> ); // un initilized
                T const1 =  ( (T) D ) * std::log(2*cn::pi<T>());
                T * x_bias = bias.get();
                T * x_prec = prec.get();
                const T * x_mu = mu.get();
                const T * x_var = var.get();
                for (uint64_t k = 0; k < K; ++k) {
                    *x_bias = const1; // initialized here.
                    for (uint64_t d = 0; d < D; ++d) {
                        *x_bias += (*x_mu) * (*x_mu) * (*x_prec);
                        *x_bias += std::log( *x_var );
                        ++x_mu;
                        ++x_var;
                        ++x_prec;
                    }
                    *x_bias *= -0.5;
                    ++x_bias;
                }
            }

            /**
             * log(mix)
             */
            void setlogprior() {
                logprior.reset( new T [ 1*K ], DeleterArray<T> ); // un-initilized
                const T *x_mix = mix.get();
                T *x_logprior = logprior.get();
                const T eps = (T) 1e-40;
                for (uint64_t k = 0; k < K; ++k) {
                    *(x_logprior++) = std::log(*(x_mix++) + eps);
                }
                
            }

            /**
             * scalar * vector of length N.
             */
            shared_ptr<T> times(const mwSize N,const T x, shared_ptr<const T> y) {
                shared_ptr <T> res( new T[ N ], DeleterArray<T> );
                const T *b = y.get();
                T *o = res.get();
                for (uint64_t i = 0; i < N; ++i ) 
                    *(o++) = x * *(b++);
                return res;
            }


            /**
             * vector * vector of length N.
             */
            shared_ptr<T> times(const mwSize N,shared_ptr<const T> x, shared_ptr<const T> y) {
                shared_ptr <T> res( new T[ N ], DeleterArray<T> ); // uninitilized
                const T *a = x.get();
                const T *b = y.get();
                T *o = res.get();
                for (uint64_t i = 0; i < N; ++i ) 
                    *(o++) = *(a++) * *(b++);
                return res;
            }

            /**
             * softmax(x){
             *    x -= max(x)
             *    x = exp(x)
             *    x /= sum(x)
             * }
             * gives the posterior
             */
            static void softmaxinplace(uint64_t N, T *q) {
                if (N<1)
                    return;

                // find max
                T *y = q;
                T max = *(y++);
                for (uint64_t i = 1; i < N; ++i) {
                    if (max < *y)
                        max = *y;
                    ++y;
                }

                // subtract max and get exponent
                y = q;
                for (uint64_t i = 0; i < N; ++i) {
                    *y = std::exp(*y - max);
                    ++y;
                }

                // calc sum of elements
                T sum = 0;
                y = q;
                for (uint64_t i = 0; i < N; ++i) {
                    sum += *(y++);
                }

                // div by the sum
                y = q;
                for (uint64_t i = 0; i < N; ++i) {
                    *y /= sum;
                    ++y;
                }

            }
 
            /**
             * Utility function to print a matrix.
             */
            template <class TT> 
            static void printMatrix( const uint64_t M, const uint64_t N, const TT * vals, const char *name ) {

                printf("%s: \n",name);
                for ( uint64_t i = 0; i < M; ++i) {
                    for ( uint64_t j = 0; j < N; ++j) {
                        printf("%4.2f ",(double) vals[M * j + i]);
                    }
                    printf("\n");
                }
                printf("\n");

            }

    };

}

