#ifndef CN_MATLAB_ARRAYS
#define CN_MATLAB_ARRAYS 1

/**
 * Author: R. Gokberk Cinbis.
 */


#include "mex.h"
#include <stdint.h>
#include <limits>
#include <stdexcept>
#include <sstream>
#include <exception>
#include <map>
using std::string;
using std::stringstream;
using std::runtime_error;

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))


/**
 * This file contains utility functions for mex interfacing.
 * In case of an error, these functions throw runtime_error (instead of matlab exception)
 * because otherwise try{} results in unreadable error messages.
 */

namespace cn{ namespace mex {

    namespace util {

        /**
         * Allows easy initialization of std::map
         */
        template <typename T, typename U> class create_map
        {
            private:
                std::map<T, U> m_map;
            public:
                create_map(const T& key, const U& val)
                {
                    m_map[key] = val;
                }

                create_map<T, U>& operator()(const T& key, const U& val)
                {
                    m_map[key] = val;
                    return *this;
                }

                operator std::map<T, U>()
                {
                    return m_map;
                }
        };
    } // end of util

    /*********************** TODO: MOVE THESE FUNCTIONS TO SOMEWHERE ELSE **********/

    /**
     * Use this rather than mxAssert. mxAssert is not good since 
     * -- it works with "-g" option which disable code optimization
     * -- it causes crash-like message in matlab rather than a error string.
     * Use assert2 instead of assert to avoid confusion.
     *
     * Although this function has nothing to do with mex, it is handly to define
     * here.
     */
    inline void assert2(bool expr, const char *errmsg="assertion failed (assert)")
    {
        if (!expr) {
            throw runtime_error(errmsg);
        }
    }



    /*********************** MIXED FUNCTIONS TO BE CATEGORIZED **********/


    /// Check whether a field exists.
    bool isfield(const mxArray *arr,const char *fieldname) {
        return mxGetFieldNumber(arr,fieldname) != -1;
    }

    // Convert from Matlab class ID to class name string. May throw std runtime_error exception!
    const string &classID2String(const mxClassID &id) {

        // initialized at the very first call
        static std::map<mxClassID,string> c = util::create_map<mxClassID,string> 
            (mxUINT8_CLASS,"uint8")(mxUINT16_CLASS,"uint16")(mxUINT32_CLASS,"uint32")(mxUINT64_CLASS,"uint64")
            (mxINT8_CLASS,"int8")(mxINT16_CLASS,"int16")(mxINT32_CLASS,"int32")(mxINT64_CLASS,"int64")
            (mxDOUBLE_CLASS,"double")(mxSINGLE_CLASS,"single")(mxLOGICAL_CLASS,"logical");

        try {
            return c.at(id);
        } catch(...) {
            throw runtime_error("mxClassID2String::Unknown class id");
        }
    }

    // Convert from Matlab class string to class id. May throw std runtime_error exception!
    const mxClassID &classString2ID(const string &name) {

        // initialized at the very first call
        static std::map<string,mxClassID> c = util::create_map<string,mxClassID> 
            ("uint8",mxUINT8_CLASS)("uint16",mxUINT16_CLASS)("uint32",mxUINT32_CLASS)("uint64",mxUINT64_CLASS)
            ("int8",mxINT8_CLASS)("int16",mxINT16_CLASS)("int32",mxINT32_CLASS)("int64",mxINT64_CLASS)
            ("double",mxDOUBLE_CLASS)("single",mxSINGLE_CLASS)("logical",mxLOGICAL_CLASS);

        try {
            return c.at(name);
        } catch(...) {
            throw runtime_error(string("mxClassString2ID::Unknown class name: ") + name);
        }

    }

    /**
     * Check that array is dense & non-complex.
     */
    void assertRealDenseArray(const mxArray *arr,const char *varname="?")
    {
        if (mxIsSparse(arr)) 
            throw runtime_error( string("assertRealDenseArrayClass::Array is sparse: ") + string(varname) );
        if (mxIsComplex(arr)) 
            throw runtime_error( string("assertRealDenseArrayClass::Array is complex: ") + string(varname) );
    }

    /**
     * Check that array is dense & non-complex and check its class
     */
    void assertRealDenseArrayClass(const mxArray *arr,const mxClassID &cls,const char *varname="?")
    {
        assertRealDenseArray(arr,varname);
        if (mxGetClassID(arr)!=cls)
            throw runtime_error( string("assertRealDenseArrayClass::Array does not have the required class: ") + string(varname) );
    }

    /**
     * Example: shared_ptr<int> p( (int *) mxMalloc(..), DeleterMex )
     *
     * Important: This has to be passed with reset() too!
     *
     * SEE cn_memory.hpp
     */
    template <class T> void DeleterMex(T *x) { if (x!=NULL) mxFree(x); }

    /**
     * Return if arr is mxCHAR_CLASS.
     */
    bool isstring(const mxArray *arr) {
        return (mxGetClassID(arr)==mxCHAR_CLASS);
    }


    /** 
     * Number of elements in a matlab array.
     */
    inline size_t numel( const mxArray * x ) {
        return mxGetNumberOfElements(x);
    }

    /**
     * Copy string array.
     */
    string getstring(const mxArray *arr,const char *varname="") {
        char *result = mxArrayToString(arr);
        if (result == NULL) {
            string e = string(varname) + string(" is not a proper string");
            throw runtime_error(e.c_str());
        }
        string s(result);
        mxFree(result);
        return s;
    }
    /**
     * Copy string of length 1.
     */

    char getchar(const mxArray *arr,const char *varname="") {
        if ( numel(arr) != 1 ) {
            string e = string(varname) + string(" has numel!=1.");
            throw runtime_error(e.c_str());
        }

        char *result = mxArrayToString(arr);
        if (result == NULL) {
            string e = string(varname) + string(" is not a proper string");
            throw runtime_error(e.c_str());
        }
        char c = result[0];
        mxFree(result);
        return c;
    }

    /**
     * Whether a given array has only a single element
     */
    bool isscalar(const mxArray *arr) {

        return (mxGetNumberOfElements(arr) == 1);
    }

    /**
     * Set mex-lock status.
     */
    void setMexLock(bool state) {
        if (state) {
            if (!mexIsLocked()) {
                mexLock();
                printf("Mex file is locked \n");
            } else {
                printf("Mex file is already locked \n");
            }
        } else {
            if (!mexIsLocked()) {
                printf("Mex file is already unlocked \n");
            } else { 
                while (mexIsLocked())
                    mexUnlock();
                printf("Mex file is unlocked \n");
            }
        }
    }

    /**
     * Get a particular field of a struct array.
     */
    const mxArray *getfield(const mxArray *arr,const mwIndex index,const char *fieldname) {
        const mxArray *result = mxGetField(arr,index,fieldname);
        if (result == NULL) {
            string e = "mxGetField: Missing field: ";
            e += fieldname;
            mexErrMsgTxt(e.c_str());
        }
        return result;
    }

    /**
     * Get a particular field of a scalar struct.
     */
    const mxArray *getfield(const mxArray *arr,const char *fieldname) {
        if (mxGetNumberOfElements(arr)!=1)
            throw runtime_error("arr is not a (struct)-scalar");
        return getfield(arr,0,fieldname);
    }

    /**
     * Returns the requested field number. If not exists, returns an error message
     */
    int getfieldnumber(const mxArray *pm,const char *fieldname) {
        int result = mxGetFieldNumber(pm,fieldname);
        if (result == -1) {
            mexPrintf("Failed: mxGetFieldNumber(...,fieldname=%s\n",fieldname);
            throw runtime_error("mxGetFieldNumber has failed, see above for the details.");
        }
        return result;
    }

    /**
     * Whether struct array has a particular field
     * TODO Currently returns false if array is empty.
     */
    bool hasfield(const mxArray *arr,const char *fieldname) {
        const mxArray *result = mxGetField(arr,0,fieldname);
        if (result == NULL) {
            return false;
        }
        return true;
    }

    /**
     * Number of dimensions in a matlab array.
     */
    inline size_t ndims( const mxArray * x ) {
        return mxGetNumberOfDimensions(x);
    }

    /**
     * Get matlab cell. index is base 0.
     */
    const mxArray * getcell( const mxArray * cellarr, size_t index, const char *cellname="" ) {
        if ( !mxIsCell(cellarr) ) // I have seen at least one case where mxGetCell returns non-NULL ptr with non-cell array.
            mexErrMsgTxt((string("Not a cell array ") + string(cellname)).c_str());
        const mxArray * res = mxGetCell(cellarr,index);
        if ( res==NULL ) 
            mexErrMsgTxt((string("cannot get cell ") + string(cellname)).c_str());
        return res;
    }

    /** 
     * Get value from a numeric scalar.
     */
    double getscalar(const mxArray *arr) {
        assert2(mxGetNumberOfElements(arr)==1,"getscalar::invalid size (not scalar)");
        if ( !mxIsNumeric(arr) && !mxIsLogical(arr) )
            throw runtime_error("getscalar::not numeric or logical");
        return mxGetScalar(arr);
    }

    /**
     * dim is in matlab coordinates. dim is base 1 (1..ndim).
     */
    mwSize getdim(const mxArray *arr,mwIndex dim) {
        assert2(dim>=1,"getdim: dim<1 is invalid");
        if (mxGetNumberOfDimensions(arr) < dim)
            return 1;
        else
            return mxGetDimensions(arr)[dim-1];
    }

    /************************* FUNCTIONS TO CONVERT MATLAB and C++ DATA TYPES ***********************************/

    // Convert from C numeric data type to Matlab's class id. Throws runtime_error in case of unknown class.
    template <class T> mxClassID type2MxClassID() { throw std::runtime_error("type2MxClassID::Unknown"); }
    template <> mxClassID type2MxClassID <uint8_t> () { return mxUINT8_CLASS; }
    template <> mxClassID type2MxClassID <uint16_t> () { return mxUINT16_CLASS; } 
    template <> mxClassID type2MxClassID <uint32_t> () { return mxUINT32_CLASS; } 
    template <> mxClassID type2MxClassID <uint64_t> () { return mxUINT64_CLASS; } 
    template <> mxClassID type2MxClassID <int8_t> () { return mxINT8_CLASS; } 
    template <> mxClassID type2MxClassID <int16_t> () { return mxINT16_CLASS; }
    template <> mxClassID type2MxClassID <int32_t> () { return mxINT32_CLASS; } 
    template <> mxClassID type2MxClassID <int64_t> () { return mxINT64_CLASS; } 
    template <> mxClassID type2MxClassID <float> () { return mxSINGLE_CLASS; } 
    template <> mxClassID type2MxClassID <double> () { return mxDOUBLE_CLASS; } 
    template <> mxClassID type2MxClassID <bool> () { return mxLOGICAL_CLASS; } 
    // char (8-bit) is NOT mxCHAR_CLASS (16-bit)

    /**
     * Return size of a single element from a particular mex class.
     */
    mwSize sizeofMxClassID(const mxClassID &id) {
        if (id==mxUINT8_CLASS) {
            return sizeof(uint8_t);
        } else if (id==mxUINT16_CLASS) {
            return sizeof(uint16_t);
        } else if (id==mxUINT32_CLASS) {
            return sizeof(uint32_t);
        } else if (id==mxUINT64_CLASS) {
            return sizeof(uint64_t);
        } else if (id==mxINT8_CLASS) {
            return sizeof(int8_t);
        } else if (id==mxINT16_CLASS) {
            return sizeof(int16_t);
        } else if (id==mxINT32_CLASS) {
            return sizeof(int32_t);
        } else if (id==mxINT64_CLASS) {
            return sizeof(int64_t);
        } else if (id==mxSINGLE_CLASS) {
            return sizeof(float);
        } else if (id==mxDOUBLE_CLASS) {
            return sizeof(double);
        } else if (id==mxLOGICAL_CLASS) {
            return sizeof(bool);
        } else {
            throw std::runtime_error("sizeofMxClassID::Unknown");
        }
    }



    /************************* FUNCTIONS TO ACCESS DATA ***********************************/

    /**
     *  return pointer to the data of a dense numeric array.
     *  example: const double *data = getDenseData<double>(arr)
     */
    template <class T> T *getDenseData(const mxArray *arr,const char *varname=NULL)
    {
        if (mxIsSparse(arr)) { // has to be dense
            std::string x = "getDenseData::Sparse array is not allowed.";
            if (varname!=NULL) { x += std::string(" (variable:") + std::string(varname) + std::string(")"); }
            throw runtime_error(x.c_str());
        }

        if (mxGetClassID(arr)!=type2MxClassID<T>()) { // has to have the correct class
            std::string x = "getDenseData::Invalid variable data class";
            x += std::string(" Required:") + std::string(classID2String(type2MxClassID<T>()));
            x += std::string(" Actual:") + std::string(classID2String(mxGetClassID(arr)));
            if (varname!=NULL) { x += std::string(" (variable:") + std::string(varname) + std::string(")"); }
            throw runtime_error(x.c_str());
        }

        T * ptr = (T *) mxGetData(arr);
        if (!mxIsEmpty(arr)) {
            assert2(ptr!=NULL,"Incorrectly non-empty array has null pointer");
        }
        return ptr;
    }

    /** 
     * returns all diagonal element values of the sparse matrix W
     * output  diagelems is an array of size min(#row,#col) newly allocated with mxCalloc.
     * returns number of non zero diagonal elements
     */
    mwSize sparseMatGetDiagonals(const mxArray *array_ptr, double **diagelems) {

        assert2(mxIsDouble(array_ptr),"Input should be double");
        assert2(mxIsSparse(array_ptr),"Input should be sparse");

        /* Get the starting positions of all four data arrays. */
        double  *pr  = mxGetPr(array_ptr); // has length nnz(W)
        mwIndex *ir  = mxGetIr(array_ptr); // has length nnz(W)
        mwIndex *jc  = mxGetJc(array_ptr); // has length size(W,2)
        mwSize total = 0;

        mwSize nrow = mxGetM(array_ptr);
        mwSize ncol = mxGetN(array_ptr);

        // init output array to zero
        mwSize numdiag = MIN(nrow,ncol);
        double *out = (double *) mxCalloc(numdiag,sizeof(double));
        if (out == NULL) {
            throw runtime_error("Cannot allocate vector for binary elements");
        }

        mwSize nnzdiag = 0;

        // go over elements in colwise order
        for (mwSize col=0; col<ncol; col++)  {
            mwIndex startRowI = jc[col];
            mwIndex stopRowI = jc[col+1];
            mwIndex matcoor_j = col+1; // in matrix coordinates
            if (startRowI == stopRowI)
                continue;
            else {
                for (mwIndex curRowI = startRowI; curRowI < stopRowI; curRowI++)  {
                    mwIndex matcoor_i = ir[curRowI]+1;
                    if (matcoor_i == matcoor_j) {
                        // set the nonzero diagnal element
                        double value = pr[total];
                        out[matcoor_i-1] = value;

                        // increment counter for additional information
                        nnzdiag += 1;
                    }
                    total = total + 1;
                }
            }
        }

        *diagelems = out;
        return nnzdiag;
    }

    /************************* MATH FUNCTIONS ***********************************/

    /************************* ALLOCATORS ***********************************/

    /**
     * Allocate real-valued matlab array without initialization.
     * createArrayWoInit(type,size1,size2,size3...)
     * At least first two 2 dimensions is required.
     */
    mxArray *createArrayWoInit(mxClassID type,mwSize s0,mwSize s1,mwSize s2=1,mwSize s3=1) {

        mxArray * a = mxCreateNumericMatrix(0,0,type,mxREAL);
        if (a==NULL)
            throw std::runtime_error("mxCreateNumericArray failed");

        // can increase #dims here
        const mwSize ndim = 4;
        mwSize dims[ndim];
        dims[0] = s0;
        dims[1] = s1;
        dims[2] = s2;
        dims[3] = s3;

        mwSize numel = 1;
        for (mwIndex i = 0; i < ndim; ++i)
            numel *= dims[i];

        void *data = mxMalloc(sizeofMxClassID(type)*numel);
        if (data==NULL && (numel>0))
            throw std::runtime_error("mxMalloc failed");
        mxSetData(a,data);
        mxSetDimensions(a,dims,ndim);

        return a;
    }

    /**
     * Allocate matlab array with zeros initialized.
     * createArrayZeros(type,size1,size2,size3...)
     * At least first two 2 dimensions is required.
     * This is just a convenient wrapper to mxCreateNumericArray()
     */
    mxArray *createArrayZeros(mxClassID type,mwSize s0,mwSize s1,mwSize s2=1,mwSize s3=1) {
        // can increase #dims here
        const mwSize ndim = 4;
        mwSize dims[ndim];
        dims[0] = s0;
        dims[1] = s1;
        dims[2] = s2;
        dims[3] = s3;

        mxArray * a = mxCreateNumericArray(ndim,dims,type,mxREAL);
        if (a==NULL)
            throw std::runtime_error("mxCreateNumericArray failed");

        return a;
    }

    /**
     * Allocate matlab cell array.
     * createCellArray(size1,size2,size3...)
     * At least first two 2 dimensions is required.
     * This is just a convenient wrapper to mxCreateCellArray()
     */
    mxArray *createCellArray(mwSize s0,mwSize s1,mwSize s2=1,mwSize s3=1) {
        // can increase #dims here
        const mwSize ndim = 4;
        mwSize dims[ndim];
        dims[0] = s0;
        dims[1] = s1;
        dims[2] = s2;
        dims[3] = s3;

        mxArray * a = mxCreateCellArray(ndim,dims);
        if (a==NULL)
            throw std::runtime_error("mxCreateCellArray failed");

        return a;
    }

    /**
     * check if allocated, throw error if not.
     * Name carefully to avoid name-clash with mxFree() and free().
     */
    void * matMalloc(mwSize n,const char *purpose="?") {
        if (n == 0)
            return NULL;
        void * res = mxMalloc(n);
        if (res == NULL) {
            stringstream ss;
            ss << "mxMalloc  failed for n=" << n << " purpose="  << purpose;
            throw runtime_error(ss.str());
        }
        return res;
    }

    /**
     * check if allocated, throw error if not.
     * Name carefully to avoid name-clash with mxFree() and free().
     */
    void * matCalloc(mwSize n,mwSize size,const char *purpose="?") {
        if ((n*size) == 0)
            return NULL;
        void * res = mxCalloc(n,size);
        if (res == NULL) {
            stringstream ss;
            ss << "mxCalloc  failed for n=" << n << "size=" << size << " purpose="  << purpose;
            throw runtime_error(ss.str());
        }
        return res;
    }

    /**
     * Dont free if NULL.
     * Name carefully to avoid name-clash with mxFree() and free().
     */
    void matFree(void *ptr) {
        if (ptr!=NULL)
            mxFree(ptr);
    }

    /************************* MISC FUNCTIONS ***********************************/

    /**
     * Prints any given (2D) matrix.
     */
    void printmatrix(const double *data,mwSize height,mwSize width) 
    {
        for (mwSize y = 0; y<height; y++) {
            for (mwSize x =0; x<width; x++) {
                mexPrintf("%.2f\t",*(data + x * height + y));
            }
            mexPrintf("\n");
        }

    }

    /**
     * Prints any given (2D) matrix.
     */
    void printmatrix(const bool *data,mwSize height,mwSize width)
    {
        for (mwSize y = 0; y<height; y++) {
            for (mwSize x =0; x<width; x++) {
                if (*(data + x * height + y))
                    mexPrintf("1\t");
                else
                    mexPrintf("0\t");
            }
            mexPrintf("\n");
        }
    }

    /**
     * Prints any given (2D) matrix as matlab array.
     */
    template <class T> void printmatrix(const mxArray *mat) 
    {
        const mwSize *dims = mxGetDimensions(mat);
        const T *data = getDenseData<T>(mat);
        assert2(mxGetNumberOfDimensions(mat)==2,"mat is not a 2D array!");
        for (mwSize y = 0; y<dims[0]; y++) {
            for (mwSize x =0; x<dims[1]; x++) {
                mexPrintf("%.2f\t",(double) *(data + x * dims[0] + y));
            }
            mexPrintf("\n");
        }
    }

} } // end of namespace cn


#endif

