#pragma once
#include <stdlib.h>
#include <stdint.h>
#include "cn_os.hpp"

/**
 * Author: R. Gokberk Cinbis.
 */


#ifdef CN_CPP11
    #include <memory>
    using std::shared_ptr;
#else
    #include <tr1/memory>
    using std::tr1::shared_ptr;
#endif

// todo: should be fine but not tested yet in detail.
// todo: convert function pointers into functor objects, which are
//       surprisingly faster... though I dont understand how! can we use
//       singleton objects?

namespace cn {

    /**
     * Example: shared_ptr<int> p( (int *) malloc(..), DeleterC )
     * Reminder: &DeleterC and DeleterC are both valid ways of passing function pointers.
     *
     * Important: This has to be passed with reset() too!
     * template is useful in some cases, though essentially void is sufficient.
     */
    template <class T> void DeleterC(T *x) { if (x!=NULL) free(x); }

    /**
     * Deleter that does not delete.
     *
     * Example: shared_ptr<int> p( external_pointer, DeleterNone )
     *
     * Important: This has to be passed with reset() too!
     * template is useful in some cases, though essentially void is sufficient.
     */
    template <class T> void DeleterNone(const T *x) { if (x!=NULL) {} } // no op 

    /** 
     * Example: shared_ptr<int> p( new int [..], DeleterArray<int>()
     *
     * Important: This has to be passed with reset() too!
     */
    template <class T> void DeleterArray(T *x) { if (x!=NULL) delete [] (x); }

    /**
     * Shared Pointer and auto-deallocator for arrays created. By default,
     * uses "delete []", therefore, input should be created via "new []".
     * Alternatively, DeleterNone, DeleterC, DeleterMex can be used.
     *
     * Although similar, it has soem differences compared to shared_ptr.
     * Here, some important details:
     *  - It is OK to pass NULL pointer and it is the state of empty SharedArray.
     *  - Allows direct array access via operator overloading.
     *  - operator[] is explicitly defined as an inline function.
     *  - Does not support Function Objects as deleters since it requires embedding 
     *  another object inside (eg, DeleterClass <T,D> del) so that SharedArray is
     *  still a single-parameter class template. To avoid the overhead of allowing
     *  function objects, it only allows function pointers, as above.
     *
     * Named with capitals to distinguish from stl/boost versions.
     * An alternative might be shared_ptr< vector< T >>, though, here we have more flexibility.
     * 
     * todo: && of C++11 support.
     *
     * Design Note: Does not allow direct pointer arithmetic since a) it causes conflicts with bool operator,
     * b) it should result in a raw pointer for efficiency reasons, which can cause mistakes.
     *
     * SEE http://svn.boost.org/svn/boost/trunk/boost/smart_ptr/shared_array.hpp
     */
    template <class T> class SharedArray {

        typedef void (*Deleter)(T *);
        
        public: 

            /**
             * Initialize with an array pointer and optionally a custom Deleter.
             * If arrptr = NULL, empty array pointer is created.
             * Deleter d can be DeleterArray,DeleterNone,DeleterC,DeleterMex.
             */
            explicit SharedArray( T * arrptr = NULL, Deleter d = DeleterArray ) {
                // "explicit" avoid accidental conversion of pointers into SharedArray objects.
                resethelper( arrptr, d );
            }

            /// Copy pointer.
            SharedArray( SharedArray & from ) {
                copyhelper( from );
            }

            /// Destructor.
            ~SharedArray() {
                removepointer();
            }

            /**
             * Assign a new raw pointer. Dealloc old array if necessary.
             * Note that new Deleter always defaults to DeleterArray.
             */
            void reset( T * arrptr = NULL, Deleter d = DeleterArray ) { 
                removepointer();
                resethelper( arrptr, d );
            }

            /// Direct array access. Undefined behaviour if contains NULL pointer.
            inline T & operator[]( const uint64_t &index ) const { 
                return ptr[index]; 
            }

            /// Copy a shared pointer.
            void operator=( SharedArray &from ) { // from is NOT const since we're NOT const.
                removepointer();
                copyhelper( from );
            }

            /// Conversion into bool, whether it contains any array currently.
            operator bool() const { 
                return ptr!=NULL;
            }

            /// Get the contained pointer.
            inline T * get() const 
            {
                return ptr;
            }



        private:

            T * ptr;
            uint64_t *counter;
            Deleter d;

            /// Delete the existing pointer if necessary, otherwise, just reduce counter.
            void removepointer() {
                if (ptr!=NULL)  {
                    if (*counter <= 1) {
                        delete counter;
                        d(ptr); // delete [] ptr;
                    } else {
                        --( *counter );
                    }
                }
                // No need: ptr = NULL; counter = NULL;
            }

            /// Do reset() while ignoring the current pointer
            void resethelper( T * arrptr, Deleter d ) {
                this->ptr = arrptr;
                this->d = d;
                if (ptr != NULL) {
                    counter = new uint64_t;
                    *counter = 1; 
                } else {
                    counter = NULL;
                }
            }

            /// Copy a shared pointer while ignoring the current pointer.
            void copyhelper( SharedArray & from ) {
                this->ptr = from.ptr;
                this->counter = from.counter;
                this->d = from.d;
                if (ptr != NULL) 
                    ++( *counter );
            }

    };


}


