#include <vector>
#include <cmath>
#include <stdio.h>
#include <stdint.h>
#include <vector>
#include <stdexcept>
#include "cn_macros.hpp"
#include "cn_memory.hpp"
#include "cn_math.hpp"  
#include "cn_blas.hpp"   
#include "cn_threads.hpp"   

using std::vector;

/**
 * Author: R. Gokberk Cinbis.
 */


namespace cn { namespace pq {

    const uint64_t ks = 256;

    namespace _ { // internals


        /**
         * Take sqrt(max(vals,eps)) in place. Does not multithread.
         */
        template <class T> void sqrtmax(const uint64_t n, const shared_ptr<T> &vals, const T eps=0);

        /**
         * Print Matrix.
         */
        template <class T> void printmat(const char *name,const uint64_t m, const uint64_t n, const T *mat);

        /**
         * If TOp==shared_dist_l2normperz_op_dotprod:   Compute dotprod_l2normperz()
         * If TOp==shared_dist_l2normperz_op_sqeucdist: Compute sqeucdist_l2normperz()
         */
        template <class T,class TOp> void shared_dist_l2normperz(
                const vector<uint64_t>   &nsq_c,
                const vector <const T *> &dotprods,
                const vector <const T *> &sqnorms,
                const vector<uint64_t>   &z2c, 
                const uint64_t nx, 
                const uint64_t ny, 
                const uint8_t *X,
                const uint8_t *Y,
                T *D);

        class shared_dist_l2normperz_op_dotprod;
        class shared_dist_l2normperz_op_sqeucdist;

        // utility function for shared_dist_l2normperz.
        template <class T,class TOp> inline void shared_dist_l2normperz_op(T &_d,const uint64_t Z);

    }

    /**
     * Return sum(nsq_c(z2c)). Assumes that z2c has proper, base-0 indexing.
     */
    uint64_t get_nsq_total(const vector<uint64_t> &nsq_c, const vector<uint64_t> &z2c);

    /** 
     * Compute the squared L2 norms of PQ-coded vectors.
     *
     * nsq_c    (1 C) Number of subquantizers in each compressor.  
     * sqnorms  (1 C) {(ks nsq_c)} Squared-norm of each center of each compressor.
     * z2c      (1 Z) Mapping from channels to compressors.
     * nx       Number of descriptors.
     * X        (nsq_total nx) where nsq_total = sum(nsq_c(z2c)).
     * perz     Whether to return squared-norm for each z separately.
     * out      perz=0: (1 nx) Should be already zero-initialized!
     *          perz=1: (Z nx) Should be already zero-initialized!
     */
    template <class T> void squaredl2norm( 
            const vector<uint64_t>  &nsq_c,
            const vector<const T *> &sqnorms,
            const vector<uint64_t>  &z2c, 
            const uint64_t nx, 
            const uint8_t *X,
            const bool perz,
            T *out);

    /**
     * nsq_c    (1 C)
     * sqeucs   (1 C) {(ks*(ks+1)/2 nsq_c)}
     * z2c      (1 Z)
     * nx
     * ny
     * X        (nsq_total nx) where nsq_total = sum(nsq_c(z2c)).
     * Y        (nsq_total ny)
     * D        (nx ny)
     *
     * "D" must be zero-initialized before!
     */
    template <class T> void sqeucdist(
            const vector<uint64_t>  &nsq_c,
            const vector<const T *> &sqeucs,
            const vector<uint64_t>  &z2c, 
            const uint64_t nx, 
            const uint64_t ny, 
            const uint8_t *X,
            const uint8_t *Y,
            T *D);

    /**
     * nsq_c    (1 C)
     * dotprods (1 C) {(ks*(ks+1)/2 nsq_c)}
     * z2c      (1 Z)
     * nx
     * ny
     * X        (nsq_total nx) where nsq_total = sum(nsq_c(z2c)).
     * Y        (nsq_total ny)
     * D        (nx ny)
     *
     * "D" must be zero-initialized before!
     */
    template <class T> void dotprod(
            const vector<uint64_t>  &nsq_c,
            const vector<const T *> &dotprods,
            const vector<uint64_t>  &z2c, 
            const uint64_t nx, 
            const uint64_t ny, 
            const uint8_t *X,
            const uint8_t *Y,
            T *D);


    /**
     * Compute sum_z < x_z/||x_z||, y_z/||y_z|| >
     * To get squared-L2 distance where X,Y is per-cell normalized implicitly,
     * do 2(Z-dotprod_l2normperz()).
     *
     * nsq_c      (1 C)
     * dotprods   (1 C) {(ks*(ks+1)/2 nsq_c)}
     * sqnorms    (1 C) {(ks nsq_c)}  
     * z2c        (1 Z)
     * nx
     * ny
     * X        (nsq_total nx) where nsq_total = sum(nsq_c(z2c)).
     * Y        (nsq_total ny)
     * D        (nx ny)
     *
     * "D" must be zero-initialized before!
     */
    template <class T> void dotprod_l2normperz(
            const vector<uint64_t>   &nsq_c,
            const vector <const T *> &dotprods,
            const vector <const T *> &sqnorms,
            const vector<uint64_t>   &z2c, 
            const uint64_t nx, 
            const uint64_t ny, 
            const uint8_t *X,
            const uint8_t *Y,
            T *D) {
        _::shared_dist_l2normperz<T,_::shared_dist_l2normperz_op_dotprod>(nsq_c,dotprods,sqnorms,z2c,nx,ny,X,Y,D);
    }

    /**
     * Compute sum_z || x_z/(||x_z||) - y_z/(||y_z||) ||^2
     *
     * Arguments are same as in dotprod_l2normperz().
     *
     * To get squared-L2 distance where X,Y is per-cell normalized implicitly,
     * this function internally does 2(Z-dotprod_l2normperz()).
     */
    template <class T> void sqeucdist_l2normperz(
            const vector<uint64_t>   &nsq_c,
            const vector <const T *> &dotprods,
            const vector <const T *> &sqnorms,
            const vector<uint64_t>   &z2c, 
            const uint64_t nx, 
            const uint64_t ny, 
            const uint8_t *X,
            const uint8_t *Y,
            T *D) {
        _::shared_dist_l2normperz<T,_::shared_dist_l2normperz_op_sqeucdist>(nsq_c,dotprods,sqnorms,z2c,nx,ny,X,Y,D);
    }


    /**
     * Precompute scores lookup table for a given linear classifier.
     * Uses LAPACK for matrix multiplication, therefore, requires equivalent data type.
     * Although using "float" can be evil for computations, here it might be OK given that
     * we're summing at most "ds_c" dimensions.
     *
     * ARGUMENTS
     * ds_c       (1 C)  
     * nsq_c      (1 C)
     * centers    (1 C)         {(ds_c ks nsq_c)}
     * z2c        (1 Z)
     * ncls       (1 1)
     * allw       (ncls d_total)  
     * alls       (ncls ks nsq_total)   Output array, preallocated without initialization.
     *
     */
    template <class T> void linclslookup(
            const vector<uint64_t>    &ds_c,
            const vector<uint64_t>    &nsq_c,
            const vector <const T *>  &centers,
            const vector<uint64_t>    &z2c, 
            const uint64_t            ncls,  
            const T *                 allw,
            T *                       alls);


    /*
     * Compute XT'*W where XT is pq-encoded.
     *
     * ARGUMENTS
     * ds_c       (1 C)  
     * nsq_c      (1 C)
     * centers    (1 C)         {(ds_c ks nsq_c)}
     * z2c        (1 Z)
     * N          (1 1)
     * K          (1 1)
     * XT         (N nsq_total)
     * [wscr]     (N Z)         Optional descriptor scaling per-Z. Set to NULL to disable. The same scaling can
     *                          be applied directly on "out" but it is faster to do it here.
     * W          (N K)  
     * out        (d_total K)   Output array, preallocated without initialization.
     *
     */
    template <class T> void XTTmtimesW(
            const vector<uint64_t>    &ds_c,
            const vector<uint64_t>    &nsq_c,
            const vector<const T *>   &centers,
            const vector<uint64_t>    &z2c, 
            const uint64_t            N,  
            const uint64_t            K,  
            const uint8_t *           XT,
            const T *                 wscr,
            const T *                 W,
            T *                       out);

    /*
     * Sum each point over z components with per-z,per-point weights.
     * d_z = ds_c[z2c[z]] * nsq_c[z2c[z]] has to be constant for all z.
     *
     * ARGUMENTS
     * ds_c       (1 C)  
     * nsq_c      (1 C)
     * centers    (1 C)         {(ds_c ks nsq_c)}
     * z2c        (1 Z)
     * N          (1 1)
     * X          (nsq_total N)
     * W          (Z N)  
     * out        (d_z N)       Output array, preallocated with all ZEROS.
     *
     */
    template <class T> void wsumz(
            const vector<uint64_t>    &ds_c,
            const vector<uint64_t>    &nsq_c,
            const vector<const T *>   &centers,
            const vector<uint64_t>    &z2c, 
            const uint64_t            N,  
            const uint8_t *           X,
            const T *                 W,
            T *                       out) ;


}}



namespace cn {
    namespace pq {
#include "private/cn_pq.templates.cpp"
    }
}

