#pragma once 

#include <stdint.h>
#include <stdlib.h>
#include <map>
#include <math.h>
#include <stdexcept>
#include "cn_sort.hpp"

/**
 * Author: R. Gokberk Cinbis.
 */


// Generates equivalent random numbers with the Matlab's implementation of Twister algorithm.
// This is just a wrapper to the code provided by the University of Hiroshima.
//
// EXAMPLE
// RandStream.setGlobalStream( RandStream.create('mt19937ar','seed',100) );
// rand(100,1)
//
// KNOWN BUG  
// Exceptionally, for seed=0, results differ. Therefore, for now, seed=0 is not allowed.
//
// R.G.Cinbis, Sep 2012

namespace cn{

    class RandGenTwister {

        public:
            
            /// Initialize with an integer.
            RandGenTwister(uint32_t seed);

            /// Initialize with a double value, which is simply floored into uint32_t.
            RandGenTwister(double seed);

            /// Destructor
            ~RandGenTwister();

            /// Generate a random number in range (0,1)
            double rand();

            /// Generate a random integer in range [0,maxval)
            template <class T>
            T randi(T upper) {
                T res = (T) ceil(rand()*upper);
                if (res < 1)
                    res = 1; // very unlikely
                --res; // from [1,N] to [0,N) for consistency.
                return res;
            }

            /**
             * Generate a random permutation of [0:N), equivalent to Matlab's "randperm(N)-1"
             * out should be pre-allocated and should have length >= N. 
             *
             * NOTE In extreme cases where rand() produces exactly the same number, result may differ slightly among equivalent numbers.
             */
            template <class T>
            void randperm(const T N, T *out) {
                if (N==0)
                    return;

                double *tmp = (double *) malloc(sizeof(double) * N);
                if (tmp==NULL)
                    throw std::runtime_error("RandGenTwister::malloc failed in randperm()");

                {
                    double *tmp2 = tmp;
                    for (T i = 0; i < N; ++i) {
                        *(tmp2++) = this->rand();
                    }
                }

                // cn::stable_getorder(N,tmp,out,"ascend"); // matlab equivalent but slower 
                cn::stable_getorder(N,tmp,out,"ascend"); 
                free(tmp);        
            }


            /**
             * Generate a random permutation of 1:N and return first k elements from it efficiently.
             * "out" should be pre-allocated and should have length >= K.
             * In Matlab, equivalent to "x=randperm(1,N)-1;x=x(1:K)".
             * The resultant random state is equivalent to the one after calling randperm(N).
             *
             * For very large N and a small K, randpermfirstk does work faster than randperm(), in addition,
             * randpermfirstk() can work well where randperm() gives out-of-memory error.
             *
             * NOTE This is NOT randperm(N,K) of Matlab, which returns different results depending on both N / and K.
             * NOTE In contrast to randperm(), I guess, here we're 100% compatible.
             */
            template <class T>
            void randpermfirstk(const T N,const T K, T *out) { 
                if (K==0)
                    return;
                if (N<K)
                    throw std::runtime_error("RandGenTwister::randpermk::N>=K is required");

                // use a map to store smallest k random numbers without generating full rand(1,N).
                // use multimap instad of map in case we get same random number again.
                // multimap guarantees to preserve the original order of items, therefore, result is guaranteed to be
                // equivalent to stable_sort.
                std::multimap <double, T> m;

                for (T i = 0; i < K; ++i) {
                    double r = this->rand();
                    m.insert( std::pair <double, T> ( r, i ) );
                    //m.emplace(r,i);
                }
                double maxsofar = m.rbegin()->first;
                for (T i = K; i < N; ++i) {
                    double r = this->rand();
                    // add this number of it is smaller than the largest among the first k 
                    if (r < maxsofar) { // use < instead of <=.
                        //m.emplace(r,i);
                        m.insert( std::pair <double, T> ( r, i ) );

                        // if we erase last element, then total complexity is ~ 2*N*log(K)==N*log(K^2)
                        // if we dont erase the last element, then total complexity is ~ N*log(N)
                        // for K<sqrt(N), it is beneficial to remove to remove the old element.
                        // more importantly, memorywise it is alwasy more efficient. so we always remove it.
                        typename std::multimap <double, T>::iterator it = m.end(); // typename is needed for obscure reasons
                        --it;
                        m.erase( it );  // remove the old, largest item

                        maxsofar = m.rbegin()->first; // this is not necessarily r.
                    }
                }

                typename std::multimap<double, T>::iterator it = m.begin();
                for (T i = 0; i < K; ++i) {
                    *(out++) = it->second;
                    ++it;
                }
            }

        private:
            unsigned long *mt; /* the array for the state vector  */
            int mti; /* mti==N+1 means mt[N] is not initialized */

            void init_genrand(unsigned long s);
            unsigned long genrand_int32();

    };

}


