#pragma once
#include <vector>
#include <algorithm>
#include <stdexcept>
#include <string.h>
#include <stdint.h>
#include <functional>

/**
 * Author: R. Gokberk Cinbis.
 */


/**
 * Functions to efficiently sort values and their sorting indices.
 * Provides simple functions to use with Matlab rand(),randi(),randperm() compabilities.
 *
 * Tcount is usually either uint64_t or size_t or uint32_t.
 */

namespace cn {

    namespace __sort__helper__ {
        // -- HELPER NAMESPACE, SEE BELOW FOR THE ACTUAL FUNCTIONS --

        template<class T,class Tcount>
        class cn_getorder_helper_ascend {
            public:
                cn_getorder_helper_ascend(const T *arr2): arr(arr2) {} 
                inline bool operator()(Tcount const& lhs, Tcount const& rhs) const {
                    return arr[lhs] < arr[rhs]; 
                }
            private:
                const T * arr;
        };

        template<class T,class Tcount>
        class cn_getorder_helper_descend {
            public:
                cn_getorder_helper_descend(const T *arr2): arr(arr2) {} 
                inline bool operator()(Tcount const& lhs, Tcount const& rhs) const {
                    // NO! return arr[lhs] >= arr[rhs]; // In this case, original order of elements is not
                    // preserved, instead, reversed.
                    return arr[lhs] > arr[rhs];
                }
            private:
                const T * arr;
        };

        /**
         * Does the main sorting and update the 'ind' array without changing the 'arr' array, which is const.
         */
        template <class T,class Tcount>
        void cn_getorder_helper_generic(const Tcount &N,const T *arr,Tcount *ind,const char *dir,bool stable) {

            for (Tcount i = 0; i < N; ++i) 
                ind[i] = i;

            if (!stable) {
                if (strcmp(dir,"ascend")==0)
                    std::sort(ind,ind+N,cn_getorder_helper_ascend<T,Tcount>(arr));
                else if (strcmp(dir,"descend")==0)
                    std::sort(ind,ind+N,cn_getorder_helper_descend<T,Tcount>(arr));
                else
                    throw std::runtime_error("Invalid dir argument to cn::getorder");
            } else {
                if (strcmp(dir,"ascend")==0)
                    std::stable_sort(ind,ind+N,cn_getorder_helper_ascend<T,Tcount>(arr));
                else if (strcmp(dir,"descend")==0)
                    std::stable_sort(ind,ind+N,cn_getorder_helper_descend<T,Tcount>(arr));
                else
                    throw std::runtime_error("Invalid dir argument to cn::stable_getorder");
            }
        }


        /**
         * First sort and get order.
         * Then apply the ordering on the values array.
         */
        template <class T,class Tcount>
        void cn_sort_helper_generic(const Tcount &N,T *arr,Tcount *ind,const char *dir,bool stable) {

            cn_getorder_helper_generic(N,arr,ind,dir,stable);

            T *tmp = (T *) malloc(sizeof(T)*N);
            if (tmp==NULL)
                throw std::runtime_error("cn::sort failed, not enough memory");

            {
                T *tmp2 = tmp;
                Tcount *ind2 = ind;
                for (Tcount i = 0; i < N; ++i) {
                    *(tmp2++) = arr[ *(ind2++) ]; 
                }
            }

            memcpy( arr, tmp, sizeof(T) * N );
            free(tmp);
        }

    }

    /**
     * Return order of the elements without changing ordering of the input elements.
     *
     * INPUT
     * N        Number of elements in the array
     * arr      A basic array containing the unsorted values. This array will not be changed.
     * ind      A pre-allocated array of length N to store output order. Output values will be in the range: 0 <= ind < N.
     * dir     'ascend' | 'descend'
     *
     */
    template <class T,class Tcount>
    void getorder(const Tcount &N,const T *arr,Tcount *ind,const char *dir) {
        // templates have to be fully specified within the header
        __sort__helper__::cn_getorder_helper_generic(N,arr,ind,dir,false);
    }


    /**
     * Sort the input array.
     *
     * INPUT
     * N        Number of elements in the array
     * arr      A basic array containing the unsorted values. This array will be sorted.
     * dir     'ascend' | 'descend'
     *
     */
    template <class T,class Tcount>
    void sort(const Tcount &N,T *arr,const char *dir) {
        if (strcmp(dir,"ascend")==0)
            std::sort(arr,arr+N,std::less<T>());
        else if (strcmp(dir,"descend")==0)
            std::sort(arr,arr+N,std::greater<T>());
        else
            throw std::runtime_error("Invalid dir argument to cn::sort");
    }

    /**
     * Sort the input array and return the order of the elements.
     *
     * INPUT
     * N        Number of elements in the array
     * arr      A basic array containing the unsorted values. This array will be sorted.
     * ind      A pre-allocated array of length N to store output order.
     * dir     'ascend' | 'descend'
     */
    template <class T,class Tcount>
    void sort(const Tcount &N,T *arr,Tcount *ind,const char *dir) {
        __sort__helper__::cn_sort_helper_generic(N,arr,ind,dir,false);
    }


    /**
     * Similar to getorder() except that uses stable_sort instead of sort.
     * "ind" will contain increasing indices among equal values.
     */
    template <class T,class Tcount>
    void stable_getorder(const Tcount &N,const T *arr,Tcount *ind,const char *dir) {
        __sort__helper__::cn_getorder_helper_generic(N,arr,ind,dir,true);
    }

    /**
     * Similar to getorder() except that uses stable_sort instead of sort
     * "ind" will contain increasing indices among equal values.
     */
    template <class T,class Tcount>
    void stable_sort(const Tcount &N,T *arr,Tcount *ind,const char *dir) {
        __sort__helper__::cn_sort_helper_generic(N,arr,ind,dir,true);
    }

    /**
     * Sort the input array using stable_sort.
     * For basic types, sort() should be used instead for efficiency, since output is indifferent.
     *
     * INPUT
     * N        Number of elements in the array
     * arr      A basic array containing the unsorted values. This array will be sorted.
     * dir     'ascend' | 'descend'
     */
    template <class T,class Tcount>
    void stable_sort(const Tcount &N,T *arr,const char *dir) {
        if (strcmp(dir,"ascend")==0)
            std::stable_sort(arr,arr+N,std::less<T>());
        else if (strcmp(dir,"descend")==0)
            std::stable_sort(arr,arr+N,std::greater<T>());
        else
            throw std::runtime_error("Invalid dir argument to cn::sort");
    }




}

