// - See cn_mycpuset() for more information about cpuset.txt
// - Assumes that cpu affinity via taskset is not altered during run time.
//   See my blog for reasons not to play with this setting. 
// - cpuset file is read each time or at certain intervals. In linux, this
//   file is usually located at /dev/shm, which reduces the overhead.
// - Never uses matlab functions, therefore, should be realtively safe.
//
// R.G.Cinbis Nov 2011
#pragma once

#include <omp.h>
#include "stdint.h"

// todo: if possible, we could regularly check cpu usage by OTHER processes.
// todo: we can use file change time avoid rereading this file too frequently
//       [this in not urgent. every 5000-10000 call (with reading file) has
//       an overhead of 1 sec]. 
// todo: number of secs passed

// dont include openmp to make sure caller does.

namespace cn {

    /**
     * * Sets the maximum number of threads and dynamic option.
     *
     * * Returns the maximum number of threads, which is the positive integer passed to omp_set_num_threads().
     * Returned value is independent of whether dynamic scheduling is enabled or not.
     *
     * * Note that dynamic scheduling does not depend on the passed "argmaxthread" option.
     *
     * --BEGIN--
     *  If argnthread==-1
     *      Results are directly loaded from the settings file.
     *      If settings file is not available, opt=0.
     *  else
     *      Do not read the settings file.
     *      opt=argnthread.
     *  end
     *
     *  If opt<0
     *      nthreads=abs(opt)
     *      use dynamic scheduling.
     *  elseif opt==0 
     *      nthreads=num cores.
     *      use dynamic scheduling.
     *  else
     *      nthreads=opt
     *      always use "nthreads"-many threads.
     *
     *  nthreads=min(nthreads,num cores).
     *
     *  call omp_set_num_threads, omp_set_dynamic
     *  return nthreads. 
     * --END--
     *
     */
    int setmaxthreads(int argnthread=-1,bool echo=false);

    // todo: change the interface. have separate functions setmaxthreads_auto(), setmaxthreads_override()

    /**
     * Sets the maximum number of threads and dynamic option based on the settings file.
     * Here we can pass an upper bound limit for the number of cores while sitt, which is not possible with setmaxthreads().
     *
     * --BEGIN--
     *  if settings file is available
     *      Read nthreads and dynamic scheduling options from the file with the rules given in setmaxthreads()
     *  otherwise
     *      nthreads=num cores.
     *      use dynamic scheduling.
     *
     *  if argnthread>0
     *      nthreads=min([nthreads,argnthread,num cores])
     *  else
     *      throw exception.
     *
     *  call omp_set_num_threads, omp_set_dynamic
     *  return nthreads.
     * --END--
     *
     * * Note that this function also does not allow having more threads than the number of cores.
     * 
     * * This is very similar to omp_set_num_threads(MIN(MAX_THREAD,setmaxthreads(-1,echo))).
     *
     * * Note that dynamic scheduling does not depend on the passed "argmaxthread" option.
     */
    int setmaxthreads_limit(int argmaxthread,bool echo=false);


    /**
     * Return path of the settings file.
     */
    const char *getcpusetpath();

    namespace _cn_threads {
        uint64_t posdivceil(uint64_t x, uint64_t y);

        // setmaxthreads() with an internal maximum number of threads constraint.
        int setmaxthreadsgeneric(int argnthread=-1,bool echo=false,int argmaxthread=-1);
    }

    uint64_t jobstarti(uint64_t njob,uint64_t nworker=omp_get_num_threads(),uint64_t thi=omp_get_thread_num());
    uint64_t jobendi(uint64_t njob,uint64_t nworker=omp_get_num_threads(),uint64_t thi=omp_get_thread_num());

}

