/*** template definitions directly included in the hpp file ***/

/**
 * Author: R. Gokberk Cinbis.
 */


namespace PoolObjDesc_Private {

    template <class mask_t>
        inline mask_t getPtWeightEps() {
            throw runtime_error("undefined");
        }

    template <>
        inline bool getPtWeightEps<bool>() {
            return false;
        }

    template <>
        inline float getPtWeightEps<float>() {
            return FvGmm_PoolObjDesc_PtWeightEps;
        }

    /**
     * V:       (4 N) Normalized locations of points.
     * spatgmm
     * tmpm     (1 N) Denotes the points inside the box.
     * Qspat    (Kspat N) Output posterior probabilities. Qspat(:,~tmpm) contains garbage values.
     */
    template <class mask_t>
        void getSpatialPosteriors( const size_t Npt, const double * V, const MixOfGauss<double> & spatgmm, const mask_t *tmpm, double * Qspat ) {

            size_t K = spatgmm.getK();
            mask_t eps = getPtWeightEps<mask_t>();
            for (size_t i = 0 ; i < Npt; ++i) {
#ifndef DEBUG_SPATFV
                if (tmpm[i] <= eps) { continue; }
#endif
                spatgmm.calcpostSingle(V + 4 * i,V + 4 * i + 2, Qspat + K * i);
            }

#ifdef DEBUG_SPATFV
            utils::printMatrix( 4, Npt, V, "Locs" );
            utils::printMatrix( K, Npt, Qspat, "Qspat" );
#endif

        }

    /**
     * Return patch locations [0,1] normalized wrt an object bb.
     *
     * obj_bb   (1 4) box coordinates.
     * N        Number of points in f_xy.
     * f_xy     (2 N)
     * tmpm     Output of getptmask4box. Points that are not inside the object are skipped.
     *          Corresponding V2(:,j) entries will contain garbage values.
     * V2       Preallocated output array of size (4 Npt)
     */
    template <class mask_t, class Tc>
        void getNormLocsWrtObject( const Tc * obj_bb, const mwSize &N, const Tc *f_xy, const mask_t *tmpm, double *V2 ) {

            /* Example:
             * Org coordinates: 
             *  5 6 7 8     (object [5 8])
             *  (5-5)/(8-5) = 0
             *  (8-5)/(8-5) = 1
             *  ie. should be normalized by height-1,width-1.
             */
            double x0 = obj_bb[0];
            double y0 = obj_bb[1];
            double Zx = obj_bb[2]-x0; if (Zx < 1) { Zx = 1; }
            double Zy = obj_bb[3]-y0; if (Zy < 1) { Zy = 1; }

            mask_t eps = PoolObjDesc_Private::getPtWeightEps<mask_t>();

            for (mwIndex pti = 0; pti < N; ++pti) {
#ifndef DEBUG_SPATFV
                if ( tmpm[pti] <= eps ) { continue; } // no need to compute
#endif

                const Tc *f_xy_pti = f_xy + 2 * pti;
                double *cur = V2 + 4 * pti;

                const Tc &fx = f_xy_pti[0];
                const Tc &fy = f_xy_pti[1];

                cur[0] = ( ((double) fx) - x0 ) / Zx;   // x
                cur[1] = ( ((double) fy) - y0 ) / Zy;   // y
                cur[2] = cur[0] * cur[0];               // x^2
                cur[3] = cur[1] * cur[1];               // y^2

            }

        }



}


template <class Tc>
PoolObjDesc<Tc>::PoolObjDesc(): patchfinal(false), minscs(NULL), maxscs(NULL), scstep(1) {
    mwSize initn = 50; // sth large enough 
    unsetmasks();
    allN.reserve(initn);
    allV.reserve(initn);
    allQ.reserve(initn);
    allQ_Ir.reserve(initn);
    allQ_Jc.reserve(initn);
    allLocs.reserve(initn);
}

/**
 * Pool descriptors.
 * Must be preallocated: out: (D K R), npts: (1 R)
 * out should have been pre-initialized to zeros.
 */
template <class Tc>
void PoolObjDesc<Tc>::poolboxdesc(double *out,npts_t *npts) {

    // printf("masktype: %d\n",(int) masktype);
    if (masktype==MaskNone || masktype==MaskBool || masktype==MaskBackground) {
#define POOL_BOXES
#include "poolobjdesc_pooldesc.cpp"
    } else if ( (masktype==MaskFloat) || (masktype==MaskImage) ) {
#define POOL_BOXES_WITH_PTWEIGHTS
#include "poolobjdesc_pooldesc.cpp"
    } else {
        throw runtime_error("Unknown masktype");
    }

}

/**
 * Pool Fisher Vectors.
 * Must be preallocated: out: (1+2*D K R), npts: (1 R)
 * out should have been pre-initialized to zeros.
 *
 * If avgfv=true, fv descriptor is normalized by the number of points in the box.
 */
template <class Tc>
void PoolObjDesc<Tc>::poolboxfv(double *out,npts_t *npts,bool avgfv) {

    if (masktype==MaskNone || masktype==MaskBool || masktype==MaskBackground) {

#define POOL_BOXES
#include "poolobjdesc_poolfv.cpp"

    } else if ( (masktype==MaskFloat) || (masktype==MaskImage) ) {

#define POOL_BOXES_WITH_PTWEIGHTS
#include "poolobjdesc_poolfv.cpp"

    } else {
        throw runtime_error("Unknown masktype");
    }




}

/**
 * Pool fishervectors according to a segmentation tree.
 * out should be pre-allocated and pre-initialized to zeros.
 * No descriptor normalization is done.
 */
template <class Tc>
void PoolObjDesc<Tc>::pooltreefv(double *out,npts_t *npts,shared_ptr<const id_t> idmap,const Tc imH,const Tc imW,const SingleTree <id_t> &tree) {

    setsegmentation( idmap, imH, imW );

    // For example, 1,2,4 can be the leaf nodes and 3 can be the root node. In this case, 
    // descriptor of node=3 will be set via accumFVsToParents(). Therefore, it should remain as
    // all-zeros until calling accumFVsToParents()
    //
    // even after the 2012-11-06 bug fix in stats2fv, we still need this behaviour
    // to avoid invalid obj_npts entries.
    pooltreefv_leaves(out,npts,false,true); 
    accumFVsToParents( out, npts, tree );

}


/**
 * Pool Spatial Fisher Vectors.
 * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
 * out should have been pre-initialized to zeros.
 */
template <class Tc>
void PoolObjDesc<Tc>::poolboxspatfv(double *out,npts_t *npts,bool avgfv) {

    if (masktype==MaskNone || masktype==MaskBool || masktype==MaskBackground) {
#define POOL_BOXES
#include "poolobjdesc_poolspatfv.cpp"
    } else if ( (masktype==MaskFloat) || (masktype==MaskImage) ) {
#define POOL_BOXES_WITH_PTWEIGHTS
#include "poolobjdesc_poolspatfv.cpp"
    } else {
        throw runtime_error("Unknown masktype");
    }

} // poolboxspatfv


/**
 * Pool Spatial Fisher Vectors over a segmentation tree.
 * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
 * out should have been pre-initialized to zeros.
 * No descriptor normalization is done.
 *
 * TODO This is somewhat slow. To speed up there are things that we can do:
 *      - For example, we can move information from lower levels to higher levels in order to find out
 *      points associated with each segment.
 *      - Similarly, we can also find out the bounding box of the upper level pretty much the same
 *      way.
 *      - Therefore, there is a lot to be done if it becomes necessary to speed up!
 */
template <class Tc>
void PoolObjDesc<Tc>::pooltreespatfv(double *out,npts_t *npts,shared_ptr<const id_t> idmap,const Tc imH,const Tc imW,const SingleTree <id_t> &tree) {

    setsegmentation( idmap, imH, imW ); // sets this->bbs

    // Since descriptor (= normalized locations) change based on the bounding box, we cannot use propagation
    // trick used above. instead, we'll use the multi-segmentation algorithm 
    //
    // We'll process each segmentation level independently. since base segments dont overlap
    // and since each segment has a unique parent, we know that segments are non-overlapping
    // within each level.

    pooltreespatfv_leaves(out,npts,false,true);
    const id_t treeheight = tree.getTreeHeight(); // level = [0,treeheight] 

    // to remember the bbs from lower levels
    vector< shared_ptr<const Tc> > all_bbs;
    vector< shared_ptr<const id_t> > all_idmaps;
    all_bbs.reserve( treeheight+1 );          // (4 R), base 0
    all_idmaps.reserve( treeheight+1 );

    // save the lowest level
    all_bbs.push_back( this->bbs );         // lowest level is the first, root will be the last
    all_idmaps.push_back( idmap );

    // utils::printMatrix( imH, imW, idmap.get(), "leafidmap" );

    // now process upper nodes
    // "level" = maxdisttoleaf
    for ( id_t level = 1; level <= treeheight; ++level) { // level=0 is already processed

        id_t * newidmap = new id_t[ imH * imW ](); // init to zero
        for ( id_t nodei = 1; nodei <= tree.getNumNode(); ++nodei ){

            if ( tree.getNodeMaxDistToLeaf(nodei) != level ) {
                continue; // node from a different level.
            }

            // propagate segmentation from the children to the current node
            id_t *chld = tree.getChildren( nodei );
            id_t numchld = tree.getNumChildren( nodei );

            // utils::printMatrix( 1, numchld, chld, ("children of " + utils::num2str(nodei)).c_str() );

            for ( id_t chj = 0; chj < numchld; ++ chj ) {

                const id_t ch = chld[chj]; // base 1
                const id_t chlevel = tree.getNodeMaxDistToLeaf( ch );
                const Tc * chbb = all_bbs.at(chlevel).get() + 4 * (ch-1);
                const id_t * chidmap = all_idmaps.at(chlevel).get();

                // fprintf(stderr,"%d ch=%d depth=%d chdepth=%d level=%d chlevel=%d \n", nodei, ch, (int) tree.getNodeDepth(nodei), (int) tree.getNodeDepth(ch), (int) level, (int) chlevel ) ;
                // utils::printMatrix(1,4,chbb,"chbb");

                for ( Tc x = chbb[0]; x <= chbb[2]; ++x ) { // loop in base 0
                    for ( Tc y = chbb[1]; y <= chbb[3]; ++y ) { // loop in base 0
                        if (chidmap[x * imH + y]==ch)
                            newidmap[x * imH + y] = nodei;
                    }
                }

            }

            // utils::printMatrix( imH, imW, newidmap, "idmap" );

        }

        // update segmentation
        shared_ptr< const id_t > newidmap_p( newidmap, DeleterArray<const id_t> );
        setsegmentation( newidmap_p, imH, imW ); // updates this->bbs
        all_idmaps.push_back( newidmap_p );
        all_bbs.push_back( this->bbs );

        // pool spatfv at the current level
        //
        // "It is important to use NonExistSegmentSkip so that descriptors from other levels
        // dont get overwritten when processing a particular level."
        //
        // even after the 2012-11-06 bug fix in stats2fv, we still need this behaviour
        // to avoid invalid obj_npts entries.
        pooltreespatfv_leaves(out,npts,false,true);

    }

}

/// Set dictionary parameters. This is not needed for just pooling without FV computation.
template <class Tc>
void PoolObjDesc<Tc>::setdictionary(const mwSize D,const mwSize K,const double *mix,const double *means,const double *vars) {
    this->D = D;
    this->K = K;
    this->mix = mix;
    this->means = means;
    this->vars = vars;

    // precompute fisher vector normalizer. store in the same ordering: (1+2*D K)
    this->fvnormalizer.reset( new double[(1+2*D) * K], DeleterArray<double> );
    utils::setfvnormalizer(K,D,mix,vars,this->fvnormalizer.get());

    meansSq.reset( new double[ D * K ], DeleterArray<double> );
    squarearray( means, D*K, meansSq.get() ); // (D K) cn_math.hpp
}

/// Forget any existing dictionary info.
template <class Tc>
void PoolObjDesc<Tc>::unsetDictionary() {
    D = 0;
    K = 0;
    mix = NULL;
    means = NULL;
    vars = NULL;
    fvnormalizer.reset();
    meansSq.reset();
}

template <class Tc>
mwSize PoolObjDesc<Tc>::getnumregions() {
    if (!bbs) // segmentation automatically sets this
        throw runtime_error("boxes or segmentation ids have not been set");
    return this->R;
}

/**
 * Set boxes for box-based pooling.
 * bbs: (4 R), base 0 coordinates.
 */
template <class Tc>
void PoolObjDesc<Tc>::setboxes(shared_ptr<const Tc> bbs,const mwSize R) {
    unsetmasks();
    this->R   = R;
    this->bbs = bbs;
}

/**
 * Remove any masks definitions done via setmasks().
 */
template <class Tc>
void PoolObjDesc<Tc>::unsetmasks() {
    masks_bool.reset();
    masks_float.reset();
    maskim.reset();
    this->masktype = MaskNone;
}

/**
 * Set per-box binary segmentation mask.
 * If this is defined, any method utilizing getptmask4box will be affected.
 */
template <class Tc>
void PoolObjDesc<Tc>::setmasks(const shared_ptr< vector< shared_ptr<const bool> > > &masks ) {
    if ( !bbs )
        throw runtime_error("cn_poolobjfvmog::first set bbs before setting masks");
    if ( !masks )
        throw runtime_error("empty pointer is passed!");
    if ( masks->size() != R )
        throw runtime_error("cn_poolobjfvmog::length(masks)!=R");
    unsetmasks();
    this->masks_bool = masks;
    this->masktype = MaskBool;
}


/**
 * Set per-box segmentation weights. Each point will be weighted based on its weight.
 * If this is defined, any method utilizing getptmask4box will be affected.
 */
template <class Tc>
void PoolObjDesc<Tc>::setmasks(const shared_ptr< vector< shared_ptr<const float> > > &masks ) {
    if ( !bbs )
        throw runtime_error("cn_poolobjfvmog::first set bbs before setting masks");
    if ( !masks )
        throw runtime_error("empty pointer is passed!");
    if ( masks->size() != R )
        throw runtime_error("cn_poolobjfvmog::length(masks)!=R");
    unsetmasks();
    this->masks_float = masks;
    this->masktype = MaskFloat;
}

template <class Tc>
void PoolObjDesc<Tc>::setmasks(const shared_ptr<const float> maskim,
        size_t maskim_H, size_t maskim_W, size_t maskim_B, 
        const shared_ptr<const int16_t> maskind) {

    if ( !bbs ) // we require "R".
        throw runtime_error("cn_poolobjfvmog::first set bbs before setting masks");

    if ( !maskim || !maskind )
        throw runtime_error("empty pointer is passed!");

    const int16_t *maskind2 = maskind.get();
    for ( size_t i = 0; i < R; ++i ) 
        if ( ( maskind2[i] < 0 )  || ( ((mwSize) maskind2[i]) >= maskim_B ) )
            throw runtime_error("masind has to be in [0,maskim_B)");

    // make sure all points have valid coordinates. this increases efficiency: we don't have to check
    // coordinate validity during pooling.
    size_t nscale = getnscale();
    for ( size_t sci = 0; sci < nscale; ++sci) {
        const mwSize &N = allN[sci];
        const Tc *f_xy = allLocs[sci].get();
        for ( size_t i = 0; i < N; ++i ) {
            const Tc &x = *(f_xy++);
            const Tc &y = *(f_xy++);
            if ( (x < 0) || (y < 0) || ( ((size_t) x) >= maskim_W) || ( ((size_t) y) >= maskim_H) ) { 
                throw runtime_error("0<=f_x<W, 0<=f_y<h is required when using MaskImage.");
            }
        }
    }

    unsetmasks();
    this->maskim = maskim;
    this->maskind = maskind;
    this->maskim_H = maskim_H;
    this->maskim_W = maskim_W;
    this->maskim_B = maskim_B;
    this->masktype = MaskImage;

}

/**
 * setmasks(MaskBackground)
 *      Pool descriptors that are strictly outside the bounding boxes.
 */
template <class Tc>
void PoolObjDesc<Tc>::setmasks(const MaskType &masktype) {
    if (masktype==MaskBackground) {
        // ok
    } else {
        mexErrMsgTxt("Forbidded masktype. See other setmasks() versions.");
    }

    unsetmasks();
    this->masktype = masktype;
}

/// make sure this is called after setboxes() or setsegmentation().
template <class Tc>
void PoolObjDesc<Tc>::setscales(const int16_t *minscs,const int16_t *maxscs,const int16_t scstep) {

    mwSize nscale = getnscale();
    if ( !bbs )
        throw runtime_error("R should have been set before calling setscales()");

    this->minscs = minscs;
    this->maxscs = maxscs;
    this->scstep = scstep;

    for (mwIndex i = 0; i < R; ++i) {
        if ( (minscs[i] < 0) || ((mwSize) minscs[i] >= nscale) || (maxscs[i] < 0) || ((mwSize) maxscs[i] >= nscale) )
            mexErrMsgTxt("minscs and maxscs should be in range [0,nscale)");
        if ( maxscs[i] < minscs[i])
            mexErrMsgTxt("maxscs < minscs does not make sense");
    }
    if (scstep <= 0)
        mexErrMsgTxt("scstep cannot be negative or zero");

}

/// remove scale information
template <class Tc>
void PoolObjDesc<Tc>::unsetscales() {
    this->minscs = NULL;
    this->maxscs = NULL;
    this->scstep = 1;
}

/// add a new scale of sparse posterior of shape (N Kapp).
/// note that in spatial fv, this->K = Kspatial, which is different from Kapp.
template <class Tc>
void PoolObjDesc<Tc>::pushPosteriors( mwSize N, mwSize Kapp, const double *Q, const mwIndex *Ir, const mwIndex *Jc ) {
    mwSize sci = allQ.size();
    pushN( N, sci );
    allQ.push_back(Q);
    allQ_Ir.push_back(Ir);
    allQ_Jc.push_back(Jc);
    if (sci == 0) {
        this->Kapp = Kapp;
    } else if ( this->Kapp != Kapp ) {
        std::stringstream ss;
        ss << "inconsistency in size(Q,2) is detected" << std::endl;
        ss << "prev: " << this->Kapp << "  now: " << Kapp << std::endl;
        throw runtime_error(ss.str());
    }
}

// add a new scale of locations. locations should be base 0.
template <class Tc>
void PoolObjDesc<Tc>::pushLoc( mwSize N, shared_ptr< const Tc > fxy ) {
    mwSize sci = allLocs.size();
    pushN( N, sci );
    allLocs.push_back( fxy );
}

// add a new scale of appearances
template <class Tc>
void PoolObjDesc<Tc>::pushDesc( mwSize Dapp, mwSize N, const double * d_cols ) {
    mwSize sci = allV.size();
    pushN( N, sci );
    allV.push_back( d_cols );
    if (sci == 0) {
        this->Dapp = Dapp;
    } else if ( this->Dapp != Dapp ) {
        throw runtime_error("inconsistency in size(d_cols,1) is detected");
    }
}

// call this immediately after adding posteriors,locations,[descriptors]
template <class Tc>
void PoolObjDesc<Tc>::finalizePatchInfo() {
    // check consistency and compute .nscale
    _nscale = allLocs.size();
    if (_nscale != allQ.size() || _nscale != allQ_Jc.size() || _nscale!=allQ_Ir.size())
        throw runtime_error("nscale is inconsistent for allQ*");
    if (allV.size() != 0 && allV.size() != _nscale )
        throw runtime_error("nscale is inconsistent for allV");
    patchfinal = true;
}

template <class Tc>
mwSize PoolObjDesc<Tc>::getnscale() {
    if (!patchfinal)
        throw runtime_error("call finalizePatchInfo() before asking nscale");
    return _nscale; // never access directly
}

template <class Tc>
mwSize PoolObjDesc<Tc>::getDapp() {
    if (!patchfinal)
        throw runtime_error("call finalizePatchInfo() before asking Dapp");
    return Dapp;
}

template <class Tc>
mwSize PoolObjDesc<Tc>::getKapp() {
    if (!patchfinal)
        throw runtime_error("call finalizePatchInfo() before asking Kapp");
    return Kapp;
}

template <class Tc>
mwSize PoolObjDesc<Tc>::getKdict() {
    if (!fvnormalizer)
        throw runtime_error("set dictionary before asking Kdict");
    return K;

}

/**
 * Get min..max inclusive scale for a particular region.
 */
template <class Tc>
void PoolObjDesc<Tc>::getMinMaxScaleForRegion( mwSize r, int16_t *minscale, int16_t *maxscale ) {
    if ( minscs!=NULL ) {
        *minscale = minscs[r];
        *maxscale = maxscs[r];
    } else {
        *minscale = 0;
        *maxscale = getnscale()-1; // getnscale() is unlikely to be ever the bottleneck.
    }
}

/**
 * A wrapper to omp_get_thread_num()
 */
template <class Tc>
mwSize PoolObjDesc<Tc>::getthreadindex( const mwSize & maxthread ) const {
    const mwSize threadi = omp_get_thread_num();
    if (threadi >= maxthread) 
        throw runtime_error("Bug: threadi >= maxthread");
    return threadi;
}

/**
 * Add new N (if needed) to allN vector. 
 */
template <class Tc>
void PoolObjDesc<Tc>::pushN( mwSize N, mwSize sci ) {
    patchfinal = false;
    try {
        mwSize N2 = allN.at( sci );
        if (N != N2)
            throw runtime_error("Inconsistency in Npt is detected");
    } catch (...) {
        if (allN.size() != sci )
            throw runtime_error("bug detected: allN is corrupt");
        allN.push_back( N );
    }
}

/// Create (2D K) matrices containing [d_cols; d_cols.^2] 
template <class Tc>
void PoolObjDesc<Tc>::catWithSquares(vector< shared_ptr<double> > *allV2) {

    if (allV2->size()>0)
        throw runtime_error("allV2 should be empty");

    mwSize nscale = getnscale();
    if (allV.size()!=nscale)
        throw runtime_error("allV is necessary");

    allV2->reserve(nscale);
    mwSize D2 = 2 * D;

    for (mwIndex sci = 0; sci < nscale; ++sci) {
        if (D2==0) {
            allV2->push_back(shared_ptr<double>());
            continue;
        }

        mwSize n = allN[sci];
        const double *from = allV[sci];
        double *to = new double [ D2 * n ];

#pragma omp parallel for
        for (mwIndex i = 0; i < n; ++i) {
            const double *f = from + D * i;
            double *t = to + D2 * i;
            memcpy(t,f,sizeof(double)*D); // copy d_cols(:,i)
            t += D;
            for (mwIndex d = 0; d < D; ++d ) {
                *(t++) = *f * *f; // concatenate d_cols(:,i).^2
                ++f;
            }
        }
        allV2->push_back( shared_ptr<double>(to,DeleterArray<double>) );

    }
}



/**
 * Convert 0-th,1-th and 2-nd order cumulated posteriors, descriptors, descriptor^2 (i.e.
 * moments*npts) into normalized fisher vectors with analytical approximation to Fisher Information
 * Matrix.  This function works on the FV of a single object.
 *
 * If there are no points inside, the obj_out will remain unchanged, assuming it is all-zeros.
 *
 * obj_out (2*D K) contains the following on the input.
 *      obj_out(1,k)         = sum_x(p(k|x))
 *      obj_out((1:D)+1,k)   = sum_x(p(k|x)*x)
 *      obj_out((1:D)+1+D,k) = sum_x(p(k|x)*x.^2)
 *
 * On the output, obj_out is either the average OR the total fisher vector.
 * 2013-04-30: Re-Confirmed (on paper) that avgfv=0 & avgfv=1 do compute fv consistent with each other.
 */
template <class Tc>
void PoolObjDesc<Tc>::stats2fv(double *obj_out,double obj_npts,bool avgfv) {
    // Important: Define obj_npts in double.

    if ( obj_npts < 1 ) {

        // BUG FIX: 2012-11-06
        //
        // If there are NO points inside, then the result should remain as all zeros (since we have
        // not pooled any features).
        //
        // Since we assume that output is zero-initialized, no need to do anything.
        //
        // NOTE: 2013-08-18
        //
        // (1) It is probably a good idea to set a truncation threshold option to here.
        // "obj_npts<1" makes sense only for non-weighted case. For the weighted cases, well,
        // it is not apriori clear what value should we assign to to this threshold.
        //
        // (2) There is actually a bug in here: If the obj_npts<threshold, it doesnot mean that
        // pooled obj_out is all zeros. In the weighted case, it might have collected a number 
        // of local FVs, but still the total weight might be below the threshold. Therefore, 
        // here it is best make "obj_out" all zeros.
        //
        // (3) I do not make such changes here for now because it is very easy to handle
        // these issues outside the code. One can easily check the npts output and assign
        // any thing with too-low npts to all zeros. 

        return; // dont make any changes!

    }


    const mwSize fvlen_k = 2 * D + 1;      // K * fvlen_k gives the total FV length. 
    double *fv_mix = obj_out;          // counts fv
    double *fv_mu = obj_out + 1;      // means fv
    double *fv_var = obj_out + 1 + D;  // variances fv (actually fv w.r.t. precision)

    const double *mix = this->mix;
    const double *means = this->means;
    const double *meansSq = this->meansSq.get();
    const double *vars= this->vars;
    const double *fvnormalizer = this->fvnormalizer.get();

    if (avgfv) {
        // for desctrans='avg', we always normalize by max(npts,1)

        for (mwSize k = 0; k < K; ++k) { // NOT parallelizable. This function is called within a par-for anyway.

            // 1) normalize counts by max(npts,1). store it. then subtract pi=*mix.
            *fv_mix /= obj_npts;
            double z = *fv_mix; // for speed, normalized counts
            *fv_mix -= *mix;

            // 2) normalize "sum_i p(k|xi) xi" by n
            for (mwIndex d = 0; d < D; ++d) 
                fv_mu[d] /= obj_npts;

            // 3) normalize "sum_i p(k|xi) xi^2" by n
            for (mwIndex d = 0; d < D; ++d) 
                fv_var[d] /= obj_npts;

            // 4) substract 2 * fv_mu*mu from fv_var
            for (mwIndex d = 0; d < D; ++d) 
                fv_var[d] -= 2 * fv_mu[d] * means[d];

            // 5) subtract z*mu from fv_mu
            for (mwIndex d = 0; d < D; ++d) 
                fv_mu[d] -= z * means[d];

            // 6) subtract fv_var from z*(vars-mu^2)
            for (mwIndex d = 0; d < D; ++d) 
                fv_var[d] = z * (vars[d] - meansSq[d]) - fv_var[d];

            // apply normalization
            for (mwIndex d = 0; d < fvlen_k; ++d)
                *(obj_out++) /= *(fvnormalizer++);

            fv_mix += fvlen_k;
            fv_mu += fvlen_k;
            fv_var += fvlen_k;
            ++mix;
            means += D;
            meansSq += D;
            vars += D;
        }
    } else {
        for (mwSize k = 0; k < K; ++k) { // NOT parallelizable. This function is called within a par-for anyway.
            double z = *fv_mix; 
            *fv_mix -= (*mix) * obj_npts; // pi*obj_npts

            // 4) substract 2 * fv_mu*mu from fv_var
            for (mwIndex d = 0; d < D; ++d) 
                fv_var[d] -= 2 * fv_mu[d] * means[d];

            // 5) subtract z*mu from fv_mu
            for (mwIndex d = 0; d < D; ++d) 
                fv_mu[d] -= z * means[d];

            // 6) subtract fv_var from z*(vars-mu^2)
            for (mwIndex d = 0; d < D; ++d) 
                fv_var[d] = z * (vars[d] - meansSq[d]) - fv_var[d];

            // apply normalization
            for (mwIndex d = 0; d < fvlen_k; ++d)
                *(obj_out++) /= *(fvnormalizer++);

            fv_mix += fvlen_k;
            fv_mu += fvlen_k;
            fv_var += fvlen_k;
            ++mix;
            means += D;
            meansSq += D;
            vars += D;
        }

    }
} // stats2fv()

/// Set number of threads. Return maximum number of threads.
template <class Tc>
mwSize PoolObjDesc<Tc>::setnumparallel( mwSize njob ) {
    mwSize maxthread = cn::setmaxthreads(); // maxthread is always positive.
    if (maxthread > njob) {
        maxthread = njob;
    } 
    if (maxthread < 1) {
        maxthread = 1;
    }
    omp_set_num_threads(maxthread); // may still do dynamic #thread adjustment if enabled by cn::setmaxthreads()
    return maxthread;
} // setnumparallel()


/// Single pass to find #points within each segment into (R nscale) matrix.
template <class Tc>
shared_ptr<mwSize> PoolObjDesc<Tc>::countPtsPerRegionPerScale(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) {

    // note: it is better for the output to remain as (R nscale) since 
    // we're looping over the image in this function. So, counting here better be efficient.

    mwSize nscale = getnscale();
    shared_ptr<mwSize> ptcounts( new mwSize[R*nscale](), DeleterArray<mwSize> ); // (R nsegment) , zero initialized!
    const id_t *idmap2 = idmap.get();
    for (mwSize sci = 0; sci < nscale; ++sci) {
        mwSize *c = ptcounts.get() + R * sci;
        const mwSize &N = allN[sci];
        const Tc *fx = allLocs[sci].get();
        const Tc *fy = fx + 1;
        for (mwIndex pti = 0; pti < N; ++pti) {
            if ( *fx < 0 || *fy < 0 || *fx >= imW || *fy >= imH )
                throw runtime_error("Some interest points are outside the image boundaries.");
            const id_t &id = idmap2[ *fx * imH + *fy ]; // coordinates in base 0, "id" in base 1.
            if ( id > 0 ) // foreground
                ++( c[ id-1 ] );
            fx += 2;
            fy += 2;
        }
    }
    return ptcounts;

}

/** 
 * Compare box cooodinates and points coordinates within the current scale
 *
 * sci          Feature scale index
 * r            Box index, base 0.
 * tmpm         Pre-allocated mask of length allN[sci]
 *
 * Returns sum(tmpm).
 */
template <class Tc>
npts_t PoolObjDesc<Tc>::getptmask4box(const mwSize sci,const mwSize r,bool *tmpm) {

    npts_t obj_npts = 0;
    const Tc * obj_bb = bbs.get() + 4 * r;
    const mwSize &N = allN[sci];
    memset(tmpm,0,sizeof(bool) * N);
    const Tc *f_xy = allLocs[sci].get();

    if (masktype==MaskNone) {

        for (mwIndex pti = 0; pti < N; ++pti) {
            const Tc *f_xy_pti = f_xy + 2 * pti;
            const Tc &fx = f_xy_pti[0];
            const Tc &fy = f_xy_pti[1];
            if ( (fx >= obj_bb[0]) && (fy >= obj_bb[1]) && (fx <= obj_bb[2]) && (fy <= obj_bb[3]) ) {
                tmpm[pti] = true;
                ++obj_npts;
            }
        }

    } else if (masktype==MaskBool) {

        const Tc obj_height = obj_bb[3]-obj_bb[1]+1; // 0:x1, 1:y1, 2:x2, 3:y2

        // per-box segmentation mask is defined
        if (!masks_bool)
            throw runtime_error("Bug: MaskBool but masks_bool is unset");
        const bool *mask = masks_bool->at( r ).get();
        for (mwIndex pti = 0; pti < N; ++pti) {
            const Tc *f_xy_pti = f_xy + 2 * pti;
            const Tc &fx = f_xy_pti[0];
            const Tc &fy = f_xy_pti[1];
            if ( (fx >= obj_bb[0]) && (fy >= obj_bb[1]) && (fx <= obj_bb[2]) && (fy <= obj_bb[3]) ) {

                if ( mask[ ( (size_t) (fx - obj_bb[0])) * obj_height + (fy - obj_bb[1]) ] ) { // should be fine both for base is irrelevant.
                    tmpm[pti] = true;
                    ++obj_npts;
                }

            }
        }

    } else if ( masktype == MaskBackground ) {

        for (mwIndex pti = 0; pti < N; ++pti) {
            const Tc *f_xy_pti = f_xy + 2 * pti;
            const Tc &fx = f_xy_pti[0];
            const Tc &fy = f_xy_pti[1];
            if ( (fx < obj_bb[0]) || (fy < obj_bb[1]) || (fx > obj_bb[2]) || (fy > obj_bb[3]) ) {
                tmpm[pti] = true;
                ++obj_npts;
            }
        }


    } else {

        throw runtime_error("unrecognized masktype for getptmask4box with bool");

    }

    // printf("sci=%d npts=%d\n",(int) sci+1,obj_npts);

    return obj_npts;

}


/**
 * Just like the bool* counter-part, but, for masktype==MaskFloat
 */
template <class Tc>
double PoolObjDesc<Tc>::getptmask4box(const mwSize sci,const mwSize r,float *tmpm) {

    double obj_npts = 0;
    const Tc * obj_bb = bbs.get() + 4 * r;
    const mwSize &N = allN[sci];
    memset(tmpm,0,sizeof(float) * N);
    const Tc *f_xy = allLocs[sci].get();
    const Tc obj_height = obj_bb[3]-obj_bb[1]+1; // 0:x1, 1:y1, 2:x2, 3:y2

    if (masktype==MaskFloat) {
        // per-box segmentation weights is defined
        const float *mask = masks_float->at( r ).get();
        for (mwIndex pti = 0; pti < N; ++pti) {
            const Tc &fx = *(f_xy++);
            const Tc &fy = *(f_xy++);
            if ( (fx >= obj_bb[0]) && (fy >= obj_bb[1]) && (fx <= obj_bb[2]) && (fy <= obj_bb[3]) ) {

                const float &w = mask[ ( (size_t) (fx - obj_bb[0])) * obj_height + (fy - obj_bb[1]) ];
                if (w > FvGmm_PoolObjDesc_PtWeightEps) { 
                    tmpm[pti] = w;
                    obj_npts += w;
                }

            }
        }
    } else if (masktype==MaskImage) {
        // get segmentation mask from the full image-mask
        const int16_t maski = maskind.get()[r];
        const float * mask = maskim.get() + (maskim_H*maskim_W) * maski;
        for (mwIndex pti = 0; pti < N; ++pti) {
            const Tc &fx = *(f_xy++);
            const Tc &fy = *(f_xy++);
            if ( (fx >= obj_bb[0]) && (fy >= obj_bb[1]) && (fx <= obj_bb[2]) && (fy <= obj_bb[3]) ) {
                const float w = mask[ (size_t) fx * maskim_H + fy ];
                if (w > FvGmm_PoolObjDesc_PtWeightEps) { 
                    tmpm[pti] = w;
                    obj_npts += w;
                }
            }
        }
 
    } else {
        throw runtime_error("unrecognized masktype for getptmask4box with float");
    }

    return obj_npts;

}

/** 
 * Compute a binary mask for point lying inside the current segment.
 * Uses segment2pts and segment2pts_index.
 * 
 * sci          Feature scale index
 * r            Region index = (Region id - 1).
 * tmpm         Pre-allocated mask of length allN[sci]
 *
 * Returns sum(tmpm).
 */
template <class Tc>
npts_t PoolObjDesc<Tc>::getptmask4segment(const mwSize nscale,const mwSize sci,const mwSize r,bool *tmpm) {

    const mwSize &N = allN[sci];
    memset(tmpm,0,N);

    mwIndex ii = nscale * r + sci; // segment2pts_index: (nscale R)

    mwSize ibeg = segment2pts_index.get()[ii];
    mwSize iend = segment2pts_index.get()[ii + 1];

    mwSize *pti = segment2pts.get() + ibeg;
    mwSize *end = segment2pts.get() + iend;
    while(pti < end) {
        tmpm[ *(pti++) ] = true;
    }
    return (npts_t) (iend-ibeg);

}

/**
 * Return total number of points within a particular segment. 
 * This is a very fast operation, utilizing the index structure.
 */
/* -- not used, not tested --
 * npts_t getsegmenttotalnpts(const mwSize r) {
 mwIndex ii = nscale * r; // segment2pts_index: (nscale R)
 return segment2pts_index.get()[ii + nscale] - segment2pts_index.get()[ii];
 }*/


/// SEE accumFVsToParents()
template <class Tc>
void PoolObjDesc<Tc>::accumFVsToParentsHelper( double *out, npts_t *npts, const SingleTree <id_t> &tree, id_t root, mwSize fvlen ) {
    id_t nchild = tree.getNumChildren(root);
    id_t *childs = tree.getChildren(root);

    // first make sure all children has their final descriptors
    for ( id_t i = 0; i < nchild; ++i ) {
        accumFVsToParentsHelper( out, npts, tree, childs[i], fvlen );
    }

    // now sum their descriptors
    double *root_out = out + fvlen * (root-1);
    npts_t &root_npts = npts[ root-1 ];
    for ( mwIndex i = 0; i < nchild; ++i ) {
        utils::addToVector( out + fvlen * (childs[i]-1), root_out, fvlen );
        root_npts += npts[ childs[i]-1 ];
    }

}


/// max of a std::vector
template <class Tc>
mwSize PoolObjDesc<Tc>::max( const vector<mwSize> &vec ) {
    if (vec.size()==0)
        throw runtime_error("max::vec cannot be empty");
    mwSize res = vec[0];
    for (mwIndex i = 1; i < vec.size(); ++i) {
        if ( vec[i] > res )
            res = vec[i];
    }
    return res;
}

/// Find region bounding boxes in base 0 and save them to this->bbs. Requires R.
template <class Tc>
void PoolObjDesc<Tc>::idmap2bbs(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) {

    unsetmasks();
    Tc *bbs = new Tc[ 4 * R ](); // zero initialized
    this->bbs.reset(  bbs, DeleterArray<Tc> ); 

    const id_t *im = idmap.get();

    {
        // initialize
        Tc *bbs2 = bbs;
        for (mwIndex r = 0; r < R; ++r) {
            bbs2[0] = imW-1;
            bbs2[1] = imH-1;
            bbs2 += 4;
        }
    }

    // column major ordering
    for (int16_t j = 0 ; j < imW; ++j ) {
        for (int16_t i = 0 ; i < imH; ++i ) {
            const id_t &r1 = *im;
            if (r1 > 0) {

                Tc * curbb = bbs + 4 * (r1-1) ;
                if ( curbb[0] > j )
                    curbb[0] = j;

                if ( curbb[1] > i )
                    curbb[1] = i;

                if ( curbb[2] < j )
                    curbb[2] = j;

                if ( curbb[3] < i )
                    curbb[3] = i;

            }
            ++im;
        }
    }

    // utils::printMatrix( 4, R, bbs, "bbs" );
}


/**
 * Pool elements towards upper nodes in the tree.
 * This should be called after setsegmentation().
 *
 * Parents denotes the parent of each node where node indices have base 1.
 * Root node has parent=0. Root is required to be unique.
 *
 * In the "out" and "npts" arrays, leaf node descriptors are already computed with desctrans=none
 * Other nodes descriptors should be set to zeros before this function is called.
 */
template <class Tc>
void PoolObjDesc<Tc>::accumFVsToParents( double *out, npts_t *npts, const SingleTree <id_t> &tree ) {

    // TT nleaf = tree.getNumLeaf(); // this may not be R!

    // accumulate recursively
    mwSize fvlen = (1+2*D)*K;
    accumFVsToParentsHelper( out, npts, tree, tree.getRoot(), fvlen );

}

/** 
 * Pool fishervectors according to a segment ids image.
 * Algorithm:
 * 1) First pass to find #points within each segment.
 * 2) Second pass to find points within each segment.
 * 3) Pool them.
 *
 * (1) and (2) is already done in setsegmentation()
 *
 * out should have been pre-initialized to zeros.
 */
template <class Tc>
void PoolObjDesc<Tc>::pooltreefv_leaves(double *out,npts_t *npts,bool avgfv,bool skipifnonexist) {
#define POOL_TREE
#include "poolobjdesc_poolfv.cpp"
} 


/**
 * Pool Spatial Fisher Vectors over a segmentation image.
 * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
 * out should have been pre-initialized to zeros.
 */
template <class Tc>
void PoolObjDesc<Tc>::pooltreespatfv_leaves(double *out,npts_t *npts,bool avgfv,bool skipifnonexist) {
#define POOL_TREE
#include "poolobjdesc_poolspatfv.cpp"
} 

/**
 * Set segmentation map. 
 * R = max( idmap(:) ). idmap==0 regions denote "background".
 * Sets this->bbs to segment bounding boxes.
 */
template <class Tc>
void PoolObjDesc<Tc>::setsegmentation(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) {
    this->R     = cn::maxofvector( idmap.get(), imH*imW );

    // clear old
    segment2pts_index.reset();
    segment2pts.reset();
    segmentexists.reset();

    mwSize nscale = getnscale();
    if ( (R * nscale) == 0 ) 
        return; // nothing to do.

    setSegmentExists(idmap,imH,imW);
    idmap2bbs(idmap,imH,imW); // region boxes needed to do location normalization and also for processing upper nodes

    shared_ptr<mwSize> ptcounts = countPtsPerRegionPerScale(idmap,imH,imW); // (R nscale)
    const mwSize totalNptsInRegions = make_segment2pts_index( ptcounts );

    // Second pass to store a list of points for each scale and each segment.
    shared_ptr< mwSize > curcounts_ptr( new mwSize[R], DeleterArray<mwSize> ); // single-scale counter.
    this->segment2pts.reset( new mwSize[totalNptsInRegions], DeleterArray<mwSize> );
    mwSize *segment2pts = this->segment2pts.get();
    for (mwSize sci = 0; sci < nscale; ++sci) {

        // store number of points that have been added to the index vectors within the current scale.
        mwSize *curcounts = curcounts_ptr.get();
        memset( curcounts, 0, sizeof(mwSize)*R );

        mwSize *ind = segment2pts_index.get() + sci; // index(sci,j) = ind + nscale * j 
        const mwSize &N = allN[sci];
        const Tc *fx = allLocs[sci].get(); 
        const Tc *fy = fx + 1; // base 0
        const id_t *idmap2 = idmap.get();
        for (mwIndex pti = 0; pti < N; ++pti) {
            const id_t &id = idmap2[ *fx * imH + *fy ]; // base 0
            if ( id > 0 ) {
                mwSize &c = curcounts[id-1];
                segment2pts[ *( ind + nscale*(id-1) ) + c ] = pti;
                ++c;
            }
            fx += 2;
            fy += 2;
        }
    }
}

/**
 * Given the number of points within each region, craete "segment2pts_index" array.
 * Note that segment2pts_index does not contain the actual segment->pts index.
 * Instead, it just stores pointers to acces segment2pts array for each region.
 *
 * Returns total number of points in all regions, ie. sum(ptcounts)
 */
template <class Tc>
mwSize PoolObjDesc<Tc>::make_segment2pts_index( const shared_ptr<mwSize> ptcounts ) {

    mwSize nscale = getnscale();
    mwSize totalNptsInRegions = 0;
    segment2pts_index.reset( new mwSize [nscale*R + 1], DeleterArray<mwSize> ); // (nscale R)

    // Build the index (nscale R), in the order useful during pooling.
    //
    // segment2pts will store in the following order 
    //      { {sc=0,id=1} ... {sc=nsc-1,id=1} ...  {sc=0,id=R} ... {sc=nsc-1,id=R} }
    //
    // For any sci and i, segment2pts_index[R*j+sci+1]-segment2pts_index[R*j+sci] will give 
    // number of points at the corresponding scale and region.
    mwSize * ind = segment2pts_index.get();
    *(ind++) = 0;
    for (mwSize r = 0; r < R; ++r) {
        const mwSize * c = ptcounts.get() + r;
        for (mwSize sci = 0; sci < nscale; ++sci) {
            totalNptsInRegions += *c;
            *(ind++) = totalNptsInRegions;
            c += R; // next scale for the current region
        }
    }

    return totalNptsInRegions;

}


/**
 * Build up "segmentexists" vector.
 * Note that segmentexists[r] refers to idmap==(r+1) where r is the segment index.
 * ie. r is base 0 in segmentexists and base 1 in idmap since for "idmap", r==0 represents background.
 * Requires "R" in accordance with idmap.
 */
template <class Tc>
void PoolObjDesc<Tc>::setSegmentExists(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) {

    this->segmentexists.reset( new bool[R](), DeleterArray<bool> );

    const id_t *im = idmap.get();
    size_t npixel = ( (size_t) imH ) * ( (size_t) imW );
    bool *out = segmentexists.get();
    for ( size_t i = 0; i < npixel; ++i ) {
        if ( *im > 0 )
            out[ *im - 1 ] = true;
        ++im;
    }


}


