
/**
 * Author: R. Gokberk Cinbis
 */



// selind = cn_nmsgreedy_mex(bbs,areas,normopt,pixelcoord,scores,sort2org,maxov,minscoreth,maxnumdet);
//
// sort2org:    Ordering of entries in decreasing scores.
// normopt:     Defines bboverlap measure to use for non-maxima suppression.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "mex.h"
#include "cn_bboverlap_functions.cpp"
#include "cn_matlabarrays.hpp"
#include "cn_math.hpp"

enum { a_bbs=0, a_areas, a_normopt, a_pixelcoord, a_scores, a_sort2org, a_maxov, a_minscoreth, a_maxnumdet };

void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{

    cn::mex::assert2(nrhs==9,"incorrect number of input arguments");
    cn::mex::assert2(nlhs==1,"incorrect number of output arguments");

    // constants
    mwSize N = cn::mex::getdim(prhs[a_bbs],2); 
    
    // size checks
    cn::mex::assert2(cn::mex::getdim(prhs[a_bbs],1)==4,"size(bbs,1)!=4");
    cn::mex::assert2(mxGetNumberOfElements(prhs[a_areas])==N,"numel(areas)!=N");
    cn::mex::assert2(mxGetNumberOfElements(prhs[a_sort2org])==N,"numel(sort2org)!=N");

    // get values
    const double *bbs = cn::mex::getDenseData<double>(prhs[a_bbs]); // (4 N)
    const double *areas = cn::mex::getDenseData<double>(prhs[a_areas]);
    const NormType normopt = (NormType) (int) cn::mex::getscalar(prhs[a_normopt]);
    const bool pixelcoord = (bool) cn::mex::getscalar(prhs[a_pixelcoord]);
    const double *scores = cn::mex::getDenseData<double>(prhs[a_scores]);
    const double *sort2org = cn::mex::getDenseData<double>(prhs[a_sort2org]);
    const double maxov = cn::mex::getscalar(prhs[a_maxov]);
    double minscoreth = cn::mex::getscalar(prhs[a_minscoreth]); // might be -inf
    const mwSize maxnumdet = (mwSize) MIN(N,cn::mex::getscalar(prhs[a_maxnumdet])); // might be inf. cast after MIN.

    // if minscoreth=-inf simply set minscoreth=minimum_finite_value so that 
    //  s<minscoreth is true only if s==-inf
    const double NegInf = -cn::infconst<double>();
    if (minscoreth==NegInf) {
        minscoreth = -std::numeric_limits<double>::max(); // not min()!
    }

    // degenerate case
    if (N==0) {
        plhs[0] = mxCreateDoubleMatrix(0,1,mxREAL);
        return;
    }

    // alloc
    bool *issupp = (bool *) cn::mex::matCalloc(sizeof(bool),N); // will be set to true for suppressed values
    mwSignedIndex *selind = (mwSignedIndex *) cn::mex::matMalloc(sizeof(mwSize)*maxnumdet); 
    mwSize nsel = 0;
    bool iserr = false;

    // main loop
    if (maxnumdet > 0) {
        for (mwIndex j = 0; j < N; ++j) {
            mwSignedIndex maxi = ((mwSignedIndex) sort2org[j]) -1; // in C coordinates

            if (issupp[maxi]) {
                continue; // suppressed during iterations
            }

            double s = scores[maxi];
            if ( s<minscoreth ) {
                // has reached lower-than-threshold OR -inf detections. just quit.
                break;
            } 

            selind[nsel] = maxi+1; // in matlab coordinates
            ++nsel;

            if (nsel >= maxnumdet) {
                break;
            }

            // suppress nearby detections
            issupp[maxi] = true;  // for small speedup for later loops
            const double *curbb = bbs + 4 * maxi;
            double curarea = areas[maxi];
            for (mwIndex i = 0; i < N; ++i) {
                if (issupp[i])
                    continue; // already suppressed. no need to check bboverlap.
                
                if ( getbboverlap(curbb,(bbs+4*i),curarea,areas[i],pixelcoord,normopt,&iserr) > maxov )
                    issupp[i] = true;
            }

        }
    }

    // set final output: selind2 = selind(1:nsel)
    plhs[0] = mxCreateDoubleMatrix(nsel,1,mxREAL);
    double *selind2 = cn::mex::getDenseData<double>(plhs[0]);
    mwSignedIndex *xi = selind;
    for (mwIndex i = 0; i < nsel; ++i) {
        *selind2 = (double) *xi;
        xi++;
        selind2++;
    }

    // dealloc
    mxFree(selind);
    mxFree(issupp);

}

