
/**
 * Author: R. Gokberk Cinbis
 */



// cn_fvgmm_poolobjdesc_mex.cpp

#include "mex.h"
#include "cn_matlabarrays.hpp"
#include "cn_fvgmm.hpp"
using cn::fvgmm::PoolObjDesc;
using cn::DeleterNone;
using cn::mex::getstring;

typedef PoolObjDesc<int16_t> Pooler;

const mxArray *getfield(const mxArray *prhs[],const char *fieldname) {
    return cn::mex::getfield(prhs[0],fieldname);
}

bool hasfield(const mxArray *prhs[],const char *fieldname) {
    return cn::mex::isfield(prhs[0],fieldname);
}

/** 
 * Add locations.
 * Check that nscale > 0 and not too high.
 * Check that number of points it not too high.
 */
void addLocs(Pooler &p,const mxArray *prhs[]) {

    const mxArray * allLocs =  getfield(prhs,"f_xy"); // {(2 N_s)} feature locations, x,y. [int16]
    if (!mxIsCell(allLocs))
        throw runtime_error("f_xy should be a cell array");
    mwSize nscale = mxGetNumberOfElements(allLocs);

    if (nscale == 0)
        mexErrMsgTxt("nscale>0 is required");
    if (nscale > 32767) // check int16
        mexErrMsgTxt("nscale is too high!"); 


    mwSize totalN = 0;
    for (mwIndex sci = 0; sci < nscale; ++sci ) {
        const mxArray *f_xy = mxGetCell( allLocs, sci );
        if (mxIsSparse(f_xy)) { mexErrMsgTxt("f_xy should be dense"); }
        if (mxGetM(f_xy)!=2) { mexErrMsgTxt("size(f_xy,1) should be 2"); }
        mwSize N = mxGetN(f_xy);
        totalN += N;
        p.pushLoc(N, shared_ptr<const int16_t>( cn::mex::getDenseData<int16_t>(f_xy,"f_xy"), DeleterNone<const int16_t> ) );
    }

}

/// add posteriors of size (N K)
void addPosteriors(Pooler &p,const mxArray *prhs[]) {

    const mxArray * allQ =  getfield(prhs,"posts"); // {(2 N_s)} feature locations, x,y. [int16]
    if (!mxIsCell(allQ))
        throw runtime_error("posts should be a cell array");
    mwSize nscale = mxGetNumberOfElements(allQ);

    for (mwIndex sci = 0; sci < nscale; ++sci) {
        // get Q, set allN and K
        const mxArray *Q = mxGetCell( allQ, sci );
        if ( !mxIsSparse(Q) ) { mexErrMsgTxt("Q should be sparse"); }
        if ( mxGetClassID(Q) != mxDOUBLE_CLASS ) { mexErrMsgTxt("Q should be double"); }

        mwSize N = mxGetM(Q);
        mwSize K = mxGetN(Q); 
        p.pushPosteriors( N, K, mxGetPr(Q),mxGetIr(Q),mxGetJc(Q) );
    }

}

/// add descriptors of size (D N)
void addDescriptors(Pooler &p,const mxArray *prhs[]) {

    const mxArray * allV =  getfield(prhs,"d_cols"); // {(2 N_s)} feature locations, x,y. [int16]
    if (!mxIsCell(allV))
        throw runtime_error("d_cols should be a cell array");
    mwSize nscale = mxGetNumberOfElements(allV);
    for (mwIndex sci = 0; sci < nscale; ++sci) {
        const mxArray *V = mxGetCell( allV, sci );
        mwSize D = mxGetM(V); 
        mwSize N = mxGetN(V);
        p.pushDesc(D,N,cn::mex::getDenseData<double>(V,"V"));
    }
}

// set dictionary parameters. required D,K to have been setup
void setdictionary(Pooler &p,const mxArray *prhs[]) {
    const mxArray *mix   = getfield(prhs,"mix");
    const mxArray *means = getfield(prhs,"means");
    const mxArray *vars  = getfield(prhs,"vars");

    mwSize D = mxGetM(means);
    mwSize K = mxGetN(means);
    if (mxGetNumberOfElements(mix)!=K)
        mexErrMsgTxt("numel(mix)!=K");
    if (mxGetM(vars)!=D)
        mexErrMsgTxt("numel(vars)!=D");
    if (mxGetN(vars)!=K)
        mexErrMsgTxt("numel(vars)!=K");

    p.setdictionary( D, K, cn::mex::getDenseData<double>(mix), cn::mex::getDenseData<double>(means), cn::mex::getDenseData<double>(vars) );
}


// note that this sets MatlabParams, not Pooler!
// Important: Do avgfv in all cases, except when desctrans=none.
bool doAvgFV(const mxArray *prhs[]) {
    if (getstring(getfield(prhs,"desctrans"))=="none")
        return false;
    else
        return true;
}


/// helper for setmasks
void setmasks_bool(Pooler &p,const mxArray *masks_arr,const int16_t *bbs,const mwSize R) {
    // we'll store in a std::vector
    shared_ptr< vector < shared_ptr<const bool> > > masks( new vector< shared_ptr<const bool> >() );
    masks->reserve(R);

    const int16_t *bbs2 = bbs;
    for (mwIndex r = 0; r < R; ++r) {
        const mxArray * m = mxGetCell(masks_arr,r);

        // check the size of the mask
        if ( (mxGetM(m)!=((mwSize) (bbs2[3]-bbs2[1]+1))) || (mxGetN(m)!=((mwSize) (bbs2[2]-bbs2[0]+1))) )
            throw runtime_error("at least one of the masks have size incompatible with its box");

        shared_ptr <const bool> ptr ( cn::mex::getDenseData<bool>( m, "masks[r] (binary mask)" ), DeleterNone<const bool> );
        masks->push_back( ptr );
        bbs2 += 4;
    }
    p.setmasks( masks );
}

/// helper for setmasks
void setmasks_image(Pooler &p,const mxArray *masks_arr,const mxArray *maskind_arr,const mwSize R) {

    // get data
    shared_ptr <const float> maskim( cn::mex::getDenseData<float>(masks_arr), DeleterNone<const float> );
    shared_ptr <const int16_t> maskind( cn::mex::getDenseData<int16_t>(maskind_arr), DeleterNone<const int16_t> );

    // get mask image size
    mwSize height = cn::mex::getdim(masks_arr,1);
    mwSize width  = cn::mex::getdim(masks_arr,2);
    mwSize nmask  = cn::mex::getdim(masks_arr,3);

    // make sure maskind are in range [0,nmask)
    if ( mxGetNumberOfElements(maskind_arr) != R )
        throw runtime_error("masind has invalid length");

    // if boxes are out-of-the-bounds, that's OK. we'll just clip them.
    p.setmasks( maskim, height, width, nmask, maskind );
  
}


/// helper for setmasks
void setmasks_weighted(Pooler &p,const mxArray *masks_arr,const int16_t *bbs,const mwSize R) {
    // we'll store in a std::vector
    shared_ptr< vector < shared_ptr<const float> > > masks( new vector< shared_ptr<const float> >() );
    masks->reserve(R);

    const int16_t *bbs2 = bbs;
    for (mwIndex r = 0; r < R; ++r) {
        const mxArray * m = mxGetCell(masks_arr,r);

        // check the size of the mask
        if ( (mxGetM(m)!=((mwSize) (bbs2[3]-bbs2[1]+1))) || (mxGetN(m)!=((mwSize) (bbs2[2]-bbs2[0]+1))) )
            throw runtime_error("at least on of the masks have size incompatible with its box");

        shared_ptr <const float> ptr ( cn::mex::getDenseData<float>( m, "masks[r] (weighted mask)" ), DeleterNone<const float> );
        masks->push_back( ptr );
        bbs2 += 4;
    }
    p.setmasks( masks );


}

/// helper for setboxesandscales()
void setmasks(Pooler &p,const mxArray *prhs[],const int16_t *bbs,const mwSize R) {

    const mxArray * masks_arr = getfield(prhs,"masks"); // (1 R) 

    if (cn::mex::isstring(masks_arr)) {

        string s = getstring(masks_arr);
        if ( s == "background" ) {
            p.setmasks( Pooler::MaskBackground );
        } else {
            throw runtime_error("Unrecognized masks option");
        }

    } else if ( mxIsCell(masks_arr) ) {

        if (mxGetNumberOfElements(masks_arr)!=R)
            throw runtime_error("cn_fvgmm_poolobjdesc::p.masks:: if a cell array, should be of length R");

        if ( mxIsLogical(  mxGetCell(masks_arr,0)  ) ) {
            // binary masks
            setmasks_bool(p,masks_arr,bbs,R);
        } else {
            // weighted masks
            setmasks_weighted(p,masks_arr,bbs,R);
        }

    } else if ( mxIsNumeric(masks_arr) ) {

        if ( !hasfield(prhs,"maskind") ) {
            throw runtime_error("If 'masks' is provided as an image, 'maskind' field is also required.");
        }
        const mxArray * maskind_arr = getfield(prhs,"maskind"); // (1 R) 
        setmasks_image(p,masks_arr,maskind_arr,R);

    } else { 

        throw runtime_error("Unknown masks option");

    }

}

void setboxesandscales(Pooler &p,const mxArray *prhs[]) {

    const mxArray * bbs_arr = getfield(prhs,"bbs"); // (4 R) each row is x1y1x2y2 box coordinates. [int16]

    // boxes
    const mwSize R = mxGetN( bbs_arr );
    if ( mxGetM(bbs_arr)!=4 ) 
        mexErrMsgTxt("bbs should be (4 R)");
    const int16_t *bbs = cn::mex::getDenseData<int16_t>( bbs_arr, "bbs" );
    p.setboxes( shared_ptr<const int16_t>( bbs, DeleterNone<int16_t> ), R );

    // scales (optional)
    if ( hasfield(prhs,"minscs") || hasfield(prhs,"maxscs") || hasfield(prhs,"scstep") ) {
        const mxArray * minscs = getfield(prhs,"minscs"); // (1 R) minimum scale index in range 0 <= minsc < nscale. [int16]
        const mxArray * maxscs = getfield(prhs,"maxscs"); // (1 R) maximum scale index in range 0 <= maxsc < nscale. [int16];
        const mxArray * scstep = getfield(prhs,"scstep"); // (1 1) scale step when building descriptors. ;
        if (mxGetNumberOfElements(minscs)!=R || mxGetNumberOfElements(maxscs)!=R)
            throw runtime_error("minscs,maxscs should be (1 R) or (R 1)");
        p.setscales( cn::mex::getDenseData<int16_t>(minscs,"minscs"),  cn::mex::getDenseData<int16_t>(maxscs,"maxscs"), (int16_t) cn::mex::getscalar(scstep) );
    } else {
        p.unsetscales();
    }

    // masks
    if ( hasfield(prhs,"masks") && R > 0 ) {
        setmasks(p,prhs,bbs,R);
    }

}

// process a single segmentation tree
void processtree(Pooler &p,const mxArray *trees,mwIndex ti,mxArray *plhs[],bool spatfv) {

    // tree info
    mwSize nblob = (mwSize) cn::mex::getscalar( cn::mex::getfield( trees, ti, "nblob" ) );

    // leaf ids image
    const mxArray *idmap_arr = cn::mex::getfield(trees,ti,"leafmap");
    if (mxGetNumberOfDimensions(idmap_arr) > 2) {
        throw runtime_error("idmap_arr should be HxW");
    }
    shared_ptr<const uint32_t> idmap( cn::mex::getDenseData<uint32_t>(idmap_arr,"idmap"), DeleterNone<const uint32_t> );
    const int16_t imH = (int16_t) mxGetM(idmap_arr);
    const int16_t imW = (int16_t) mxGetN(idmap_arr);

    // allocate current output cells. allocate output sufficient to hold "nblob" fisher vectors. we'll only set
    // the first nleaf nodes in it.
    mxArray *cur_out_arr;
    mxArray *cur_npts_arr;
    if (!spatfv) {
        cur_out_arr = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1+2*p.getDapp(),p.getKapp(),nblob);
        cur_npts_arr = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1,nblob);
    } else {
        cur_out_arr = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1+2*2,p.getKdict()*p.getKapp(),nblob);
        cur_npts_arr = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1,nblob);
    }
    mxSetCell( plhs[0], ti, cur_out_arr );
    mxSetCell( plhs[1], ti, cur_npts_arr );

    // get tree structure
    const mxArray *parents_arr = cn::mex::getfield(trees,ti,"parents");
    if (mxGetNumberOfElements(parents_arr)!=nblob)
        throw runtime_error("length(parents_arr) is incorrect");
    shared_ptr<const uint32_t>parents( cn::mex::getDenseData<uint32_t>(parents_arr), DeleterNone<uint32_t> );
    cn::SingleTree <uint32_t> tree( parents, (uint32_t) nblob );

    // process leaf nodes
    double *out = cn::mex::getDenseData<double>(cur_out_arr,"out");
    double *npts = cn::mex::getDenseData<double>(cur_npts_arr,"npts");

    if (!spatfv) {
        p.pooltreefv( out, npts,idmap, imH, imW, tree ); // desctrans=none
    } else {
        p.pooltreespatfv( out, npts,idmap, imH, imW, tree ); // desctrans=none
    }

}

void mexFunction2(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

    if (nrhs != 1)
        mexErrMsgTxt("Invalid number of input arguments");
    if (nlhs > 2)
        mexErrMsgTxt("Too many output args");

    string op = getstring(getfield(prhs,"op")); // operation
    Pooler p;

    // alloc output and process
    // we should init to zeros in case there is no point to contribute to a particular dimension.
    // (plus, we assume in the code that it is zero initialized)
    if (op=="poolboxes") {
        addLocs(p,prhs);
        addPosteriors(p,prhs);
        addDescriptors(p,prhs);
        p.finalizePatchInfo();
        setboxesandscales(p,prhs);

        plhs[0] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,p.getDapp(),p.getKapp(),p.getnumregions());
        double * out = cn::mex::getDenseData<double>(plhs[0],"out");
        plhs[1] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1,p.getnumregions());
        double *npts = cn::mex::getDenseData<double>(plhs[1],"npts");
        p.poolboxdesc( out, npts );
    } else if ( op == "fvboxes" ) {
        addLocs(p,prhs);
        addPosteriors(p,prhs);
        addDescriptors(p,prhs);
        p.finalizePatchInfo();
        setdictionary(p,prhs);
        setboxesandscales(p,prhs);

        plhs[0] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1+2*p.getDapp(),p.getKapp(),p.getnumregions());
        double *out = cn::mex::getDenseData<double>(plhs[0],"out");
        plhs[1] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1,p.getnumregions());
        double *npts = cn::mex::getDenseData<double>(plhs[1],"npts");
        p.poolboxfv( out, npts, doAvgFV(prhs) );

    } else if (op == "fvtrees") {
        addLocs(p,prhs);
        addPosteriors(p,prhs);
        addDescriptors(p,prhs);
        p.finalizePatchInfo();
        setdictionary(p,prhs);

        if (doAvgFV(prhs)) // in the mex file we require 'none' since it is necessary to be able to sum towards higher trees
            throw runtime_error("for fvtrees in mex implementation, desctrans=none is required");

        // get trees
        const mxArray *trees = getfield(prhs,"trees");
        const mwSize ntree = mxGetNumberOfElements( trees );

        if ( hasfield(prhs,"minscs") || hasfield(prhs,"maxscs") || hasfield(prhs,"scstep") )
            throw runtime_error("fvtrees does not utilize scales right now");

        // set output
        plhs[0] = cn::mex::createCellArray(1,ntree);
        plhs[1] = cn::mex::createCellArray(1,ntree);
        for ( mwIndex ti = 0; ti < ntree; ++ti ) {
            processtree( p, trees, ti, plhs, false );
        }

    } else if (op == "spatfvboxes") {
        addLocs(p,prhs);
        addPosteriors(p,prhs);
        // not needed: addDescriptors(p,prhs);
        p.finalizePatchInfo();
        setdictionary(p,prhs);
        setboxesandscales(p,prhs);

        plhs[0] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1+2*2,p.getKapp()*p.getKdict(),p.getnumregions());
        double *out = cn::mex::getDenseData<double>(plhs[0],"out");
        plhs[1] = cn::mex::createArrayZeros(mxDOUBLE_CLASS,1,p.getnumregions());
        double *npts = cn::mex::getDenseData<double>(plhs[1],"npts");
        p.poolboxspatfv( out, npts, doAvgFV(prhs) );


    } else if (op == "spatfvtrees") {
        addLocs(p,prhs);
        addPosteriors(p,prhs);
        // dont: addDescriptors(p,prhs);
        p.finalizePatchInfo();
        setdictionary(p,prhs);

        if (doAvgFV(prhs)) // in the mex file we require 'none' since it is necessary to be able to sum towards higher trees
            throw runtime_error("for fvtrees in mex implementation, desctrans=none is required");

        // get trees
        const mxArray *trees = getfield(prhs,"trees");
        const mwSize ntree = mxGetNumberOfElements( trees );

        if ( hasfield(prhs,"minscs") || hasfield(prhs,"maxscs") || hasfield(prhs,"scstep") )
            throw runtime_error("fvtrees does not utilize scales right now");

        if ( hasfield(prhs,"bbs") ) // avoid bugs in the long term
            throw runtime_error("do not provide bbs for the trees (to avoid confusion)"); 

        // set output
        plhs[0] = cn::mex::createCellArray(1,ntree);
        plhs[1] = cn::mex::createCellArray(1,ntree);
        for ( mwIndex ti = 0; ti < ntree; ++ti ) {
            processtree( p, trees, ti, plhs, true );
        }

    } else {
        mexErrMsgTxt("bug: Unknown p.op constant");
    }
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

    try {
        mexFunction2(nlhs,plhs,nrhs,prhs);
    } catch (runtime_error &e) {
        mexErrMsgTxt( e.what() );
    }

}

