function imgs = cn_impyramid(im,method,varargin)
% imgs = cn_impyramid(im,method,varargin)
%
% Construct a multiscale image pyramid.
%
% USAGE
% imgs = cn_impyramid(im,'cvCubic/cvArea',scales)           *** Suggested Method ***
%
%   Build image pyramid using cn_imresize.
%   scales is a vector of image scales. eg, scales=(1/1.2).^(-1:23).
%
%   ALGORITHM
%   Each downscaling is done from the full image using cvArea.
%   Exceptionally, when upsampling, uses cvCubic of cn_imresize.
%
%   OUTPUT
%   imgs{i}         cn_imresize(im,ceil([size(im,1) size(im,2)]*scales(i)))
% 
% 
% imgs = cn_impyramid(im,'cvCubic/cvArea+cvGaussPyr',scales,pybias)
%
%   Build image pyramid just like method='cvArea'. The only difference is that, instead of down-scaling from the
%   full image, uses a close level from the gaussian pyramid to save computation.
%   However, note that downscaling from the previous level can cause over-smoothing and other artifacts and this
%   approach does little speed up in most cases.
%   Unless image is huge, 'cvCubic/cvArea' should be chosen, as it is already very fast.
%
%   ALGORITHM
%   If pybias=0 (not good), closest pyramid level with larger scale is used.
%   If pybias=-1 (recommended), closest second pyramid level with larger scale is used.
%   and so on for any pybias value.
%
%
% imgs = cn_impyramid(im,'cvGaussPyr',nlevel)
%
%   Build Gaussian image pyramid using openCV's implementation (cv::buildPyramid).
%   todo: Why does it look blurred?
%
%   OUTPUT
%   imgs{1}         Original image itself.
%   imgs{nlevel}    nlevel-1 times downsampled image. At each new level, image has new_size=ceil(prev_size/2).
%                   If nlevel is too large, images at smallest scales will be repetitions.
%
%
% imgs = cn_impyramid(im,'imresize',scales)
%
%   Build image pyramid just like method='cvArea' but uses Matlab's image toolbox's 
%   imresize() with its default options.
%
% R.G.Cinbis, Sep 2012

switch(method)
    case 'cvGaussPyr'
        assert(length(varargin)==1,'incorrect number of arguments');
        nlevel = varargin{1};
        imgs = [{im} cn_impyramid_cvgaussian(im,nlevel)];
        assert(length(imgs)==nlevel);

    case 'cvCubic/cvArea'
        assert(length(varargin)==1,'incorrect number of arguments');
        scales = rowvec( varargin{1} );
        sizes = check_scales(im,scales);

        imgs = cell(1,length(scales));
        for si = 1:length(scales)
            s = scales(si);
            if s > 1 % upscaling
                imgs{si} = cn_imresize(im,sizes(si,:),'cvCubic');
            elseif s < 1 % downscaling
                imgs{si} = cn_imresize(im,sizes(si,:),'cvArea');
            else 
                imgs{si} = im;
            end
        end

    case 'cvCubic/cvArea+cvGaussPyr'
        assert(length(varargin)==2,'incorrect number of arguments');
        scales = rowvec( varargin{1} );
        pybias = varargin{2}; assert(isscalar(pybias)); assert(pybias<=0,'pybias<=0 is forced to avoid mistakes');
        sizes = check_scales(im,scales);
        imgs = cell(1,length(scales));

        % Calculation example:
        % scales=(1/1.2).^(-4:20);
        pyi = fix(log(scales)/log(.5)); % base 0 pyramid indices, including upscaling.
        % [scales; .5.^pyi; pyi]

        gpyr = cn_impyramid(im,'cvGaussPyr',max([1 max(pyi)+1+pybias]));

        for si = 1:length(scales)
            s = scales(si);
            if s > 1 % upscaling
                imgs{si} = cn_imresize(im,sizes(si,:),'cvCubic');
            elseif s < 1 % downscaling
                from = gpyr{max([pyi(si)+1+pybias,1])};
                imgs{si} = cn_imresize(from,sizes(si,:),'cvArea');
            else 
                imgs{si} = im;
            end
        end

    case 'imresize'
        assert(length(varargin)==1,'incorrect number of arguments');
        scales = rowvec( varargin{1} );
        sizes = check_scales(im,scales);

        imgs = cell(1,length(scales));
        for si = 1:length(scales)
            s = scales(si);
            if s ~= 1 
                imgs{si} = imresize(im,sizes(si,:));
            else 
                imgs{si} = im;
            end
        end
    otherwise
        error(['unknown method: ' method]);
end



function [sizes] = check_scales(im,scales)

assert(all(scales>0),'scale<=0 is not allowed');
sizes = ceil( bsxfun(@times, [size(im,1) size(im,2)], scales(:)) ); % never change this!

