function [pts mapinfo] = cn_mappts(refbb,targetbb,pts)
% [pts mapinfo] = cn_mappts(refbb,targetbb,pts)
%
% Map point coordinates from one reference bounding box to another.
% 
% INPUT
% refbb         Reference bb as x1y1x1y2 (#ref x 4) [Note #ref=1 usually).
% targetbb      Target bb as x1y1x2y2 (1 x 4).
% [pts]         2 x #point x #ref. Each column is [x y]'. 
%
% OUTPUT
% pts          Normalized points. (So, it is safe to use round() on them)
% mapinfo      Normalization info to use with cn_mapptsundo()
%               .shift  [shift_x shift_y]
%               .scale  [scale_x scale_y]
%
% Example:
% [x1 mapinfo] = cn_mappts([5 5 14 15],[1 100 1 100],[5 14; 14 10; 4 15]')
% cn_mapptsundo(x1,mapinfo)
%
% NOTE #ref support is provided for efficiency only but applicable in quite special cases.
%
% SEE cn_mapptseasy() cn_normalizedata() cn_mappts_undo() cn_mappts_apply() cn_mappts_applytobb
%
% Note: In cn_normalizedata(), the range is found from the provided data.  Here, we give the range explicitly using the
%       reference bounding box.
%
% R.G.Cinbis Oct 2010

[t_x sc_x] = i_onedim(targetbb(1),targetbb(3),refbb(:,1),refbb(:,3)); 
[t_y sc_y] = i_onedim(targetbb(2),targetbb(4),refbb(:,2),refbb(:,4)); 

% first shift than scale:
old_shift = [t_x; t_y];
old_scale = [sc_x; sc_y];

% first scale than shift
mapinfo.scale1 = old_scale;
mapinfo.shift2 = old_shift .* old_scale;

if nargin < 3
    pts = [];
end

[pts] = cn_mappts_apply(pts,mapinfo);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [shift scale] = i_onedim(L,H,minx,maxx)
% minx,maxx: source range (reference bb)
%            Note: these can be an array
% L,H: target range.
%            Note: each has to be a scalar.

% The normalization is a copy paste from cn_normalizedata()
%
%   shift = s = -min(x) + L * (max(x)-min(x)) / (H-L)
%   scale = t = (H - L) / (max(x)-min(x)). Note that t >= 0.
% cn_assert(L < H,'LOW < HIGH is not satisfied.');
%
% NOTE Dont apply +1 here! For example, suppose that we did.
%      Let ref be [0 1] and target be [2 10]
%      Current mapping: 0->2, 1->10
%      With +1 based mapping: 0->2,1->6.5!
%      So, don't change the code! 

Rx = (maxx-minx);
Rr = H - L;

% if the target is constant, the current form is not good.
% but given that this should not be the case normally, we can 
% simply approximate it:
Rr(abs(Rr) < 0) = eps;

shift = -minx + L .* Rx ./ Rr;
scale = Rr ./ Rx;

% if a dimension is constant (like bias) then this becomes crazy.
% in this case, shouldnt try stretching it. 
% we set it to upper bound in this case so that feature becomes H.
m = (abs(Rx) < eps);
if any(m)
    shift(m) = 0;
    k = maxx(m); k(k==0) = 1;
    scale(m) = H ./ k;
end

% NOTE
shift = reshape(shift,1,1,[]);
scale = reshape(scale,1,1,[]);

