function data2 = cn_mapptseasy(op,imsize,datatype,data)
% Simplified interface to "cn_mappts" for converting point and bounding box coordinates between actual and normalized
% coordinates.
%
% INPUT
% op        A predefined operation:
%           'im->uint8'  From image coordinates (1..imsize) to uint8 0-255 normalized coordinates.
%                        Out-of-range bounding boxes are cropped to the inside of the image due to uint8 restrictions.
%           'uint8->im'  Input data is uint8, output is single/double in range 1..imsize.
%           'im->unitsz' From image coordinates (1..imsize) to single/double 0-1 normalized coordinates. 
%           'unitsz->im' Input data is single/double in range [0 1], output is single/double in range 1..imsize.
%           'im->im'     Map from one image size to another. In this case
%                        imsize should be a cell array of length 2.
% imsize    [height width] or [height width nband] ("nband" is ignored).
% datatype  Information about the "data" argument:
%           'pts'   (2 npts) point locations, each column is [x;y]
%           'bbs'   (nbb 4) bounding box coordinates, each row is [x1 y1 x2 y2]
% data
%
% OUTPUT
% data2     Data with mapped coordinates.
%                           
% R.G.Cinbis Jan 2012 

if ~strcmp(op,'im->im')
    imsize = checkimsize(imsize);
else
    assert(iscell(imsize));
    assert(numel(imsize)==2);
    imsize{1} = checkimsize(imsize{1});
    imsize{2} = checkimsize(imsize{2});
end

switch(op)
    case 'im->uint8'
        % [1 imsz] -> [0 255]

        data = double(data); % input may be int16,etc.
        [~,map] = cn_mappts([1 1 imsize([2 1])],[0 0 255 255]);
        data2 = i_apply(datatype,detach(data),map);

        % OK: uint8(-1)=0 , uint8(300)=255
        data2 = uint8( round( data2 ) );

    case 'uint8->im'
        % [0 255] -> [1 imsz]

        assert(isa(data,'uint8'),'input data has incorrect type (not uint8)');
        %if isa(imsize,'double')
        %    data = double(data); % work in double
        %else
        %    data = single(data); % work in single
        %end
        data = double(data);
        imsize = double(imsize(:)');

        map.scale1 = (imsize([2 1])'-1)/255; map.shift2 = ones(2,1); % slightly faster

        %[~,xmap] = cn_mappts([0 0 255 255],[1 1 imsize([2 1])]); 
        %assert(isequal(map,xmap));

        data2 = i_apply(datatype,detach(data),map);

    case 'im->unitsz'
        % [1 imsz] -> [0 1]

        data = double(data); % input may be int16,etc.
        [~,map] = cn_mappts([1 1 imsize([2 1])],[0 0 1 1]);
        data2 = i_apply(datatype,detach(data),map);

    case 'unitsz->im'
        % [0 1] -> [1 imsz]
        
        assert(isfloat(data),'input data has incorrect type');
        [~,map] = cn_mappts([0 0 1 1],[1 1 imsize([2 1])]);
        data2 = i_apply(datatype,detach(data),map);

    case 'im->im'
        data = double(data); % input may be int16,etc.
        [~,map] = cn_mappts([1 1 imsize{1}([2 1])],[1 1 imsize{2}([2 1])]);
        data2 = i_apply(datatype,detach(data),map);

    otherwise
        error('unknown op');
end



function data2 = i_apply(datatype,data,map)

switch(datatype)
    case 'pts'
        assert(ndims(data)==2,'input data has incorrect size');
        assert(size(data,1)==2,'input data has incorrect size');
        data2 = cn_mappts_apply(data,map);
    case 'bbs'
        data2 = cn_mappts_applytobb(data,map);
    otherwise
        error('unknown datatype');
end


function imsize = checkimsize(imsize)
% check and standardize image size

imsize = double(imsize);
if numel(imsize)==3
    imsize = imsize(1:2);
end
assert(isequal(size(imsize),[1 2]),'imsize should be [height width] or [height width nband]');

