function out = cn_vldsifthelper(op,p)
% Various functions to help using vl_dsift()
% Current aimed at single scale operations.
% 
% USAGE
% bounds = cn_dsiftboxes('centralize',p)
%   Return 'bounds' so that SIFT points will be centralized without changing the number of points.
%   This is usually the first step to call since it reduces out-of-bounds areas.
%   If p.bounds is already provided, this will realign within the bounds and return new bounds.
%
% f_xy = cn_dsiftboxes('centers',p)
%   Compute SIFT center locations that are equivalent to those returned by vl_dsift.
%   f_xy:   (2 npt)     Each column is x,y.
%
% f_bb = cn_dsiftboxes('boxes',p)
%   Compute SIFT bounding box, where coordinates are continuous values.
%   Here, (f_bb(:,3)-f_bb(:,1))==p.size*4.
%   Usually this only useful for visualization.
%   f_bb:   (4 npt)     Each column is x1,y1,x2,y2.
%
% f_bb = cn_dsiftboxes('intboxes',p)
%   Compute SIFT boxes such that boundaries are pixel coordinates.
%   This can be interpreter as the centers of the boundary pixels of the boxes returned by 'boxes'.
%   Therefore, here, (f_bb(:,3)-f_bb(:,1)+1)==p.size*4. (ie, boundaries are inclusive.)
%   Coordinates for boundary cells can be outside the image coordinates.
%   f_bb:   (4 npt)     Each column is x1,y1,x2,y2.
%
% SHARED PARAMS STRUCT
% p
%   .imsz       (1 2)       [height width]
%   .size       (scalar)    Spatial bin size. Note that SIFT patch size is 4*size.
%   .step       (scalar)    Sampling size.
%   .[bounds]   [xmin ymin xmax ymax]   
%   .[debug]    (bool)      Enable debugging or not.
% 
% R.G.Cinbis, Sep 2012

p = cn_setfielddefaults(p,true,'debug',false);

if 1 && p.debug
    mywarn('debug: modifying bounds');
    p.bounds = cn_vldsifthelper('centralize',setfield(p,'debug',false))
end

cn_assertsize('d',p.imsz,[1 2]);
cn_assertsize('d',p.size,[1 1]);
cn_assertsize('d',p.step,[1 1]);

assert(prod(p.imsz)>0,'empty image is not supported');

if isfield(p,'bounds');
    cn_assertsize('d',p.bounds,[1 4]);
    assert(all(p.bounds(1:2)<=p.bounds(3:4)),'invalid bounds');

    % clip the bounds to the image size, just like vl_dsift()
    p.bounds(1:2) = max(p.bounds(1:2),1);
    p.bounds(3) = min(p.bounds(3),p.imsz(2));
    p.bounds(4) = min(p.bounds(4),p.imsz(1));
else
    p.bounds = [1 1 p.imsz(2) p.imsz(1)]; % x1,y1,x2,y2
end

% bound width and height
bh = p.bounds(4)-p.bounds(2)+1;
bw = p.bounds(3)-p.bounds(1)+1;

% short cut names
b = p.size;
t = p.step;

% for the very first SIFT descriptor with default image bounds [considering only x coordinate]:
%   - center of leftmost cell is at 1
%   - center of rightmost cell is at (1+3b)
%   - center of whole thing is (1+1.5b)
%
% note that center of the descriptor does not correspond to a descriptor pixel. it is the mid
% crossing point of 4x4 cells. however, it may correspond to an actual pixel.
%
% note that center of a cell does not correspond to a descriptor pixel when bin size is even.
% however, it is defined to be always image pixel aligned. (though, low-level implementation is
% not explained). for example, if size=4, center of a cell is the center of a 4x4 box whose center lies on a
% pixel. this is at least what I see using vl_plotsiftdescriptor().

% number of sift points that can fit into the image
if 1+3*b <= bw
    nx = floor( (bw-1-3*b) / t ) + 1;
else
    nx = 0;
end
if 1+3*b <= bh
    ny = floor( (bh-1-3*b) / t ) + 1;
else
    ny = 0;
end

switch(op)
    case 'centralize'
        % new bounds
        p.bounds(1) = p.bounds(1) + max(0,floor( (bw-(nx-1)*t-3*b-1) / 2 ));
        p.bounds(2) = p.bounds(2) + max(0,floor( (bh-(ny-1)*t-3*b-1) / 2 ));
        out = p.bounds;
        
    case 'centers'
        f_xy = getcenters(p,nx,ny);
        out = f_xy;

    case 'boxes'

        % boundary of the leftmost pixel of the leftmost cell is 1-b/2
        [x0,y0] = meshgrid(p.bounds(1)-0.5*b+(0:(nx-1))*t,p.bounds(2)-0.5*b+(0:(ny-1))*t);
        x0 = x0(:)';
        y0 = y0(:)';
        f_bb = [x0; y0; x0+4*b; y0+4*b];
        out = f_bb;

    case 'intboxes'
        % since vl_dsift has a tendency to make out-of-bounds at the left and top corners
        % of the image, we assume that, center pixel of a cell is the pixel at floor((size+1)/2).
        % this is also similar to mathematical morphology convention of matlab.

        [x0,y0] = meshgrid(p.bounds(1)+1-floor((b+1)/2)+(0:(nx-1))*t,p.bounds(2)+1-floor((b+1)/2)+(0:(ny-1))*t);
        x0 = x0(:)';
        y0 = y0(:)';
        f_bb = [x0; y0; x0+4*b-1; y0+4*b-1];
        out = f_bb;
 
    otherwise
        error('unknown');
end

if p.debug
    % -- debug, quick check --
    [f,d] = vl_dsift( zeros(p.imsz,'single'),'size',p.size,'step',p.step,'fast','bounds',p.bounds);
    f
    if nx*ny > 0
        assert(length(unique(f(1,:)))==nx);
        assert(length(unique(f(2,:)))==ny);
    else
        assert(isempty(f));
    end

    switch(op)
        case 'centralize'
            %  OK: nx,ny shouldn't have change with new p.bounds.
            mywarn('passed internal tests');

        case 'centers'
            assert(isequal(f,f_xy));
            mywarn('passed internal tests');

        case 'boxes'

            mywarn('see plot,red lines are computed by us');
            for j = 1:size(f_bb,2)

                im = rand(p.imsz);
                clf; imagesc(im); colormap(gray);
                f(3,:) = p.size/3;
                f(4,:) = 0;
                vl_plotsiftdescriptor(d(:,j),f(:,j));
                imrectplot([],'ep',f_bb(:,j)',[],'r',true);
                cn_guipause
            end

        case 'intboxes'
            mywarn('see plot,red pixels denote four corners,yellow denote new center');

            for j = 1:size(f_bb,2)
                im = repmat( rand(p.imsz), [1 1 3] );
                im = debug_setcolor(im,f_bb(2,j),f_bb(1,j),[1 0 0]);
                im = debug_setcolor(im,f_bb(4,j),f_bb(1,j),[1 0 0]);
                im = debug_setcolor(im,f_bb(2,j),f_bb(3,j),[1 0 0]);
                im = debug_setcolor(im,f_bb(4,j),f_bb(3,j),[1 0 0]);
                clf; imagesc(im,[0 1]); colormap(gray);
                f(3,:) = p.size/3;
                f(4,:) = 0;
                vl_plotsiftdescriptor(d(:,j),f(:,j));
                hold on;plot(mean(f_bb([1 3],j)),mean(f_bb([2 4],j)),'xy');
                cn_guipause
            end

        otherwise
            error('unknown');
    end

end



function f_xy = getcenters(p,nx,ny)

b = p.size;
t = p.step;
[x,y] = meshgrid(p.bounds(1)+1.5*b+(0:(nx-1))*t,p.bounds(2)+1.5*b+(0:(ny-1))*t);
f_xy = [x(:)';y(:)'];



function im = debug_setcolor(im,y,x,clr)
% ignores if out-of-bounds

if x <= 0 || y <= 0 || x > size(im,2) || y > size(im,1)
    return
end
im(y,x,:) = clr;




