function cellbb = cn_spatialgrid(objbb,grids,prm)
% cellbb = cn_spatialgrid(objbb,grids)
%
% Returns a carefully splitted spatial grid coordinates for a given set of
% bounding boxes. Fast.
% 
% INPUT
% objbb     Each column is [x1,y1,x2,y2]' of bounding boxes for which descriptor
%           to be calculated.
% grids     Cell array such that {[h1 w1],[h2 w2],...etc}. Each entry depicts
%           H x W spatial grid within each bounding box. 
%           H is the number of vertical cells, W is the number of horizontal cells.
% prm
%   .normcoor    Whether we're working on [0 1]-like normalized continuous coordinates or whether in discrete pixel coordinates.
%                If false, cellbb will be directly pixel coordinates to be used with <= and >=.
%                If true, usually used as cn_spatialgrid([0;0;1;1],grids,prm);
%   .[fixcoor]   Fix input coordinates.
%                default=~normcoor (dont change the default for consistency!)
%                If true, normcoor=0: round and trim to objbb coor>=1.
%                         normcoor=1: trim to objbb coor>=0.
%   .[avoidov]   Avoids overlap. default=~normcoor. (Use default)
%   
% OUTPUT
% cellbb:   (4 #cells #bb). Each column is [x1,y1,x2,y2]'. Each is for a bounding box in double.
% 
% * Ordering of cells doesnt depend on "objbb". See cn_spatialgrid_info().
%   Guaranteed to produce col-major order of cells within each grid.
%
% SEE cn_partlocationsgrid cn_spatialgrid_cellnames cn_spatialgrid_selpts
% cn_spatialgrid_display
% 
% R.G.Cinbis, Feb 2010. Updated Sep'11.

% OBSOLETE normcoor=~validcoord, where validcoord option is removed.

prm = cn_setfielddefaults(prm,1,'display',0,'avoidov',~prm.normcoor,'fixcoor',~prm.normcoor);

% input check
cn_assert(size(objbb,1)==4);
if ~iscell(grids)
    grids = {grids};
end

objbb = double(objbb); % added on 2012-09-11.

% find # grid cells
nbb = size(objbb,2);
ncells = zeros(1,length(grids));
ngrid = length(grids);
for gi = 1:ngrid
    ncells(gi) = prod(grids{gi});
end

% make integers
if prm.fixcoor
    if prm.normcoor
        objbb(objbb < 0) = 0;
    else
        objbb = round(objbb);
        objbb(objbb < 1) = 1;
    end
end

% get coordinates separately for efficiency.
t = objbb.';
bbx1 = t(:,1);
bby1 = t(:,2);
bbx2 = t(:,3);
bby2 = t(:,4);

% init output
cellbb = zeros(4,sum(ncells),nbb);
if nbb == 0
    return;
end
cellbb_j1 = 1;
for gridi = 1:ngrid

    % assume cgrid(:) > 0
    cgrid = grids{gridi};
    cellbb_j2 = cellbb_j1 + ncells(gridi) - 1;

    % compute grid as indices
    % it is likely be more efficient in use to give col-major ordering
    % of cells. meshgrid does this.
    c_ncelly = cgrid(1);
    c_ncellx = cgrid(2);
    [meshx meshy] = meshgrid(1:c_ncellx,1:c_ncelly);
    meshx = meshx(:);
    meshy = meshy(:);
    
    % find corresponding real coordinates
    % ally(1,1):   bounding box' y1
    % ally(end,1): bounding box' y2
    % observe transpose: each col is for an object
    ally = cn_linspace(bby1,bby2,c_ncelly+1).'; % length(y1) = # vert cells + 1
    allx = cn_linspace(bbx1,bbx2,c_ncellx+1).';

    if ~(prm.normcoor)
        % y1(1) and y2(end) are known to be valid.
        ally = floor(ally); % dont use round here
        allx = floor(allx);
        ally1 = ally(1:(end-1),:);
        allx1 = allx(1:(end-1),:);
        ally2 = ally(2:end,:);
        allx2 = allx(2:end,:);
        
        if prm.avoidov
            ally1(2:end,:) = ally1(2:end,:) + 1;
            allx1(2:end,:) = allx1(2:end,:) + 1;
        end
    else
        ally1 = ally(1:(end-1),:);
        allx1 = allx(1:(end-1),:);
        ally2 = ally(2:end,:);
        allx2 = allx(2:end,:);

        if prm.avoidov
            % this looks good but that I'm not 100% percent sure yet.  in general, for normcoor, one
            % might allow boundary overlaps. overlaps isn't an issue in most applications.
            ally1(2:end,:) = ally1(2:end,:) + eps(ally1(2:end,:));
            allx1(2:end,:) = allx1(2:end,:) + eps(allx1(2:end,:));
        end
    end
    clear ally allx

    % some crazy index calculation
    % reshape(meshy,[],1) -> all y1 linear indices for first bounding box' mesh grid
    % bsxfun -> to quickly calculate for other bounding boxes.
    yi = bsxfun(@plus,reshape(meshy,[],1),[0:(nbb-1)] * size(ally1,1)); % #vertcels x #bb, y1 location indices for mesh
    xi = bsxfun(@plus,reshape(meshx,[],1),[0:(nbb-1)] * size(allx1,1));
    y1 = reshape(ally1(yi),1,[],nbb);
    x1 = reshape(allx1(xi),1,[],nbb);
    y2 = reshape(ally2(yi),1,[],nbb);
    x2 = reshape(allx2(xi),1,[],nbb);

    cellbb(:,cellbb_j1:cellbb_j2,:) = [x1; y1; x2; y2];

    cellbb_j1 = cellbb_j2 + 1;
end


