function out = cn_assertsize(info,arr,varargin)
% Get and assert size of the variables easily.
% Checking fewer dimensions than ndims(arr) is an error.
%
% USAGE
%
% cn_assertsize('c',arr,d1,d2,...) 
%   Store dimension variables in the caller workspace. 'c' is case insensitive.    
%   dj     If scalar: Assert size(arr,j)==dj.
%          If string: If exists and not [], assert size(arr,j)==evalin('caller',dj)
%                     Otherwise, assignin('caller',dj,size(arr,j))
% 
% cn_assertsize('d',arr,dims)      
%   Dimensions are provided directly. 'd' is case insensitive.
%   Assert size(arr,j)==dims(j). 
%
% info = cn_assertsize(info,arr,d1,d2,...)
%   Store dimension variables in a struct.
%   info    [] or output from an earlier cn_assertsize() call.
%   dj      If scalar: Assert size(arr,j)==dj.
%           If string: If exists, assert size(arr,j)==info.(dj)
%                      Otherwise, info.(dj) = size(arr,j)
%   Returns updated info structure.
%
% R.Gokberk Cinbis,Sep 2011. Update:May 2012.

% todo: Allow "either 1 or D" support.

name = inputname(2);

if ischar(info)
    switch(lower(info))
        case 'd'
            assert(length(varargin)==1,'Incorrect number of arguments for info==''D''');
            assert(nargout==0);
            D = varargin{1};
            x = cn_sizend(arr, length(D));
            if ~isequal( x, D )
                error(sprintf('%s: Size assertion fails: Required: [%s] Actual: [%s] \n',name,num2str(D),num2str(x)));
            end
            return;
        case 'c'
            assert(ndims(arr)<=length(varargin),'All non-singleton dimensions should be checked.');
            for d = 1:length(varargin)
                n = varargin{d};
                if isnumeric(n)
                    assert(isscalar(n));
                    i_assert(name,arr,d,n);
                elseif ischar(n)
                    if evalin('caller',['exist(''' n ''',''var'')'])~=0 && ~isempty(evalin('caller',n))
                        i_assert(name,arr,d,evalin('caller',n));
                    else
                        assignin('caller',n,size(arr,d));
                    end
                else
                    error('unsupported entry');
                end
            end
        otherwise
            error('unknown info option');
    end
else
    assert(ndims(arr)<=length(varargin),'All non-singleton dimensions should be checked.');
    for d = 1:length(varargin)
        n = varargin{d};
        if isnumeric(n)
            assert(isscalar(n));
            i_assert(name,arr,d,n);
        elseif ischar(n)
            if isfield(info,n)
                i_assert(name,arr,d,info.(n));
            else
                info.(n)=size(arr,d);
            end
        else
            error('unsupported entry');
        end
    end
    out = info;
end




function i_assert(name,arr,d,val)

assert(~isnan(val));
if size(arr,d)~=val
    error(sprintf('%s: Size-assertion fails in dimension %d Required:%d Actual:%d\n',...
        name,d,val,size(arr,d)));
end



