/*
 * R.G.Cinbis Jan 2012. 
 *
 * Also see cn_powsum() for a similar implementaiton.
 */

#include "mex.h"
#include <cmath>
#include <math.h>
#include <vector>
#include <algorithm>
#include <cstring>
#include <omp.h>
#include "cn_threads.hpp"
#include "cn_macros.hpp"

/* currently I use a linear search (backwards) to find whether
 * to add the item. this should be good because we assume k to be really small
 * (otherwise, insertion step is slow). still, we may consider 
 * sth like quick sort but I doubt that it would actually be faster 
 */

/**
 * TODO
 *  Use a smarter algorthim.
 *  Use inline function templates instead of macros.
 */

#define INSERT(OP) T *x_v; \
    double *x_ind; \
    mwSize k2 = MIN(itemi,k); /* number of entries in the list */ \
    mwIndex newj = 0; /* mwIndex is unsigned! default=0 is important */ \
    /* search position to add from the end to the begggining. since we assume k << n, many items is expected to never enter the list */ \
    x_v = vals + (k2-1) * step; \
    for (mwSignedIndex j = (k2-1); j >= 0; --j) { /* j has to be signed for >= 0 to work. */ \
        if (newval OP *x_v) { \
            newj = j + 1; /* replace the one item to the right. shift them to right. */ \
            break;  \
        } \
        x_v -= step; \
    } \
    if ( (itemi < k)  ) { /* expand list */ \
        ++k2; \
    } \
    /* printf("newj=%d k2=%d k=%d itemi=%d newval=%f\n",newj,k2,k,itemi,(double) newval);  */  \
    if (newj < k2) { \
        x_v = vals + (k2-1) * step; \
        x_ind = ind + (k2-1) * step; \
        for (mwIndex j = (k2-1); j > newj; --j ) { /* shift & insert */ \
            *x_v = *(x_v-step); \
            *x_ind = *(x_ind-step); \
            x_v -= step; \
            x_ind -= step; \
        } \
        *x_v = newval; \
        *x_ind = itemi+1; /* c->matlab coords */ \
    } \

// insert item if it is larger/smaller than the current top k
// newval   new value candidate to insert to the list
// vals     a vector of k elements
// step     stride between elements in vals
// k        length of vals vector.
// itemi     the insert-attempt index. if itemi<k, num elements in vals is itemi, otherwise it is k.
template <class T>
void insert_ascend(T newval,T *vals,double *ind,mwSize step, mwSize k, mwSize itemi) {
    // >= and <= guarantees the following:
    // When more than one element has the same value, the order of the
    // elements are preserved in the sorted result and the indexes of
    // equal elements will be ascending in any index matrix.
    INSERT(>=)
}

template <class T>
void insert_descend(T newval,T *vals,double *ind,mwSize step, mwSize k, mwSize itemi) {
    INSERT(<=)
}

#define LOOP(insertfun) for (mwIndex p = 0; p < (d1*d3); ++p) {  \
            mwSize j3 = p / d1; /* 0...d3-1 */ \
            mwSize j1 = p - j3*d1; /* 0...d1-1 */ \
            T *xx = X + d1d2*j3 + j1; \
            T *xY = Y + d1k*j3 + j1; /* d2=k in the Y */ \
            double *xI = I + d1k*j3 + j1; \
            for (mwIndex j2 = 0; j2 < d2; ++j2) { \
                insertfun( *xx, xY, xI, d1, k, j2 ); \
                xx += d1; \
            } \
        }

template <class T>
void helper(T *X, T *Y, double *I, mwSize k, mwSize d1, mwSize d2, mwSize d3, bool isascend ) {

    cn::setmaxthreads();

    mwSize d1d2 = d1*d2;
    mwSize d1k = d1*k;

    if (isascend) {
#pragma omp parallel for
        LOOP(insert_ascend)
    } else { 
#pragma omp parallel for
        LOOP(insert_descend)
    }
}

// check if a is an integer valued double scalar and return.
double getintvaldouble(const mxArray *a) {
    if (mxGetNumberOfElements(a)!=1) {
        mexErrMsgTxt("k and dim arguments must be a scalar.");
    }
    double x = mxGetScalar(a);
    if (x != lrint(x)) {
        mexErrMsgTxt("k and dim arguments must be an integer-valued scalar.");
    }
    return x;
}

#define argX 0
#define argk 1
#define argdim 2
#define argIsAscend 3

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

    /*  check for proper number of arguments */
    if ( (nrhs!=4) ) {
        mexErrMsgTxt("Invalid #input");
    }

    // input data
    mxClassID type = mxGetClassID(prhs[argX]);
    void *X = mxGetData(prhs[argX]);
    mwSize Xndim = mxGetNumberOfDimensions(prhs[argX]);
    const mwSize *Xdims = mxGetDimensions(prhs[argX]);
    if (mxIsComplex(prhs[argX])) {
        mexErrMsgTxt("Complex data is not supported.");
    }
    if (mxIsSparse(prhs[argX])) {
        mexErrMsgTxt("Sparse data is not supported.");
    }

    double tmp = getintvaldouble(prhs[argk]);
    if (tmp < 0) {
        mexErrMsgTxt("k argument should be nonnegative integer");
    }
    mwSize k = (mwSize) tmp;

    tmp = getintvaldouble(prhs[argdim]);
    if (tmp < 0) {
        mexErrMsgTxt("dim argument should be a positive integer");
    }
    mwSize dim = (mwSize) tmp;

    bool isascend = mxGetScalar(prhs[argIsAscend])!=0;

    // we think simply the case as
    // 
    //   z = reshape(x,[prod(sz(1:d-1)),sz(d),prod(sz(d+1:end))
    //
    // ie. z has 3 dmensions and we sort in dimension 2.

    mwSize d1 = 1; // size(z,1)
    for (mwIndex i = 1; i < MIN(dim,Xndim+1); ++i ) {
        d1 *= Xdims[i-1];
    }
    mwSize d2 = 1; // size(z,2)
    if ( dim <= Xndim ) {
        d2 = Xdims[dim-1];
    }
    mwSize d3 = 1; // size(z,3)
    for (mwIndex i = (dim+1); i <= Xndim; ++i ) {
        d3 *= Xdims[i-1];
    }

    mwSize outnumdim = MAX(Xndim,dim);
    mwSize *outdims = (mwSize *) mxMalloc(sizeof(mwSize)*outnumdim);
    if (outdims==NULL) { 
        mexErrMsgTxt("Cannot allocate outdims"); 
    }
    memcpy((void *) outdims,(const void *) Xdims,sizeof(mwSize)*Xndim);
    for (mwSize d=(Xndim+1); d <= outnumdim; ++d) {  // base 1
        outdims[d-1] = 1;
    }
    outdims[dim-1] = k;

    // check if k is meaningful
    if ( dim > Xndim ) {
        if (k > 1) 
            mexErrMsgTxt("Too high k is provided.");
    } else {
        if ( Xdims[dim-1] < k ) 
            mexErrMsgTxt("Too high k is provided.");
    }

    plhs[0] = mxCreateNumericArray(outnumdim, outdims, type, mxREAL); // "Y" of type T, initialized to 0
    plhs[1] = mxCreateNumericArray(outnumdim, outdims, mxDOUBLE_CLASS, mxREAL); // "I" of type double, initialized to 0
    void *Y = mxGetData(plhs[0]); // can be NULL if output is going to be an empty matrix.
    double *I = (double *) mxGetData(plhs[1]); // can be NULL if output is going to be an empty matrix.
    if (Y == NULL) {
        return; // nothing left to do. this is important, the above loop may fail in this bounday case.
    }

    // call the main routie
    switch(type) {
        case mxDOUBLE_CLASS:
            helper((double *)X,(double *)Y,I,k,d1,d2,d3,isascend);
            break;
        case mxSINGLE_CLASS:
            helper((float *)X,(float *)Y,I,k,d1,d2,d3,isascend); // cast k to float?
            break;
        default:
            mexErrMsgTxt("Unsupported data class -- easy to implement!");
    }


}


