function out = cn_getmixmodelpbs(data,opts)
% out = cn_getmixmodelpbs(data,opts)
%
% Utility function to evaluate GMM-based models. 
%
% INPUT
% data
%   .logcond       (N K) p(x|c) or p(x|c,y) entries.
%   .logprior      (1 K) p(c)   or (N K) p(c|y) entries.
% opts             A cell array of requested output names, eg. {'logcond','post'} 
%
% OUTPUT
% out
%   .logcond    Same as input.
%   .logprior   Same as input.
%   .[logj]     (N K) logp(x,c) or logp(x,c|y)  -- Sufficient for cluster assignments.
%   .[post]     (N K) p(c|x)    or p(c|x,y)     -- sum(post,2)==1
%   .[marg]     (N 1) p(x)      or p(x|y)       -- i.e. pdf.
%   .[logm]     (N 1) log(marg)                 -- i.e. log pdf. May slightly differ whether [logm] is also requested or not.
%
% SEE cn_selections2struct
%
% R.G.Cinbis, Aug 2011

% note: this function has a little overhead due to asking logcond and logprior
% separately. however, this is worth it to simplify matters.

ALL = {'logcond','logprior','logj','post','marg','logm'};

% cp_init is a struct containing "false" for each possible output field.
persistent cp_init
if isempty(cp_init)
    cp_init = cn_selections2struct(ALL,{});
end

cp = cn_selections2struct(cp_init,cellifnot(opts));

% check data
N = size(data.logcond,1);
K = size(data.logcond,2);
assert(ndims(data.logcond)==2);
assert(ndims(data.logprior)==2);
assert(size(data.logprior,2)==K);
if (size(data.logprior,1)~=N) && (size(data.logprior,1)~=1)
    error('logprior has invalid size in dim 1');
end

out.logcond = data.logcond;
out.logprior = data.logprior;
clear data;

out.logj = bsxfun(@plus,out.logcond,out.logprior);

% to release memory
out = i_clean(out,cp,'logcond');
out = i_clean(out,cp,'logprior');

if cp.post 
    out.post = cn_softmax(out.logj,2);
end

if cp.marg 
    out.marg = sum(exp(out.logj),2);
end

if cp.logm
    if cp.marg
        out.logm = log(realmin+out.marg);
    elseif cp.post
        out.logm = i_post2logm(out.logj,out.post,N,K);
    else
        out.post = cn_softmax(out.logj,2);
        out.logm = i_post2logm(out.logj,out.post,N,K);
    end
end

% full clean-up, remove all unneeded fields.
fnames = fieldnames(out);
for i = 1:length(fnames)
    out = i_clean(out,cp,fnames{i});
end







function out = i_clean(out,cp,fname)
% remove variable if not needded in the output.
% call this function, whenever there is no more variable needed.

if ~(cp.(fname))
    out = rmfield(out,fname);
end



function logm = i_post2logm(logj,post,N,K);

% out.logm = log( sum(exp(out.logj),2) );
% out.logm = cn_logofsumofexp(out.logj,2);

% note: alternatives to get log-marginals
% p(x) = p(x,c)/p(c|x) = p(c|x)p(x)/p(c|x)
% simple method: O = logj(:,1)-log(post(:,1));
% if p(c|x) is close to 0, this can fail.
%
% instead:
% find the dimension where p(c|x) is maximum
% then do this operation at that dimension.
[maxpost,i]=max(post,[],2);
i = sub2ind([N K],1:N,i(:).');
%old: logm = reshape(  logj(i) - log(post(i)), N,1);
logm = reshape(logj(i),N,1) - log(maxpost(:));






