function [nsamples,masks] = cn_subsamplesets(p)
% [nsamples,masks] = cn_subsamplesets(p)
%
% Given sizes of sets, returns #items to sample from each set. Number of
% samples is proportional to number of items within each set.
% Overall random state is preserved.
%
% INPUT
% p
%   .[method]   Approach to used.
%               'determ+round'  nsamples is deterministic, behaves pretty well.
%   .nitems     Original #items in each set. (1 #set)
%   .maxtotal   Maximum number of features to preserve at the end.
%   .[randseed] Defines the internal randomizer. 
%
% OUTPUT
% nsamples      Number of samples to get from each set. (1 #set)
%               if maxtotal<=sum(p.nitems), maxtotal=sum(nsamples)
% [masks]       Cell of masks. (1 #set). masks{j} is a logical 
%               mask (1 p.nitems(j)).
% 
% Good practices:
% - When loading the data, usually it is good to subsample a certain
%   ratio of items within each set, before utilizing this function.
%
% EXAMPLES
% cn_subsamplesets(struct2('nitems',[1 2 3],'maxtotal',3))
% cn_subsamplesets(struct2('nitems',[1 2 10],'maxtotal',3))
%
% SEE cn_setrandseed cn_saverandstream cn_loadrandstream
% 
% R.G.Cinbis July 2011

p = cn_setfielddefaults(p,true,'randseed',[],'method','determ+round');

randinfo = cn_saverandstream();
if ~isempty(p.randseed)
    % all operations are deterministic given this seed.
    cn_setrandseed(p.randseed);
end

% avoid boundary cases due to boundary cases.
nitems = p.nitems;
assert(min(nitems)>=0);
assert(p.maxtotal>=0);
nset = length(nitems);
ntotalorg = sum(nitems);

if ntotalorg <= p.maxtotal
    % no sampling
    nsamples = nitems;
else
    % ratios
    ks = min(nitems / ntotalorg,1);
    assert(all(isfinite(ks)));

    switch(p.method)
        case 'determ+round'
            % ceil and floor doesn't well.
            % floor: we may end up sampling from a single set.
            % ceil: up to a point we continue sampling 1 from each set.
            % round suggests a more distributed sampling but may sample too-many or too-little.
            nsamples = round(p.maxtotal.*ks);
            dif = sum(nsamples)-p.maxtotal;
            if dif > 0
                % too many samples
                while(dif>0)
                    [x,i] = sort(nsamples,'descend');
                    i = i(1:min(length(i),dif));
                    i = i(x(1:length(i))>0);
                    nsamples(i) = nsamples(i)-1;

                    dif = sum(nsamples)-p.maxtotal;
                end
            elseif dif < 0
                % too little smaples
                while(dif < 0)
                    navail = max(nitems-nsamples,0);

                    % use floor here because we don't have too-many-samples any more.
                    x = floor(-dif.*ks);
                    if ~any(x) 
                        [~,i] = max(navail);
                        assert(navail(i)>0);
                        x(i) = 1;
                    end

                    nsamples = min(nitems,nsamples+x);
                    dif = sum(nsamples)-p.maxtotal;
                end
            end

            % exact #sample is achieved.
            assert(dif==0);
        otherwise
            error('unknown method');
    end
end

% todo: alternative randomization based: 
% x = rand(1,maxtotal);
% thene find nsamples by partitioning x based on #

if nargout > 1
    masks = cell(1,nset);
    for j = 1:nset
        masks{j} = cn_randmask(nitems(j),nsamples(j));
    end
end

cn_loadrandstream(randinfo);

