
function train_classifiers()

  index = dataset_definition();
  
  % make output directory
  mkdir data/classifiers
  
  % channel = descriptors + PQ centroids to decode them
  % should be in sync with the ones computed in compute_descriptors.m 
  descriptors = prepare_descriptors(1); 
 
  
  % geometrical data for training boxes
  train_candbbs = load_candidate_boxes(index, index.all_trainmask);
     
  fprintf('Loaded %d training candidate boxes\n', sample_size(train_candbbs));

  % adapted from Gokberk's fv_multiclass_objdet_pqfvmog
  
  for classi = 1:length(index.clsnames)
    
    fprintf('*** training classifier for class %s\n', index.clsnames{classi});

    % training boxes for this class
    gt_samples = prepare_gt_samples(index, classi);
    
    
    fprintf('Training from %d GT boxes\n', sample_size(gt_samples));
           
    
    % 1 st training: only GT bounding boxes and negatives candidate boxes 
    % negative boxes = candidate boxes that have ~25 % overlap with ground truth    
    neg_samples = select_negative_samples_random(index, gt_samples, train_candbbs);

    fprintf('    Loading %d gt + %d neg samples\n', ...
            sample_size(gt_samples), sample_size(neg_samples));
    
    gt_data = get_sample_descriptors(descriptors, gt_samples, 0);
    neg_data = get_sample_descriptors(descriptors, neg_samples, 1);
      
    % remove data produced from too few local features
    gt_data = enforce_minpts(gt_data); 
    neg_data = enforce_minpts(neg_data); 

    % hard mining iterations
    nhardmineiter = 4;
    for minei = 0:nhardmineiter

      
      fprintf('  Hard negative mining iteration %d: training with %d gt + %d neg samples\n', ...
              minei, size(gt_data.wscr, 2), size(neg_data.wscr, 2));
                      
      pos_data = gt_data;
     
      
      classifier = train_classifier(descriptors, pos_data, neg_data);

      
      if minei == nhardmineiter
        fprintf('  Last hard negative mining iteration\n');        
        % no need to continue mining
        break        
      end
            
      % save intermediate classifier
      save([ 'data/classifiers/classifier_' index.clsnames{classi} '_minei' num2str(minei) '.mat'], 'classifier');
      
      fprintf('    Scoring all %d candidate windows\n', sample_size(train_candbbs));
      scores = classify_all_boxes(classifier, descriptors, train_candbbs);      
      
      fprintf('    Selecting hard negatives\n');
      new_neg_samples = find_hard_negatives(gt_samples, train_candbbs, scores, merge_samples(gt_samples, neg_samples));
      
      n_new_neg_samples = sample_size(new_neg_samples);
      fprintf('    %d new negative samples\n', n_new_neg_samples);
      if n_new_neg_samples == 0 
        fprintf('  Hard negative mining converged\n');
        break; 
      end     
      
      neg_samples = merge_samples(neg_samples, new_neg_samples);
      
      fprintf('    Loading new negative descriptors\n');
      new_neg_data = get_sample_descriptors(descriptors, new_neg_samples, 1); 
      new_neg_data = enforce_minpts(new_neg_data); 
            
      neg_data = merge_train_data(neg_data, new_neg_data);
      
    end

    outfilename = [ 'data/classifiers/classifier_' index.clsnames{classi} '.mat'];
    fprintf('  Storing classifier %s\n', outfilename);
    save(outfilename, 'classifier');
    
  end
  
    
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% functions on samples (= sets of bounding boxes)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    
function ss = sample_size(samples)
  ss = 0;
  for i = 1:length(samples)
    ss = ss + size(samples(i).bbuint8, 1);
  end
end


function gt_samples = prepare_gt_samples(index, classi)
  % collect training candidate bboxes and ground-truth boxes
  
  gt_samples = struct;
        
  train_imi = find(index.all_trainmask);

  for outi = 1:length(train_imi)      
    imi = train_imi(outi);
    imsize = index.imsizes(imi, :);
    
    % gt boxes
    gtboxes = index.gtboxes(imi);
    box_subset = (gtboxes.clsind == classi); % interested only in current class
    gt_samples(outi).orgimi = imi;
    gt_samples(outi).bbuint8 = bbs_to_uint8(gtboxes.bbs(box_subset, :), imsize);
    gt_samples(outi).clsind = gtboxes.clsind(box_subset);
    gt_samples(outi).view = gtboxes.view(box_subset);
    gt_samples(outi).orgbbind = find(box_subset);
            
  end

end  



function bbs_uint8 = bbs_to_uint8(bbs, imsize)
  % rescale image bboxes to encode them in uint8 and save some space
  
  bbs = double(bbs);
  xscale = 255 / (imsize(2) - 1);
  yscale = 255 / (imsize(1) - 1);
  scales = [xscale yscale xscale yscale];
  bbs = bsxfun(@plus, bsxfun(@times, bbs, scales), -scales);
  bbs_uint8 = uint8(bbs);
  
end




function samples = empty_samples(index, gt_samples)
  % randomly select negative samples
  
  nimg = length(gt_samples);
  samples = struct;
  
  for imi = 1:nimg
    % set the output
    samples(imi).orgimi   = 0;
    samples(imi).bbuint8  = zeros(0, 4, 'uint8');
    samples(imi).clsind   = [];
    samples(imi).orgbbind = [];
    samples(imi).gtind    = [];
  end
  samples = reshape(samples,1,nimg);

end


function data = merge_train_data(data1, data2)
  data = struct;
  data.codes = [data1.codes data2.codes];
  data.wscr = [data1.wscr data2.wscr];
end

function samples1 = merge_samples(samples1, samples2)
  assert(isequal([samples1.orgimi], [samples2.orgimi]));
  samples = struct;
  for i = 1:length(samples1)
    samples1(i).bbuint8 = [samples1(i).bbuint8 ; samples2(i).bbuint8];
    samples1(i).clsind = [samples1(i).clsind ; samples2(i).clsind];
    samples1(i).orgbbind = [samples1(i).orgbbind ; samples2(i).orgbbind];
  end
    
end

function out = enforce_minpts(data) 
  % do not rely on Fisher descriptors computed from regions fewer than 20
  % local features    
  maxofnpts_th = 20; 
  m = (data.maxofnpts >= maxofnpts_th); 
  out.codes = data.codes(:, m); 
  out.orgimind = data.orgimind(:, m); 
  out.wscr = data.wscr(:, m); 
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Random/guided selection of training boxes 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function samples = select_negative_samples_random(index, gt_samples, train_candbbs)
  % randomly select negative samples
  
  orgimind = [gt_samples.orgimi];
  nimg = length(gt_samples);
  samples = struct;
  
  for imi = 1:nimg
    cn_setrandseed(100 + orgimind(imi));
    imname = index.imnames{orgimind(imi)};

    % optionally add positive samples with at least p.objdet.posminov overlap
    % add negative samples without a big overlap
    x_cand = double(train_candbbs(imi).bbuint8);
    ncand = size(x_cand,1);
    x_gt = double(gt_samples(imi).bbuint8);
    if isempty(x_gt)
      % there is no groundtruth object in the image. simply assign inf. (since -nan doesnt exist)
      posovrat = zeros(ncand,1);
      gtind = zeros(ncand,1);
    else
      O = cn_bboverlap(x_gt,x_cand); % (ngt ncand)
      [posovrat, gtind] = max(O, [], 1);
      posovrat = reshape(posovrat,[ncand 1]);
      gtind = reshape(gtind,[ncand 1]);
    end
    
        
    % keep overlaps of about 25 % with GT
    m = (posovrat <= 0.3) & (posovrat >= 0.2);
    
    % mask negatives that do not have the proper overlap
    neg_uint8 = x_cand(m, :); % keep in double 
    neg_orgbbind = find(m(:));
    neg_gtind = gtind(m, :);

    % maximum overlap between two negative samples
    scores = rand(size(neg_uint8, 1), 1);
    x = samples_nonmax_suppression(neg_uint8, scores, 0.6, -Inf, Inf);
    neg_uint8 = neg_uint8(x,:);
    neg_orgbbind = neg_orgbbind(x);
    neg_gtind = neg_gtind(x);

    % set the output
    samples(imi).orgimi   = orgimind(imi);
    samples(imi).bbuint8  = uint8(neg_uint8);
    samples(imi).clsind   = -ones(size(neg_uint8, 1), 1);
    samples(imi).orgbbind = neg_orgbbind(:);
    samples(imi).gtind    = neg_gtind(:);

  end
  samples = reshape(samples,1,nimg);

end



function newsamples = find_hard_negatives(gt_boxes, scored_bbs, S, samples); 
  
  ntrainimg = length(samples);
  
  for imi = ntrainimg:-1:1
    assert(gt_boxes(imi).orgimi == samples(imi).orgimi);

    newsamples(imi).orgimi   = samples(imi).orgimi;
    newsamples(imi).clsind   = [];
    newsamples(imi).bbuint8  = zeros(0,4,'uint8');
    newsamples(imi).orgbbind = zeros(0,1);

    cand_bbs = scored_bbs(imi).bbuint8;
    cand_bbs = double(cand_bbs);
    scores   = S(imi).scores;
    ncand    = size(scores,1);
    
    
    realgt_bbs = gt_boxes(imi).bbuint8;
    pos_bbs = samples(imi).bbuint8; % stupid.. contains negative boxes as well. M12: kept for compatibility
    % pos_bbs = [samples(imi).bbuint8(samples(imi).clsind>0,:); gt_boxes(imi).bbuint8];
    
    % windows already in the samples array
    neg_mask = samples(imi).clsind<=0; % ignore positive samples.
    neg_bbs  = samples(imi).bbuint8(neg_mask, :);

    minscoreth = -1;
    mcur       = false(ncand,1);
    
    % tolerate little overlap 
    selind = samples_nonmax_suppression(cand_bbs, scores, 0.3, -1, Inf);
    mcur(selind) = true;
        
    if ~isempty(pos_bbs)
      % max overlap with a groundtruth. dont do check against all positive samples_bbs in case possamples='gt+all'.
      O = cn_bboverlap(double(pos_bbs), cand_bbs(mcur, :));
      mcur(mcur) = max(O,[],1) < 0.3; % use < in case hardnegposmaxov=0.5
    end
    
    if ~isempty(neg_bbs)
      % max overlap with a negative window in the sample set
      % we do not check overlap with a positive sample window taken around the groundtruth because that would
      % make hard negatives weaker than we want.
      O = cn_bboverlap(double(neg_bbs),cand_bbs(mcur,:));
      mcur(mcur) = max(O,[],1) < 0.6; % use < in case negnegmaxov=1 to avoid self-replicates.
    end
    
    maxhardnegperim = 2; % max nb of hard negs. Keep highest scores
    [~,x] = sort( scores(mcur), 'descend' );
    y = false( length(x), 1 );
    y(x(1:min(length(x), maxhardnegperim))) = true;
    mcur(mcur) = y;
    
    mcand = mcur;
    nhardneg = sum(mcand);
    if nhardneg == 0
        continue;
    end
    
    % add new negatives
    newsamples(imi).bbuint8  = uint8(cand_bbs(mcand, :));
    newsamples(imi).clsind   = -ones(nhardneg, 1);
    newsamples(imi).orgbbind = colvec( find(mcand) );
    
  end
  
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Training 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function classifier = train_classifier(descriptors, pos_data, neg_data)
  npos = size(pos_data.codes, 2);
  nneg = size(neg_data.codes, 2);
  nchan = length(descriptors.channels);
  
  z2c = descriptors.z2c;
  p2.quiet = true;
  p2.costp = 1000; % C
  p2.epsilon = 0.1;
  p2.clstype = 3; % 3:Hinge loss.
  
  wnegcls = 10; % weight of neg classes
  p2.samplew = zeros(1, npos+nneg);
  p2.samplew(1:npos) = 1/(npos*(1 + wnegcls));
  p2.samplew(npos+1:npos+nneg) = wnegcls/(nneg*(1 + wnegcls));
  
  Y = double((1:npos + nneg) <= npos) * 2 - 1;
  pqcenters = {};
  for i = 1:length(descriptors.pqs)
    pqcenters{i} = descriptors.pqs(i).cent3s;
  end
  
  traindata = merge_train_data(pos_data, neg_data);
  
  classifier = cn_pqliblinear_train(pqcenters, z2c, traindata.codes, traindata.wscr, ones(1, npos + nneg), Y, p2);
  classifier.alls_pq = reshape(classifier.alls_pq, 256, []);
end


