#pragma once

/**
 * Author: R. Gokberk Cinbis.
 */


#include <vector>
#include <limits>
#include <stdexcept>
#include "omp.h"
#include <stdint.h>
#include "cn_macros.hpp"
#include "cn_threads.hpp"

// C++11: #include <cmath>
#include <math.h>

namespace cn {

    /**
     * Multithread max(vector) where length(vec) is N
     * Throw exception if N is 0.
     */
    template <class T> T maxofvector(const T * vec,uint64_t N);


    /**
     * Compute to=from.^2 where input and output arrays are pre-allocated
     */
    template <class T> void squarearray( const T *from, const uint64_t len, T *to );

    /**
     * Input:  (M N) X
     * Output: (N M) XT
     */
    template <class T> void transposematrix(const T *X,const uint64_t M,const uint64_t N,T *XT);

    /**
     * PI Constant at the desired precision.
     */
    template <class T> T pi();

    /**
     * Given in the col-major linear indices over the lower triangle (including diagonal)
     * elements in a square matrix, returns row/column indices. 
     *
     * See cn_trilind2sub for more details.
     *
     * INPUT
     * n        Size of the (n n) matrix.
     * ind      Linear index 0 <= i < (n*(n+1)/2). If it is outside of this range, result will be nonsense.
     * oi       Output row index, 0 <= oi < n
     * oj       Output col index, 0 <= oi < n
     *
     * todo: write an iterator version for C++11
     */
    template <class T>
    void trilind2sub(const double n,const double ind, T *oi, T *oj);

    /**
     * Return the linear index for the col-major storage of the lower-triangular (including diagonal)
     * (n n) symmetric matrix. Note that trilsub2ind(n,i,j)==trilsub2ind(n,j,i).
     *
     * INPUT
     * n    Size of the (n n) matrix.
     * i    Index 0 <= i < n.
     * j    Index 0 <= j < n.
     *
     * OUTPUT
     * Return linear index.
     *
     * EXAMPLE
     * For n = 4:
     *      0          
     *      1 4 
     *      2 5 7 
     *      3 6 8 9
     *
     * See cn_trilsub2ind for more details.
     */
    template <class Tsz, class Tind>
    Tsz trilsub2ind( const Tsz &n, const Tind &i,const Tind &j ) ;

    /**
     * Compute out+=X'*X.
     * Particularly useful for covariance calculation.
     * Exploits the fact that output is symmetric.
     * Increments the output by the result. If this is undesired, simply memset out to zeros.
     *
     * X:   (N D)
     * out: (D D)
     */
    template <class T>
    void mtimes_accum_XTX(const T *X,const uint64_t N,const uint64_t D,T *out) ;

    /**
     * Return maximum of two scalar values.
     */
    template <class T> inline T max(const T &x, const T &y );


    /**
     * Return minumum of two scalar values.
     */
    template <class T> inline T min(const T &x, const T &y );

    /**
     * Return +inf constant for the required type, if available.
     * inf and -inf can be treated like any other number (unlike NaN where comparisons fail).
     * Don't call within openmp loop.
     */
    template <class T> inline T infconst();

    /* Simply do dot product between two contiguously stored vectors:
     * 
     *  out += A' * B;
     *
     * A and out has to be in double. B can have any time.
     * Note that N is unsigned.
     *
     * NOTE THAT IT UPDATES "out", WITHOUT INITIALIZING TO ZERO.
     * THIS IS DONE INTENTIONALLY TO AVOID OVERHEAD OF CALLING MULTIPLE TIMES OVER SEVERAL SMALL VECTORS.
     * Defining 'out' as a reference also avoid explicitly passing template arguments.
     *
     * The code is based on the following part of ddot.f in BLAS

     M = MOD(N,5)
     IF (M.NE.0) THEN
     DO I = 1,M
     DTEMP = DTEMP + DX(I)*DY(I)
     END DO
     IF (N.LT.5) THEN
     DDOT=DTEMP
     RETURN
     END IF  
     END IF
     MP1 = M + 1
     DO I = MP1,N,5
     DTEMP = DTEMP + DX(I)*DY(I) + DX(I+1)*DY(I+1) + DX(I+2)*DY(I+2) + DX(I+3)*DY(I+3) + DX(I+4)*DY(I+4)
     END DO
     */
    template <class Ta,class Tb,class To> inline void ddot(const size_t &N,const Ta *A,const Tb *B,To &out); // SEE THE NOTE ABOUT "out" ABOVE.
}

/*** IMPLEMENTATIONS ****/


template <class T> T cn::maxofvector(const T * vec,uint64_t N) {
    std::vector<T> temps;
    uint64_t nthread;
    T minval;

    if ( N < 1 )
        throw std::runtime_error("Cannot find max() of empty vector");

#pragma omp parallel 
    {

        uint64_t thi=omp_get_thread_num();

        // initailize temps to minimum possible value
        if (thi == 0) {
            nthread = omp_get_num_threads();

            if (std::numeric_limits<T>::has_infinity)
                minval = -std::numeric_limits<T>::infinity();
            else
                minval = std::numeric_limits<T>::min();

            temps.reserve( MIN(nthread,N) );
            for (uint64_t j = 0; j < nthread; ++j) 
                temps.push_back( minval );
        }
#pragma omp barrier

        T x = minval;
        uint64_t endi = jobendi(N);
        for (uint64_t i = jobstarti(N); i < endi; ++i) {
            if ( vec[i] > x )
                x = vec[i];
        }
        temps[thi] = x;

    } // end of parallel 

    T final = temps[0];
    for (uint64_t j = 1; j < temps.size(); ++j) {
        if ( temps[j] > final )
            final = temps[j];
    }

    return final;
    
}


template <class T> void cn::squarearray( const T *from, const uint64_t len, T *to ) {
#pragma omp parallel for
    for (uint64_t d = 0; d < len; ++d) {
        const T &x = from[d];
        to[d] = x * x;
    }
}


template <class T> void cn::transposematrix(const T *X,const uint64_t M,const uint64_t N,T *XT) {
    for (uint64_t j = 0; j < N; ++j) {
        T *row = XT + j;
        for (uint64_t i = 0; i < M; ++i) {
            *row = *(X++);
            row += N;
        }
    }
}

template <class T> T cn::pi() {
    static const double pi_d = atan(1)*4; // MUST be computed in double.
    return (T) pi_d;
}


template <class T>
void cn::trilind2sub(const double n,const double ind, T *oi, T *oj) {
    double delta = 4 * n * (n + 1) + 1 - 8 * ind; 
    double j = floor( (2 * n + 1 - sqrt(delta)) / 2 ); // get column in c coordinates
    double i = ind - j * n + j * (j - 1) / 2; // get row within lower triangle
    i = round(i + j); // push row from top to convert into full matrix row index. round to avoid precision errors.
    *oi = (T) i;
    *oj = (T) j;
}


template <class Tsz, class Tind>
Tsz cn::trilsub2ind( const Tsz &n, const Tind &i,const Tind &j ) {
    Tind r, c;
    if ( i > j ) {
        r = i;
        c = j;
    } else {
        r = j;
        c = i;
    }
    return c*( (2*n-1) - c )/2+r;
}

template <class T>
void cn::mtimes_accum_XTX(const T *X,const uint64_t N,const uint64_t D,T *out) {

    uint64_t ntril = D * (D+1) / 2;

    // there is no significant memory overhead to multiprocess here.
#pragma omp parallel for
    for ( uint64_t ti = 0; ti < ntril; ++ti) 
    {
        // we'll process out(ix,iy)=out(iy,ix)
        uint64_t ix, iy;
        cn::trilind2sub(D,ti,&ix,&iy);

        // --debug-- if ( (ix >= D) || (iy >= D) ) { throw std::runtime_error("ix or iy is invalid"); }

        // get pair of vectors of length (N). 
        const T *_x = X + N * ix;
        const T *_y = X + N * iy;

        T _o = 0;
        for ( uint64_t j = 0; j < N; ++j) {
            _o += *(_x++) * *(_y++);
        }
        out[ D * ix + iy ] += _o; 
        if (ix != iy) { // needed
            out[ D * iy + ix ] += _o; // symmetric pair
        }
    }

}


/**
 * Return maximum of two scalar values.
 */
template <class T> inline T cn::max(const T &x, const T &y ) {
    if ( x < y ) 
        return y;
    return x;
}


/**
 * Return minumum of two scalar values.
 */
template <class T> inline T cn::min(const T &x, const T &y ) {
    if ( x < y ) 
        return x;
    return y;
}


template <class T> inline T cn::infconst()
{
    if ( !(std::numeric_limits<T>::has_infinity) )
        throw std::runtime_error("Requested class does not have Inf");
    return std::numeric_limits<T>::infinity();
}


template <class Ta,class Tb,class To>
inline void cn::ddot(const size_t &N,const Ta *A,const Tb *B,To &out)  // SEE THE NOTE ABOUT "out" ABOVE.
{
#ifdef CN_DDOT_NAIVE
#warning In cn_matlabarrays.hpp, naive, reference implementation for cn_ddot is selected.

    // naive implementation
    for (mwIndex i = 0; i < N; ++i) {
        out += *(A++) * *(B++);
    }

#else
    // This version is 2x faster for vectors of at least 1000 dimensions.
    // For very small vectors, it has neglicible overhead.

    // default implementation
    size_t M = N % 5;

    // consume a few indices so that length becomes divisible by 5.
    //for (mwIndex i = 0; i < M; ++i) {
    const Ta *e = A+M;
    while(A < e) {
        out += *(A++) * *(B++);
    }

    // do efficient processing using unrolled loops in chunks of size 5
    //for (mwIndex i = M; i < N; i+=5) {
    e = A+(N-M);
    while(A<e) {
        out += A[0]*B[0] + A[1]*B[1] + A[2]*B[2] + A[3]*B[3] + A[4]*B[4];
        A += 5;
        B += 5;
    }

#endif
}


