
/**
 * Author: R. Gokberk Cinbis
 */



#include "mex.h"
#include "matrix.h"
#include <string.h>
#include <string>
#include <stdint.h> // to get uint8_t
#include <cn_macros.hpp>
#include <cn_matlabarrays.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

using std::string;

#define a_im        0
#define a_sz        1
#define a_method    2

void mexFunction(int , mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nrhs != 3)
        mexErrMsgTxt("Invalid nargin");

    // input dims
    if (mxGetNumberOfDimensions(prhs[a_im]) > 3) {
        mexErrMsgTxt("im should have at most 3 dimensions");
    }
    mwSize imH = cn::mex::getdim(prhs[a_im],1);
    mwSize imW = cn::mex::getdim(prhs[a_im],2);
    mwSize nband = cn::mex::getdim(prhs[a_im],3);

    // output dims
    if (mxGetNumberOfElements(prhs[a_sz])!=2)
        mexErrMsgTxt("sz array should contain [height width]");
    const double *m_outhw = cn::mex::getDenseData<double>(prhs[a_sz]);
    mwSize outH = m_outhw[0];
    mwSize outW = m_outhw[1];

    // data type
    mxClassID imtype = mxGetClassID(prhs[a_im]);
    mwSize elbytes = cn::mex::sizeofMxClassID(imtype);
    int ocv_type;
    if (imtype==mxUINT8_CLASS) {
        ocv_type = CV_8UC1;
    } else if (imtype==mxDOUBLE_CLASS) {
        ocv_type = CV_64FC1;
    } else if (imtype==mxSINGLE_CLASS) {
        ocv_type = CV_32FC1;
    } else {
        mexErrMsgTxt("not supported data type");
    }

    // allocate output image
    mxArray *outarray = cn::mex::createArrayWoInit(imtype,outH,outW,nband);
    plhs[0] = outarray;

    // data pointers
    const uint8_t *imdata = (uint8_t *) mxGetData(prhs[a_im]);
    uint8_t *outdata = (uint8_t *) mxGetData(outarray);

    // interpolation method
    char *methodstr = mxArrayToString(prhs[a_method]);
    int interpolation;
    if ( strcmp(methodstr,"cvNearest")==0 ) {
        interpolation = cv::INTER_NEAREST;
    } else if ( strcmp(methodstr,"cvLinear")==0 ) {
        interpolation = cv::INTER_LINEAR;
    } else if ( strcmp(methodstr,"cvArea")==0 ) {
        interpolation = cv::INTER_AREA;
    } else if ( strcmp(methodstr,"cvCubic")==0 ) {
        interpolation = cv::INTER_CUBIC;
    } else if ( strcmp(methodstr,"cvLanczos4")==0 ) {
        interpolation = cv::INTER_LANCZOS4;
    } else {
        interpolation = 0; // avoid warning.
        mexErrMsgTxt((string("Unknown method string") + string(methodstr)).c_str());
    }
    mxFree(methodstr);

    if (outH*outW*nband == 0)
        return; // nothing to process

    // process each band separately
    for (mwSize b = 0; b < nband; ++b ) {

        // here we exploit openCV's nice memory allocation interface.
        // this is a safe and recommended way of handling this situtaion in openCV documation.
        //
        // resize() will call Mat.create()
        // Since Mat already has the correct data array contained, it will not do any new allocation.
        // Since we have passed already allocated memory to Mat at construction, it will not do deallocation
        // either.
        //
        // we modify each band in transposed format since opencv uses row-major ordering.

        // wrap the band in opencv
        cv::Mat src(imW,imH,ocv_type,(void *) (imdata+imH*imW*elbytes*b));
        cv::Mat dst(outW,outH,ocv_type,(void *) (outdata+outH*outW*elbytes*b));
        cv::resize(src,dst,dst.size(),0,0,interpolation);

    }

}





